require 'middleclass'

local config = require 'system_config'

local ok, curl = pcall(require, 'curl')
if not ok then
    local ffi = require 'ffi'
    cl = ffi.load('libcurl.so')

    ffi.cdef[[
        void *curl_easy_init(void);
        int curl_easy_setopt(void *, int, ...);
        int curl_easy_perform(void *);

        struct curl_slist {
            char                *data;
            struct curl_slist   *next;
        };

        struct curl_slist *curl_slist_append(struct curl_slist *list, const char *string);
        void curl_slist_free_all(struct curl_slist *list);

        typedef int      size_t;

        typedef size_t (*rcb)(char *buffer, size_t size, size_t nitems, void *instream);
        typedef size_t (*wcb)(char *ptr, size_t size, size_t nmemb, void *userdata);
    ]]

    local N = class('curllib')

    function N:initialize()
        self.handle = cl.curl_easy_init()
    end
    function N:setopt(option, parameter)
        if option == curl.OPT_READFUNCTION then
            return cl.curl_easy_setopt(self.handle, option,
                ffi.new('rcb', function(b, s, n, i)
                        local l, s = parameter(s * n)
                        ffi.copy(b, s, l)
                        return l
                    end))
        elseif option == curl.OPT_WRITEFUNCTION then
            return cl.curl_easy_setopt(self.handle, option,
                ffi.new('wcb', function(b, s, n, i)
                        local len = s * n
                        local l, s = parameter(ffi.string(b, len), len)
                        return l
                    end))
        elseif option == curl.OPT_HTTPHEADER then
            local slist = ffi.new('struct curl_slist *')
            for _, v in ipairs(parameter) do
                slist = cl.curl_slist_append(slist, v)
            end
            local ret = cl.curl_easy_setopt(self.handle, option, slist)
--[[
            if slist ~= nil then
                cl.curl_slist_free_all(slist)
            end
]]
            return ret
        end
        if type(parameter) == 'number' then
            parameter = ffi.new('unsigned long', parameter)
        end
        return cl.curl_easy_setopt(self.handle, option, parameter)
    end
    function N:perform()
        local ret = cl.curl_easy_perform(self.handle)
        return ret, ret ~= 0 and 'ERROR'
    end

    curl = {
        easy_init = function() return N:new() end,
        OPT_READFUNCTION = 20012,
        OPT_WRITEFUNCTION = 20011, 
        OPT_URL = 10002,
        OPT_TIMEOUT = 13,
        OPT_HTTPHEADER = 10023,
        OPT_CUSTOMREQUEST = 10036,
        OPT_SSL_VERIFYHOST = 81,
        OPT_SSL_VERIFYPEER = 64,
        OPT_UPLOAD = 46,
        OPT_INFILESIZE = 14,
    }
end
local cjson = require 'cjson'

local M = class('db_client')

local function url_escape(s)
    return string.gsub(s, '([^A-Za-z0-9_])', function(c)
        return string.format('%%%02x', string.byte(c))
    end)
end

function M:initialize(settings)
    self:set(settings)

    self.cu = curl.easy_init()
    self.tx = nil
    self.rx = nil

    self.cu:setopt(curl.OPT_READFUNCTION,
        function(len)
            if len < string.len(self.tx) then
                local s = string.sub(self.tx, 1, len)
                self.tx = string.sub(self.tx, len + 1)
                return len, s
            end
            local s = self.tx
            self.tx = ''
            return string.len(s), s
        end
    )
    self.cu:setopt(curl.OPT_WRITEFUNCTION,
        function(s, len)
            if len > 0 then
                if self.ofh then
                    self.ofh:write(s)
                else
                    self.rx[#self.rx + 1] = s
                end
            end
            return len, nil
        end
    )
end

function M:set(settings)
    for k, v in pairs(settings) do
        self[k] = v
    end
end

function M:http(url, opt)
    opt = opt or {}

    -- Default query stuff
    if self.query then
        opt.query = opt.query or {}
        for k, v in pairs(self.query) do
            if opt.query[k] == nil then
                opt.query[k] = v
            end
        end
    end

    if type(opt.data) == 'table' then
        opt.data = cjson.encode(opt.data)
    end
    self.tx = opt.data or ''

    local headers = {
    }
    local query = ''
    if opt.query then
        local t = {}
        for k, v in pairs(opt.query) do
            t[#t + 1] = url_escape(k) .. '=' .. url_escape(v)
        end
        query = '?' .. table.concat(t, '&')
    end
    local curl_opts = {
        OPT_URL = string.format('%s%s', url, query),
        OPT_TIMEOUT = 10,
        OPT_HTTPHEADER = headers,
        OPT_CUSTOMREQUEST = opt.method,
        OPT_SSL_VERIFYHOST = 0,
        OPT_SSL_VERIFYPEER = 0,
        OPT_UPLOAD = opt.data and 1 or 0,
    }

    if opt.data then
        curl_opts.OPT_INFILESIZE = string.len(self.tx)
    else
        headers[#headers + 1] = 'Content-Length: ' .. string.len(self.tx)
    end

    if opt.filename then -- KLUDGE allow Content-disposition???
        self.ofh = io.open(opt.filename, 'wb')
        curl_opts.OPT_TIMEOUT = 0 -- File transfers can be BIG
    end

    for k, v in pairs(opt.curl or {}) do
        curl_opts[k] = v
    end

    for k, v in pairs(curl_opts) do
        self.cu:setopt(curl[k], v)
    end

    self.rx = {}
    local got, err = self.cu:perform()
    if got ~= 0 then
        error(err)
    end

    if opt.filename and self.ofh then
        assert(self.ofh:close())
        self.ofh = nil
        -- KLUDGE check size here!!!
        return {ok=true}
    end

    local j = table.concat(self.rx, '')
    local v = nil
    pcall(function() v = cjson.decode(j) end)
    return v, j
end

function M:request(method, path, opt)
    opt = opt or {}
    opt.method = method
    local url
    if config.device_type == 'edaq2' then
        url = string.format('http://%s%s', self.host, path)
    else
        url = string.format('https://%s%s', self.host, path)
    end
    return self:http(url, opt)
end

function M:raw_db(method, path, data)
    return self:request(method, '/api/raw_db/' .. path, {data=data})
end

function M:tracked_db(method, path, data)
    return self:request(method, '/api/tracked_db/' .. path, {data=data})
end

return M
