local __ = require '__'
local log = require 'log'
local ev = require 'ev'
local ts = require 'timestamps'
local sample_clock = require 'sample_clock'
local cca_handler = require 'cca_handler'
local channel_proxy = require 'channel_proxy'

local aux = require 'aux'

local update_rate = 10 -- Hz

local dumper = require 'data_dumper'

local ffi = require 'ffi'

local C = ffi.C

ffi.cdef[[
struct proxy_aux_msg {
    uint32_t id;
    uint8_t msg[8];
    uint8_t ff;
};
]]

local proxy_aux_msg = ffi.typeof('struct proxy_aux_msg')

local function dummy_message_sink(loop, timer)
    local self = timer.proxy
    if self.aux_port_mode == 'edaq_sync' then
        local connector = '@' .. self.proxy_hostname
        self.ts = ts.ts_to_stamp(ts.get_ts())
        self.sync_clock_data[0] = self.sync_clock_rate
        if self.channel_sinks[connector] then
            self.channel_sinks[connector](self.ts - self.start_time,
                                          self.sync_clock_data, 8)
        end
    end
end

local aux_proxy = class('aux_proxy', channel_proxy)

function aux_proxy:initialize(name, devname)
    cca_handler.new(devname, self, aux, 'VBM_0.bin', sample_clock.master_rate)
    self.proxy_type = 'aux'
    self.streaming = false
    self.ts = 0
    self.hb_ts = 0
    self.channels = {}
    self.channel_sinks = {}
    self.aux_port_mode = 'edaq_sync'
    self.sample_rate_domain = 'decimal'
    self.sync_clock_data = ffi.new('double [1]')
    self:cmd_block('txdr_power_on')
    table.insert(self.channels, {
        name = 'raw',
        description = 'Raw Vehicle Bus Message',
        electrical_units = '',
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/msg_aux',
    })
    self.channels['raw'] = self.channels[1]
    channel_proxy.initialize(self, name)
    self.timer = ev.Timer.new(dummy_message_sink, 0.1, 0.1)
    self.timer.proxy = self
end

function aux_proxy:buffer(msg)
    if not self.streaming then
        -- Idle
        return
    end
    local last_non_hb = msg.element_count - (msg.heartbeat and 2 or 1)
    if msg.cmd == C.VBM_RX_START then
        self.ts = msg.bufdata[last_non_hb + 1].ts
        if not self.ts then return end
        if msg.heartbeat then
        end
        -- All other channels processed here
    elseif msg.cmd == C.VBM_EVENT_START then
    end
end

function aux_proxy:start_stream(w, spec)
    w.proxy = self
    if not self.streaming then
        self.start_time = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
        if self:setup_sync(spec) then
            if self.sample_rate_domain == 'binary' then
                self.start_time = self.start_time - 50004476
            else
                self.start_time = self.start_time - 40004000
            end
            self:cmd('setup_sync', self.sync_clock_rate * 2, self.start_time)
        end
        self:cmd('event_setup', {cable_plugged=true, cable_unplugged=true})
        self:cmd('event_start', self.start_time, update_rate, 16, 128)
        if self.aux_port_mode ~= 'edaq_sync' then
            self.hardware_interface = spec.hardware_interface
            self.bit_rate = spec.bit_rate
            self.internal_termination = spec.internal_termination
            self:cmd('rx_start', self.start_time, update_rate, 16, 128)
        end
        log:notice(__"AUX: Started in ({0}) Sample Rate Domain",
                   self.sample_rate_domain)
        self.streaming = true
        self.timer:start(ev.Loop.default)
    end
end

function aux_proxy:stop_stream(w)
    if self.streaming then
        self.streaming = false
        self:cmd('event_stop')
        if self.aux_port_mode == 'edaq_sync' then
            self:cmd('setup_sync', 0, 0)
        else
            self:cmd('rx_stop')
        end
        self.hb_ts = 0
        self.ts = 0
        self.timer:stop(ev.Loop.default)
    end
end

function aux_proxy:start_comet_stream(w, channels, sink)
-- required channel_proxy callback function
end

function aux_proxy:start_channel_stream(w, spec, md_sink, sink)
    local name = spec.name:match('(%a+)_%a+')
    if self.channels[name] then
        table.insert(self.channel_sinks, sink)
        self.channel_sinks[spec.connector] = sink
        md_sink(spec)
        w.cleanup = function()
            self.channel_sinks[spec.connector] = nil
            for i = #self.channel_sinks, 1, -1 do
                if self.channel_sinks[spec.connector] == sink then
                    table.remove(self.channel_sinks, i)
                end
            end
            self:stop_stream()
            self.start_callback = nil
        end
        self.start_callback = function(self, start_time)
            spec.start_time = start_time
            self:start_stream(w, spec)
        end
        self:announce_ready(spec.connector)
    end
end

function aux_proxy:available_channels(hostname)
    local o = {}
    for _, ch in ipairs(self.channels) do
        if ch.name == 'raw' then
            ch.connector = '@' .. self.proxy_hostname
        else
            ch.connector = ch.name .. '@' .. self.proxy_hostname
        end
        table.insert(o, {
            name = ch.name .. '_' .. hostname:gsub('%W', '_'),
            connector = ch.connector,
            description = ch.description,
            electrical_units = ch.electrical_units,
            data_type = ch.data_type,
            category = ch.category,
        })
    end
    return o
end

function aux_proxy:setup_sync(spec)
    self.sample_rate_domain = spec.sample_rate_domain or self.sample_rate_domain
    self.aux_port_mode = spec.aux_port_mode or self.aux_port_mode
    if self.aux_port_mode ~= 'edaq_sync' then
        -- sync clock not required
        log:notice(__"AUX: Sync Clock Disabled")
        return false
    end
    if self.sample_rate_domain == 'binary' then
        self.sync_clock_rate = 98304
    elseif self.sample_rate_domain == 'decimal' then
        self.sync_clock_rate = 100000
    else
        -- unknown sample rate domain
        return false
    end
    log:notice(__"AUX: Sync Clock Enabled @ {0} Hz", self.sync_clock_rate)
    return true
end

return aux_proxy
