--
-- AXIS HTML (old) 5.2.4.6 MJPG video CGI reguest
--
-- Method: GET
-- Syntax:
--     http://<camera_ip>/axis-cgi/mjpg/video.cgi
--             [?<parameter>=<value>[&<parameter>=<value>...]]
--
-- Parameter Pairs
-- ---------------------------------------------------------------------
-- [ resolution=<string> ]
--     Values: 640x480, 480x360, 320x240, 240x180, 176x144, 160x120
-- [ camera=<string> ]
--     Values: 1, 2, 3, 4, and quad
-- [ compression=<int> ]
--     Values: 0 - 100 (where 0 = largest image size / best quality)
-- [ fps=<int> ]
--     Values: 0, ... (where 0 = unlimited or [30])
--
-- FIXME: The remaining CGI API from documentation

local __ = require '__'
local log = require 'log'
local class = require 'class'
local camera = require 'camera/common'
local net = require 'net_ffi'
local ffi = require 'ffi'
local C = ffi.C
local cjson = require 'cjson'

ffi.cdef[[
    typedef int32_t     pid_t;

    // DNS types
    enum {
        DNS_TYPE_A = 1,
        DNS_TYPE_NS = 2,
        DNS_TYPE_CNAME = 5,
        DNS_TYPE_SOA = 6,
        DNS_TYPE_PTR = 12,
        DNS_TYPE_HINFO = 13,
        DNS_TYPE_MX = 15,
        DNS_TYPE_TXT = 16,
        DNS_TYPE_AAAA = 28,
        DNS_TYPE_SRV = 33,
        DNS_TYPE_OPT = 41,
        DNS_TYPE_SSHFP = 44,
        DNS_TYPE_NSEC = 47,
        DNS_TYPE_NSEC3 = 50,
        DNS_TYPE_AXFR = 252,
    };

    // DNS Classes
    enum {
        DNS_CLASS_IN = 1,
        DNS_CLASS_CH = 3,
        DNS_CLASS_ANY = 255
    };

    // Flags, MSB staring bit number in flags
    enum {
        DNS_FLAG_RA     = 0x80,
        DNS_FLAG_RD     = 0x100,
        DNS_FLAG_TC     = 0x200,
        DNS_FLAG_AA     = 0x400,
        DNS_FLAG_QR     = 0x8000,
    };

    // Opcode stuff
    enum {
        DNS_OPCODE_QUERY = 0,
        DNS_OPCODE_IQUERY,
        DNS_OPCODE_STATUS,
        DNS_OPCODE_UNASSIGNED,
        DNS_OPCODE_NOTIFY,
        DNS_OPCODE_UPDATE,
        DNS_OPCODE_MAX,
        DNS_OPCODE_SHIFT        = 11,
        DNS_OPCODE_MASK         = 0xf,
    };

    // Response code stuff
    enum {
        DNS_RCODE_NOERROR = 0,
        DNS_RCODE_FORMERR,
        DNS_RCODE_SERVFAIL,
        DNS_RCODE_NXDOMAIN,
        DNS_RCODE_NOTIMP,
        DNS_RCODE_REFUSED,
        DNS_MAX_RCODE,
        DNS_RCODE_MASK  = 0xf,
    };
        

    typedef struct dns_header {
        uint16_t    id;             /* identification number */
        uint16_t    flags;          /* flags */
        uint16_t    qdcount;        /* number of question entries */
        uint16_t    ancount;        /* number of answer entries */
        uint16_t    nscount;        /* number of authority entries */
        uint16_t    arcount;        /* number of resource entries */
    } dns_header_t;

    typedef struct dns_question {
        uint16_t qtype;
        uint16_t qclass;
    } dns_question_t;

    typedef struct dns_resource_data {
        uint16_t type;
        uint16_t class;
        unsigned int   ttl;
        uint16_t len;
    } dns_resource_data_t;

    typedef struct dns_resource_record {
        unsigned char            name[256];
        struct dns_resource_data *resource;
        unsigned char            *data;
    } dns_resource_record_t;

    typedef struct dns_query {
        unsigned char       *name;
        struct dns_question *question;
    } dns_query_t;

    in_addr_t   inet_addr(const char *cp);
    char        *strerror(int);
    size_t      strlen(const char *s);
    pid_t       getpid(void);
    char        *inet_ntoa(struct in_addr in);

    void *malloc(int);
]]

local m = class(camera)

local credentials = {
    username = 'root',
    password = 'insecure',
    base64 = 'cm9vdDppbnNlY3VyZQ==' -- base64 of "root:insecure"
}

function m:setup(vendor, model, serial, host, port)
    camera:setup(vendor, model, serial, host, port)
    local query = '?' ..
        'compression=' .. self.camera.spec_image_compression .. '&' ..
        'resolution=' ..  self.camera.spec_image_size .. '&' ..
        'fps=' ..  self.camera.spec_frame_rate

    if self.camera.query ~= query then self.camera.query = query end
    self.camera.page = 'axis-cgi/mjpg/video.cgi'
    return self
end

local function decode_string(dns_header, data)
    local function dec(dns_header, data, limit)
        local partlen
        local parts = {}
        local part

        limit = limit or 10
        if limit < 0 then return data, nil end

        partlen, data = ffi.cast('uint8_t *', data)[0],
            data + ffi.sizeof('uint8_t')
        while partlen ~= 0 do
            if partlen < 64 then
                part = ffi.string(data, partlen)
                data = data + partlen
                if not part then return data end
                parts[#parts + 1] = part
                partlen, data = ffi.cast('uint8_t *', data)[0],
                    data + ffi.sizeof('uint8_t')
            else
                partlen, data = ffi.cast('uint16_t *', data - 1)[0], data + 1
                local offset = partlen - 0xc000
                local _, part = dec(dns_header,
                    ffi.cast('uint8_t *', dns_header) + offset, limit - 1)
                if not part then return data end
                parts[#parts + 1] = part
                partlen = 0
            end
        end
        return data, table.concat(parts, '.')
    end
    return dec(dns_header, data)
end

local decoder = {}

decoder[C.DNS_TYPE_A] = function(entry)
    local ptr = ffi.cast('uint8_t *', entry.data)
    entry.ip = string.format("%d.%d.%d.%d", ptr[0], ptr[1], ptr[2], ptr[3])
end

decoder[C.DNS_TYPE_AAAA] = function(entry)
    local ptr = ffi.cast('uint16_t *', entry.data)
    entry.ipv6 = string.format("%x:%x:%x:%x:%x:%x:%x:%x", ptr[0], ptr[1],
        ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7])
end

decoder[C.DNS_TYPE_SSHFP] = function(entry)
    entry.SSHFP = {
        algorithm = ffi.cast('uint8_t *', entry.data)[0],
        fptype = ffi.cast('uint8_t *', entry.data)[1],
        fingerprint = ffi.cast('uint16_t *', entry.data)[1],
    }
end

decoder[C.DNS_TYPE_SOA] = function(entry, dns_header, data)
    local np = data - ffi.sizeof('dns_header_t')

    entry.SOA = {}
    np, entry.SOA.mname = decode_string(dns_header, np)
    np, entry.SOA.rname = decode_string(dns_header, np)
    local v = ffi.cast('uint16_t *', np)
    entry.SOA.serial, entry.SOA.refresh, entry.SOA.retry, entry.SOA.expire,
        entry.SOA.minimum = v[0], v[1], v[2], v[3], v[4]
end

local function decode_domain(entry, dns_header, data)
    if type(data) == 'string' then
        entry.domain = decode_string(dns_header, data - #entry.data)
    else
        entry.domain = decode_string(dns_header, data - entry.data_len)
    end
end

decoder[C.DNS_TYPE_CNAME] = decode_domain
decoder[C.DNS_TYPE_NS] = decode_domain
decoder[C.DNS_TYPE_PTR] = decode_domain

decoder[C.DNS_TYPE_TXT] = function(entry, dns_header, data)
    local len = entry.data:len()
    local np = data - #entry.data
    local txt_len, txt

    if len > 0 then
        entry.TXT = {}
        entry.TXT.text = {}
    end

    while len > 0 do
        txt_len, np = ffi.cast('uint8_t *', np)[0], np + ffi.sizeof('uint8_t')
        txt, np = ffi.string(np, txt_len), np + txt_len
        len = len - txt_len - 1
        entry.TXT.text[#entry.TXT.text + 1] = txt
   end
end

decoder[C.DNS_TYPE_MX] = function(entry, dns_header, data)
    local np = data - #entry.data + 2
    local _
    entry.MX = {}
    _, entry.MX.pref = ffi.cast('uint16_t *', entry.data)[0]
    _, entry.MX.server = decode_string(dns_header, np)
end

decoder[C.DNS_TYPE_SRV] = function(entry, dns_header, data)
    local np = data - #entry.data
    entry.SRV = {
        prio = ffi.cast('uint16_t *', np)[0],
        weight = ffi.cast('uint16_t *', np)[1],
        port = ffi.cast('uint16_t *', np)[2]
    }
    np = np + (ffi.sizeof('uint16_t') * 3)
    np, entry.SRV.target = decode_string(dns_header, np)
end

local function decode_rr(dns_header, count, data)
    local ans = {}
    for i = 1, count do
        local crr = {}
        data, crr.name = decode_string(dns_header, data)
        crr.type, crr.class, crr.ttl, data = ffi.cast('uint16_t *', data)[0],
            ffi.cast('uint16_t *', data)[1], ffi.cast('uint32_t *', data)[1],
            data + 8

        local reslen
        reslen, data = ffi.cast('uint16_t *', data)[0],
            data + ffi.sizeof('uint16_t')
        crr.data = ffi.string(data, reslen) or data
        data = data + reslen
        crr.data_len = reslen

        if decoder[crr.type] then
            decoder[crr.type](crr, dns_header, data)
        end
        ans[#ans + 1] = crr
    end
    return data, ans
end

local function decode_questions(dns_header, count, data)
    local q = {}
    for i = 1, count do
        local cq = {}
        data, cq.name = decode_string(dns_header, data)
        local d = ffi.cast('uint16_t *', data)
        cq.type, cq.class = d[0], d[1]
        data = data + (ffi.sizeof('uint16_t') * 2)
        q[#q + 1] = cq
    end
    return data, q
end

local function dns_decode(dns_header)
    local pkt = {
        id = dns_header.id,
        flags = {
            QR = bit.band(dns_header.flags, C.DNS_FLAG_QR) == C.DNS_FLAG_QR,
            AA = bit.band(dns_header.flags, C.DNS_FLAG_AA) == C.DNS_FLAG_AA,
            TC = bit.band(dns_header.flags, C.DNS_FLAG_TC) == C.DNS_FLAG_TC,
            RD = bit.band(dns_header.flags, C.DNS_FLAG_RD) == C.DNS_FLAG_RD,
            RA = bit.band(dns_header.flags, C.DNS_FLAG_RA) == C.DNS_FLAG_RA,
            opcode = bit.band(bit.rshift(
                dns_header.flags, C.DNS_OPCODE_SHIFT), C.DNS_OPCODE_MASK),
            rcode = bit.band(dns_header.flags, C.DNS_RCODE_MASK),
        },
    }
    if pkt.opcode == C.DNS_OPCODE_UPDATE then return pkt end
    local ptr = ffi.cast('uint8_t *', dns_header + 1)
    ptr, pkt.questions = decode_questions(dns_header, dns_header.qdcount, ptr)
    ptr, pkt.answers = decode_rr(dns_header, dns_header.ancount, ptr)
    ptr, pkt.auth = decode_rr(dns_header, dns_header.nscount, ptr)
    ptr, pkt.add = decode_rr(dns_header, dns_header.arcount, ptr)
    return pkt
end

local on = ffi.cast('int *', C.malloc(ffi.sizeof('int')))
local ttl = on
local addr = ffi.cast('struct sockaddr_in *',
    C.malloc(ffi.sizeof('struct sockaddr_in')))
local mlen = ffi.sizeof('struct sockaddr_in')
local buf = ffi.cast('unsigned char *', C.malloc(65536))
local buf_size = 65536
local header = ffi.cast('struct dns_header *', buf)
local qname = ffi.cast('unsigned char *', header + 1)
local hsize = ffi.sizeof('struct dns_header')
local qsize = ffi.sizeof('struct dns_question')
local slen = ffi.cast('int *', C.malloc(ffi.sizeof('int')))
local sd

function m:probe(loop, timeout, cb)
    local rc
    local o = {}

    if not sd then
        -- Create the socket
        sd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
        if sd < 0 then
            log:log(log.ERR, string.format("axis_camera:mdns(): socket %s",
                ffi.string(C.strerror(ffi.errno()))))
            sd = nil
            cb{}
            return
        end

        -- Disable multicast loop back
        on[0] = 0
        rc = C.setsockopt(sd, C.IPPROTO_IP, C.IP_MULTICAST_LOOP, on,
            ffi.sizeof('int'))
        if rc < 0 then
            log:log(log.ERR,
                string.format("axis_camera:mdns(): IP_MULTICAST_LOOP %s",
                ffi.string(C.strerror(ffi.errno()))))
            C.close(sd)
            sd = nil
            cb{}
            return
        end

        -- Make multicast go through routers
        ttl[0] = 64
        rc = C.setsockopt(sd, C.IPPROTO_IP, C.IP_MULTICAST_TTL, ttl,
            ffi.sizeof('int'))
        if rc < 0 then
            log:log(log.ERR,
                string.format("axis_camera:mdns(): IP_MULTICAST_TTL %s",
                ffi.string(C.strerror(ffi.errno()))))
            C.close(sd)
            sd = nil
            cb{}
            return
        end
    end

    -- Function called when it is time to check for responses
    ev.Timer.new(function(l, t, e)
            t:stop(l)
            local info = {}
            repeat
                slen[0] = ffi.sizeof('struct sockaddr_in')
                local amt = C.recvfrom(sd, buf, buf_size, C.MSG_DONTWAIT,
                    ffi.cast('struct sockaddr *', addr), slen)
                if amt > 0 then

                    local response = dns_decode(ffi.cast('dns_header_t *', buf))
                    if not response.flags.QR then
                        -- Isn't a response...don't bother with it
                        break
                    end
                    for _, ans in ipairs(response.answers) do
                        if ans.name and
                            ans.name == '_axis-video._tcp.local' then
                            local i = {}
                            for _, add in ipairs(response.add) do
                                if add.ip and add.ip:find('^169%.') then
                                    i.ip = add.ip
                                end
                                if add.TXT and add.TXT.text then
                                    for _, str in ipairs(add.TXT.text) do
                                        i.serial = str:match('macaddress=(%w+)')
                                        if i.serial then break end
                                    end
                                end
                            end
                            if i.ip and i.serial then
                                i.type = 'axis'
                                info[#info + 1] = i
                            end
                        end
                    end
                elseif ffi.errno() == C.EAGAIN then
                    amt = 0
                end
            until amt <= 0
            cb(info)
        end, timeout, 0):start(loop)

    -- build DNS header
    ffi.fill(header, hsize, 0)
    header.id = C.htons(math.random())
    header.qdcount = C.htons(1)

    -- build DNS question
    ffi.copy(qname, "\011_axis-video\004_tcp\005local\0")
    local question = ffi.cast('struct dns_question *',
        qname + C.strlen(qname) + 1)
    question.qtype = C.htons(C.DNS_TYPE_PTR)
    question.qclass = C.htons(0x8001)

    -- Build multicast address to send the DNS question to
    ffi.fill(addr, ffi.sizeof('struct sockaddr_in'), 0)
    addr.sin_family = C.AF_INET
    addr.sin_port = C.htons(5353)
    addr.sin_addr.s_addr = C.inet_addr('224.0.0.251')

    -- Send the question
    rc = C.sendto(sd, buf, hsize + C.strlen(qname) + 1 + qsize, 0,
        ffi.cast('struct sockaddr *', addr), mlen)
    if rc < 0 then
        log:log(log.ERR, string.format("axis_camera:mdns(): sendto %s",
            ffi.string(C.strerror(ffi.errno()))))
        C.close(sd)
        cb{}
    end
end

function m:_http_request_stream(http_info, callback, ecb, data)
    local s = net.socket{
        ip = http_info.host,
        port = http_info.port or 80,
        family = C.AF_INET,
        type = C.SOCK_STREAM,
        flags = 0}

    local ok, socket = pcall(s.connect, s)
    if not ok then
        log:log(log.ERR, string.format("Can not connect to %s on port %s: %s",
            http_info.host, http_info.port or 80, socket))
        ecb("Unable to connect to video stream")
        return
    end
    local state = 'wait4response'
    local sbuf = net.StreamBuffer:new(socket)
    self.sbuf = sbuf
    local boundary_marker
    local frame_size
    local image_info = {}
    local header = {}
    sbuf.read_cb = function(loop)
        local avail, buf = sbuf:read()
        local avail_save = avail
        if avail then
            if state == 'wait4response' then
                local s, e = ffi.string(buf, avail):find('\r\n\r\n')
                avail = 0
                if s and e then
                    local str = ffi.string(buf, e)
                    for line in str:gmatch('[^\r\n]+') do
                        local code, msg =
                            line:match('^HTTP/1%.[01] (%d+) (.+)$')
                        if code and msg then
                            if tonumber(code) ~= 200 then
                                ecb(string.format(
                                    "HTTP error response: %d (%s)", code, msg),
                                    code)
                                return 0
                            end
                            header = {}
                        else
                            local k, v = line:match('^(.-):%s-(.+)$')
                            if k and v then
                                header[k:lower()] = v
                            end
                        end
                    end
                    if header['cache-control'] and header['pragma'] and
                        header['expires'] and header['connection'] and
                        header['content-type'] then

                        boundary_marker =
                            header['content-type']:match('^.+=(.+)')
                        if boundary_marker then
                            boundary_marker = '--' .. boundary_marker
                            image_info = {}
                            state = 'wait4boundary'
                            avail = e
                        else
                            ecb("Unable to find boundary marker in " ..
                                "HTTP response header.")
                            return 0
                        end
                    else
                        ecb("One or more items expected in HTTP " ..
                            "response header was not received")
                        return 0
                    end
                end
            elseif state == 'wait4boundary' then
                local s, e = ffi.string(buf, avail):find('\r\n\r\n')
                avail = 0
                if s and e then
                    local str = ffi.string(buf, e)
                    for line in str:gmatch('[^\r\n]+') do
                        if line == boundary_marker then
                            image_info = {}
                        else
                            local k, v = line:match('^(.+):%s-(.+)$')
                            if k and v then
                                image_info[k:lower()] = v
                            end
                        end
                    end
                    if image_info['content-type'] and
                        image_info['content-length'] then
                        state = 'wait4frame'
                        frame_size = tonumber(image_info['content-length'])
                        avail = e
                    else
                        ecb("One or more items expected in boundary " ..
                            "header was not received")
                        return 0
                    end
                end
            elseif state == 'wait4frame' then
                if avail >= frame_size then
                    callback(buf, frame_size)
                    avail = frame_size
                    state = 'wait4boundary'
                    frame_size = 0
                else
                    avail = 0
                end
            end
        end
        return avail or 0
    end
    sbuf:start(ev.Loop.default)
    local req
    if http_info.method and http_info.method == 'POST' then
        if http_info.extra_header then
            req = string.format('%s %s HTTP/1.1\r\n' ..
                'Host: %s\r\n' ..
                'Content-Length: %d\r\n' ..
                'Content-Type: application/json\r\n' ..
                '%s\r\n' ..
                '%s\r\n',
                http_info.method, http_info.absolute_path,
                http_info.host, #data + 2, http_info.extra_header, data)
        else
            req = string.format('%s %s HTTP/1.1\r\n' ..
                'Host: %s\r\n' ..
                'Content-Length: %d\r\n' ..
                'Content-Type: application/json\r\n\r\n' ..
                '%s\r\n',
                http_info.method, http_info.absolute_path,
                http_info.host, #data + 2, data)
        end
    elseif http_info.method and http_info.method ~= 'GET' then
        req = string.format('%s %s HTTP/1.1\r\nHost: %s\r\n\r\n%s\r\n',
            http_info.method, http_info.absolute_path, http_info.host, data)
    else
        if http_info.extra_header then
            req = string.format('GET %s HTTP/1.1\r\nHost: %s\r\n%s\r\n',
                http_info.absolute_path, http_info.host, http_info.extra_header)
        else
            req = string.format('GET %s HTTP/1.1\r\nHost: %s\r\n\r\n',
                http_info.absolute_path, http_info.host)
        end
    end
    sbuf:write(req)
end

local resolution_table = {
    "1280x1024", "1280x960", "1280x720", "768x576", "4cif", "704x576",
    "704x480", "vga", "640x480", "640x360", "480x360", "cif", "384x288",
    "352x288", "352x240", "320x240", "240x180", "qcif", "192x144", "176x144",
    "176x120", "160x120"
}

function m:process_video(settings, host, cb, ecb)
    settings.resolution = settings.resolution or '640x480'
    settings.fps = settings.fps or 30
    settings.compression = settings.compression or '20'
    self:_http_request_stream(
        {
            host = host,
            method = 'GET',
            absolute_path = string.format('/axis-cgi/mjpg/video.cgi?' ..
                'resolution=%s&fps=%s&compression=%s&rotation=%s&camera=%s',
                settings.resolution,
                settings.fps,
                settings.compression,
                settings.rotation,
                settings.connector:match('^video_(.-)@') or 1),
        },
        function(data, len)
            if cb(data, len) == true then
                -- Error from data handler callback
                self.sbuf:close()
                self.sbuf.read_cb = nil
            end
        end,
        -- Error detected in read from connection to camera
        -- We don't really know why the request is rejected we are assuming it
        -- is because the request resolution is not supported.  So we try the
        -- next one from the list
        function(msg, code)
            if self.sbuf then
                self.sbuf:close()
                self.sbuf.read_cb = nil
            end
            if tonumber(code) == 400 then
                for i, res in ipairs(resolution_table) do
                    if res == settings.resolution:lower() then
                        if i - 1 > 0 then
                            settings.resolution = resolution_table[i - 1]
                            break
                        end
                        ecb(msg)
                        return
                    end
                end
                self:process_video(settings, host, cb, ecb)
            else
                ecb(msg)
            end
        end)
end

function m:_set_admin_password(host, cb)
    self:http_request(
        {
            host = host,
            method = 'GET',
            absolute_path = '/axis-cgi/pwdroot/pwdroot.cgi' ..
                '?action=update&user=root&pwd=insecure&return_page=/'
        },
        function(data)
            -- No response data - error
            if not data then
                cb()
                return
            end

            -- Success
            cb(true)
        end, nil, credentials)
end

function m:_enable_anonymous_viewing(host, cb)
    -- After the administrator's password is set, a short delay is required
    -- before anonymous viewing can be enabled.
    ev.Timer.new(function(l, t, e)
            self:http_request(
                {
                    host = host,
                    method = 'GET',
                    absolute_path = '/axis-cgi/admin/param.cgi?action=update' ..
                        '&Network.RTSP.ProtViewer=anonymous'
                },
                function(data)
                    -- No response data - error
                    if not data then
                        cb("Unable to enable anonymous viewer on RTSP stream")
                        return
                    end

                    -- Success
                self:http_request(
                    {
                        host = host,
                        method = 'GET',
                        absolute_path =
                            '/axis-cgi/admin/param.cgi?action=update' ..
                            '&System.BoaProtViewer=anonymous'
                    },
                    function(data)
                        -- No response data - error
                        if not data then
                            cb("Unable to enable anonymous viewer on BOA")
                            return
                        end

                        -- Success
                        cb()
                    end, nil, credentials)
                end, nil, credentials)
        end, 2, 0):start(ev.Loop.default)
end

function m:configure_camera(host, cb)
    -- First check to see if we are already configured
    self:info(host, function(data)
            if data and type(data) == 'table' then
                -- Already configured, return OK
                cb()
                return
            end

            -- Make sure the administrator password is set
            self:http_request(
                {
                    host = host,
                    method = 'GET',
                    absolute_path = '/pwdroot/pwdRoot.shtml'
                },
                function(data)
                    -- No response data - error
                    if not data then
                        cb("Unable to determine if admin password is set")
                        return
                    end
        
                    -- Set administrator account password if it has not been
                    -- set yet.  The text below only shows up on the page if
                    -- the password has NOT been set.
                    if data:find('.+checkPassword.+') then
                        self:_set_admin_password(host, function(data)
                                -- Unable to set - error
                                if not data then
                                    cb("Unable to set adminstrator password")
                                    return
                                end
        
                                -- Enable anonymous viewing
                                self:_enable_anonymous_viewing(host,
                                    function(data)
                                        cb(data)
                                    end)
                            end)
                    end

                    -- Enable anonymous viewing
                    self:_enable_anonymous_viewing(host, function(data)
                            cb(data)
                        end)
                end, nil, credentials)
        end)
end

function m:info(host, cb)
    self:http_request(
        {
            host = host,
            method = 'GET',
            absolute_path = '/axis-cgi/admin/param.cgi?action=list' ..
                '&group=Brand.ProdNbr,Properties.System.SerialNumber,' ..
                'Properties.Firmware.Version,' ..
                'ImageSource.NbrOfSources' ..
                '&responseformat=rfc'
        },
        function(data)
            -- Callback has been called, just return
            if not cb then return end

            -- No response data - error
            if not data then
                cb("HTTP request failed")
                cb = nil
                return
            end
            local ret = {}
            ret.camera_ports = 1
            for line in data:gmatch('[^\r\n]+') do
                local k, v = line:match('^(.+)=(.+)$')
                if k and v then
                    if k == 'Brand.ProdNbr' then
                        -- Remove forward slashes from model name
                        ret.model = v:lower():gsub('/', '')
                    elseif k == 'Properties.System.SerialNumber' then
                        ret.serial = v:lower()
                    elseif k == 'Properties.Firmware.Version' then
                        ret.fw_version = v:lower()
                    elseif k == 'ImageSource.NbrOfSources' then
                        ret.camera_ports = v:lower()
                    end
                end
            end
            cb(ret)
            cb = nil
        end, nil, credentials)
end

return m
