-- MX Channel proxy for MX1601, MX1615, MX1609, MX411

local ev = require 'ev'
local log = require 'log'
local channel_proxy = require 'channel_proxy'
local cjson = require 'cjson'
local rational = require 'rational'
local config = require 'system_config'
local sys_notify = require 'sys_notify'

-- Map of DRD "magic" numbers to strings
local drd_map = {
    ['0'] = 'hbm_legacy',
    ['2'] = 'decimal',
    ['3'] = 'binary',
}

local mxChannelProxy = class('mxChannelProxy', channel_proxy)

function mxChannelProxy:initialize(mxDevice, channel_num)
    self.mxdev = mxDevice
    self.timestamp = 0
    self.index = channel_num
    self.proxy_type = 'mxChannelProxy'
    self.comet_sinks = {}

    self:apply_non_teds_state()

end


function mxChannelProxy:apply_teds_state()

    self.spec = self.mxdev.mx.control:get_teds_spec(self.index)
    self.spec.data_type = 'sequential_float32'
    if not self.spec.teds_sensor.bin_data then
        self.spec = self.mxdev.mx.control:get_spec(channel_num)
        self.spec.name = string.format('%s_ch_%d',
            self.mxdev.name:gsub('-', '_'), channel_num)
        self.spec.teds_sensor = {
            status = 'connected',
            use_teds = tonumber(self.spec.teds_sensor.use_teds),
        }
        error(string.format(
            "A TEDS device was installed into channel %d and no TEDS data " ..
            "has been supplied by the MX device.  The channel will be " ..
            "treated as though no TEDS device is installed.", self.index))
    else
        if self.spec.teds_sensor.name then
            self.spec.name = string.format('%s (%s)',
                self.spec.teds_sensor.name, self.spec.teds_sensor.serial_number)
        else
            self.spec.name = string.format('%s_%s_%s',
            self.spec.teds_sensor.manufacturer or 'unknown',
            self.spec.teds_sensor.type or 'unknown',
            self.spec.teds_sensor.serial_number or 'unknown'):lower()
        end
        self.spec.teds_sensor.status = 'connected'
        if self.spec.teds_sensor.use_teds then
            self.spec.teds_sensor.use_teds =
                tonumber(self.spec.teds_sensor.use_teds)
        end
    end

end


function mxChannelProxy:apply_non_teds_state()

    self.spec = self.mxdev.mx.control:get_spec(self.index)
    self.spec.data_type = 'sequential_float32'
    self.spec.name = string.format('%s_ch_%d',
        self.mxdev.name:gsub('-', '_'), self.index)
    if self.spec.teds_sensor then
        self.spec.teds_sensor.status = 'not connected'
        if self.spec.teds_sensor.use_teds then
            self.spec.teds_sensor.use_teds =
                tonumber(self.spec.teds_sensor.use_teds)
        end
    else
        self.spec.teds_sensor = {
            status = 'not connected',
            use_teds = 0,
        }
    end

end


function mxChannelProxy:available_channels(hostname)
    self.stream_reference = self.spec._reference
    self.spec.connector = string.format('@%s', self.proxy_hostname)
    local cnum, base = self.proxy_hostname:match('connector(%d+)%.(mx.+)$')
    self.spec.category = string.format('source/hbm_%s', self.mxdev.category_name)
    self.spec.description = string.format("%s Connector %d", base:upper(), cnum)
    self.channel_name = self.spec.connector
    self.spec.valid_controls = {
        initialize_channel = true,
    }
    self.spec.is_logical = false
    local controls = self.mxdev.mx.control.device.controls and
        self.mxdev.mx.control.device.controls.channel or {}
    for c, f in pairs(controls) do
        self.spec.valid_controls[c] = true
        self[c] = f
    end

    if self.spec.teds_sensor and self.spec.teds_sensor.use_teds > 0 and
        self.spec.teds_sensor.status == 'connected' then
        self.spec.description =
            string.format("TEDS Device: %s %s %s",
                self.spec.teds_sensor.manufacturer or "Manufacturer unknown",
                self.spec.teds_sensor.type or "Type unknown",
                self.spec.teds_sensor.serial_number or "Serial number unknown")

        if config.use_long_teds_connector_names then
            self.spec.connector =
                string.format('%s-%s-%s@%s', self.spec.teds_sensor.manufacturer,
                    self.spec.teds_sensor.type,
                    self.spec.teds_sensor.serial_number, self.proxy_hostname)
        else
            self.spec.connector = string.format('@%s', self.proxy_hostname)
        end
        self.channel_name = self.spec.connector
        self.spec.electrical_units = nil
    elseif self.spec.teds_sensor and self.spec.teds_sensor.use_teds > 0 and
        self.spec.teds_sensor.status == 'not connected' then
        local ns = {
            use_teds = self.spec.teds_sensor.use_teds,
            status = 'not connected',
        }
        self.spec.teds_sensor = ns
    end

    -- Stupid defaults that really should be done by the GUI!
    local device = self.mxdev.mx.control.devinfo.name:lower()
    local set_defaults = {
        mx1615b = true,
        mx1601b = true,
        mx840b = true,
        mx411b = true
    }
    if set_defaults[device] then
        local drd = drd_map[self.mxdev.mx.control.devinfo.current_drd or 5]
        if drd == 'hbm_legacy' then
            self.spec.sample_rate = "150"
            self.spec.digital_filter_attenuation_frequency = "20"
            self.spec.digital_filter_type = "bessel_iir"
        elseif drd == 'decimal' then
            self.spec.sample_rate = "100"
            self.spec.digital_filter_attenuation_frequency = "33"
            self.spec.digital_filter_type = "linear_phase_fir"
        end
    end

    local dev = self.mxdev.mx.control.device
    if dev.insert_additional_channel_spec_data then
        dev:insert_additional_channel_spec_data(self.spec)
    end

    -- Don't include the _reference stuff
    self.spec._reference = nil
    self.spec.is_little_endian = false
    return { self.spec }
end

function mxChannelProxy:start_comet_stream(w, chans, sink)
    if not config.mx_comet_stream_enable then return end
    for _, chan in ipairs(chans) do
        if chan == self.spec.connector then
            self.comet_sinks[w.socket] = sink
            w.cleanup = function()
                self.comet_sinks[w.socket] = nil
            end
        end
    end
end

function mxChannelProxy:initialize_channel(params, cb)
    local spec = params[1]
    local idx = spec.connector:match("connector(%d)")
    self.mxdev.mx.control:setup_channel(idx, spec, function(result)
            cb(result)
        end)
end

function mxChannelProxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            result = 'error',
            msg = 'Control request with not request specified',
        }
    end
    if not self[spec.request] or
        type(self[spec.request]) ~= 'function' then
        return {
            result = 'error',
            msg = string.format("Control function not supported: %s",
                spec.request)
        }
    end
    return self[spec.request](self, spec.params, cb)
end

function mxChannelProxy:_abort_test_start(l, t, e)
    self.mxdev.cleanup_was_called = true
    self.start_channel_list = {}
end

-- Function used to start a channel stream for a test
function mxChannelProxy:_start_channel_stream(w, spec, md_sink, sink)
    self.mxdev.cleanup_was_called = nil
    -- First attach a cleanup function to be used until replaced at a later
    -- point in the test start sequence.
    w.cleanup = function()
            self:_abort_test_start()
        end

--[[
    -- For DVM operation, set start_async to true and
    -- disregard_sync_status_changes to true
    --
    -- For asynchronous test purposes, start_async should be true and
    -- disregard_sync_status_changes can be set to either state.
    -- 
]]
    -- A purpose of 'digital_display' means asynchronous start
    self.mxdev.start_async = nil
    self.mxdev.disregard_sync_status_changes = nil
    if spec.purpose then
        if spec.start_time_delay == 0 then
            self.mxdev.start_async = true
        end
        self.mxdev.disregard_sync_status_changes = true
    end

    -- Set the timestamp for this channel
    self.timestamp = 0

    -- Do some initialization of the list of channels that are in this test
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}
    self.mxdev.start_channel_list[spec._test_id] =
        self.mxdev.start_channel_list[spec._test_id] or {}

    -- Make sure all of the "required" parameters have been given
    local spec_required = {
        'source_physical_units',
        'connector',
        'sensor_input'
    }
    for _, k in pairs(spec_required) do
        if not spec[k] then
            local msg = string.format(
                "Required parameter '%s' is not included in setup for %s",
                k, spec.name or 'UNKNOWN NAME')
            log:log(log.ERR, msg)

            -- Send back the message on ALL previously successfully added
            -- channels
            local start_channel_list = self.mxdev.start_channel_list
            for tid, t in pairs(start_channel_list) do
                for _, c in ipairs(t) do
                    c.md_sink{error_message = msg}
                end
            end
            self.mxdev.start_channel_list = {}

            -- Send back for this channel too!
            md_sink{error_message = msg}
            return
        end
    end

    -- Make sure all of the channels starting up are part of the same test id
    local num = 0
    for k, v in pairs(self.mxdev.start_channel_list) do
        num = num + 1
        if num > 1 then
            -- Yikes, more than one test is trying to start up at the same time!
            local msg = 'Can not start more than one test/MX device at a time'
            log:log(log.ERR, msg)

            -- Send back the message on ALL previously successfully added
            -- channels
            local start_channel_list = self.mxdev.start_channel_list
            for tid, t in pairs(start_channel_list) do
                for _, c in ipairs(t) do
                    c.md_sink{error_message = msg}
                end
            end

            -- Send back for this channel too!
            md_sink{error_message = msg}
            return
        end
    end

    local start_channel_list = self.mxdev.start_channel_list[spec._test_id]

    -- Deal wit rationally expressed sample rates
    if spec.sample_rate then
        spec.sample_rate = rational:new(spec.sample_rate):tonumber()
    end

    -- Set up the information that will be used to configure and start this
    -- channel after this function has been called for ALL of this devices
    -- channels in this test!
    start_channel_list[#start_channel_list + 1] = {
        w = w,
        spec = spec,
        md_sink = md_sink,
        sink = sink,
        chan_obj = self,
    }

    -- Make sure changes are saved
    self.mxdev.start_channel_list[spec._test_id] = start_channel_list

    -- If this function has been called for all of the channels on this
    -- device for this test, then we can start the test!
    if #start_channel_list == spec._channel_count then
        self.mxdev:_start_channels(start_channel_list)
    end
end

function mxChannelProxy:start_channel_stream(w, spec, md_sink, sink, cb)
    if self.shunt then
        self:shunt({0},
            function(res)
                if res and res.result and res.result == 'success' then
                    local ok, err = pcall(self._start_channel_stream, self,
                        w, spec, md_sink, sink)
                    if ok then
                        if cb then cb() end
                        return
                    end
                end
                self.mxdev.start_channel_list = nil
                local msg = "Unable to remove shunt"
                md_sink{error_message = msg}
                log:log(log.ERR, msg)
            end)
        return
    end
    local ok, err = pcall(self._start_channel_stream, self, w, spec, md_sink,
        sink)
    if not ok then
        self.mxdev.start_channel_list = nil
        local msg = string.format("%s: %s", spec.connector, err)
        md_sink{error_message = msg}
        log:log(log.ERR, msg)
        return
    else
        if cb then cb() end
    end
end

return mxChannelProxy
