require 'middleclass'
local sqlite3 = require 'lsqlite3'
local lpeg = require 'lpeg'
local cjson = require 'cjson.safe'
local config = require 'system_config'
local ev = require 'ev'

local NetworkDB = class('NetworkDB')

local sdm, expected_setup_verison
if config.device_type == 'cx23r' then
    sdm = require 'setupdb_migrate'
    expected_setup_version = sdm.get_info().expected_setup_file_version
end

local fpga_version

local function get_fpga_version()
-- EDAQ2 FIXME
if config.device_type and config.device_type == 'edaq2' then
    return 'FIXME', true
end
    local fh = io.open('/var/log/fpga-hardware', 'r')
    if not fh then return 'Unknown' end
    local last_line
    for line in fh:lines() do
        if line:find('FPGA') then last_line = line end
    end
    fh:close()
    if not last_line then return 'Unknown' end
    local version = last_line:match('%[FPGA%-(.+)%]')
    fh = io.popen('spi-flash-tool verify /usr/share/fpga-config/digcomm.bin')
    if fh then
        for line in fh:lines() do
            if line:find('Verified') then
                fh:close()
                return version or 'Unknown', true
            end
            break
        end
        fh:close()
    end
    return version or 'Unknown'
end

local function get_platform_info()
    -- append the setup version to the response
    local platform_info = {}
    local f = io.open('/etc/version.json')
    if f then
        platform_info = cjson.decode(f:read('*a')) or {}
        f:close()
    end
    f = io.open('/tmp/power-micro-version.json', 'r')
    platform_info.power_micro = {fw_version = 'Unknown'}
    if f then
        local t = cjson.decode(f:read('*a'))
        if t then
            platform_info.power_micro.fw_version =
                string.format("%d.%d", t.running.major, t.running.minor)
        end
        f:close()
-- EDAQ2 FIXME
elseif config.device_type and config.device_type == 'edaq2' then
platform_info.power_micro.fw_version = 'FIXME'
    end
    if not fpga_version then
        fpga_version = get_fpga_version()
    end
    platform_info.fpga = {version = fpga_version}
    if config.device_type == 'cx23r' then
        platform_info.setup = {version = expected_setup_version or 'Unknown'}
    end
    platform_info.device = config.device_type:upper()
    local function get_sn()
        local fh = io.popen('ip link show dev eth0')
        if fh then
            for line in fh:lines() do
                local sn = line:match('link/ether (.-) brd')
                if sn then
                    fh:close()
                    return sn:gsub(':', ''),
                        sn:match('^.-:.-:.-:(.+)$'):gsub(':', '')
                end
            end
            fh:close()
            return 'Unknown'
        end
    end
    local function get_ipv4_addr()
        local fh = io.popen('ip addr show dev ' ..
            (config.device_type == 'cx23r' and 'eth0' or 'ethint'))
        if fh then
            for line in fh:lines() do
                local ipv4 = line:match('inet (.-)/%d+ brd ')
                if ipv4 then
                    fh:close()
                    return ipv4
                end
            end
            fh:close()
        end
    end
    local id
    platform_info.serial_number, id = get_sn()
    platform_info.ip_address = get_ipv4_addr()
    platform_info.device_id = (config.device_type or 'cx23r') .. '-' .. id
    return platform_info
end

function NetworkDB:initialize(network, db, hostname, ip)
    self.network = network
    self.stmt_cache = {}
    self.db = assert(sqlite3.open(db))

    self.hostname = hostname
    self.ip = ip
    self.soas = {}

    assert(self.db:exec[==[
        DELETE FROM unified WHERE owner = '_armada-device-info';
        DELETE FROM unified WHERE owner LIKE '_system-notifcation%';
        CREATE TEMP TABLE unified_deltas (
            op TEXT NOT NULL,
            owner TEXT NOT NULL,
            type TEXT NOT NULL,
            rdata TEXT NOT NULL,
            source TEXT NOT NULL,
            local_source TEXT NOT NULL
        );
        CREATE TEMP TRIGGER unified_update AFTER UPDATE ON unified BEGIN
            INSERT INTO unified_deltas (
                op, owner, type, rdata, source, local_source
            ) VALUES (
                'D', old.owner, old.type, old.rdata, old.source,
                old.local_source
            );
            INSERT INTO unified_deltas (
                op, owner, type, rdata, source, local_source
            ) VALUES (
                'A', new.owner, new.type, new.rdata, new.source,
                new.local_source
            );
        END;
        CREATE TEMP TRIGGER unified_insert AFTER INSERT ON unified BEGIN
            INSERT INTO unified_deltas (
                op, owner, type, rdata, source, local_source
            ) VALUES (
                'A', new.owner, new.type, new.rdata, new.source,
                new.local_source
            );
        END;
        CREATE TEMP TRIGGER unified_delete AFTER DELETE ON unified BEGIN
            INSERT INTO unified_deltas (
                op, owner, type, rdata, source, local_source
            ) VALUES (
                'D', old.owner, old.type, old.rdata, old.source,
                old.local_source
            );
        END;
        CREATE TEMP TABLE staging (
            owner TEXT NOT NULL,
            type TEXT NOT NULL,
            rdata TEXT NOT NULL
        );
        CREATE INDEX staging_owner ON staging (owner);
        CREATE INDEX staging_type ON staging (type);
        CREATE INDEX staging_rdata ON staging (rdata);
    ]==])
    self.select_master_rr =
        assert(self.db:prepare[[
            SELECT owner, type, rdata, source FROM unified
                WHERE owner GLOB ? AND type GLOB ? AND rdata GLOB ?
                    AND source GLOB ?;
        ]])
    self.select_local_rr =
        assert(self.db:prepare[[
            SELECT owner, type, rdata, source FROM unified
                WHERE owner GLOB ? AND type GLOB ? AND rdata GLOB ?
                    AND local_source <> '' AND local_source GLOB ?
                        ORDER BY ROWID ASC;
        ]])
    self.insert_local_rr =
        assert(self.db:prepare[[
            INSERT INTO unified (owner, type, rdata, source, local_source)
                VALUES (?, ?, ?, ?, ?);
        ]])
    self.delete_local_rr =
        assert(self.db:prepare[[
            DELETE FROM unified
                WHERE owner = ? AND type = ? AND rdata = ?
                    AND local_source = ?;
        ]])
    self.delete_local_rr_owner =
        assert(self.db:prepare[[
            DELETE FROM unified WHERE owner = ? AND local_source <> '';
        ]])
    self.delete_local_rr_otr =
        assert(self.db:prepare[[
            DELETE FROM unified
                WHERE owner = ? AND type = ? AND rdata = ?
                    AND local_source <> '';
        ]])
    self.local_hadv_response =
        assert(self.db:prepare[[
            SELECT owner || X'09' || type || X'09' || rdata || X'0a' AS rr
                FROM unified
                WHERE local_source <> '';
        ]])
    self.insert_master_rr =
        assert(self.db:prepare[[
            INSERT INTO unified (owner, type, rdata, source, local_source)
                VALUES (?, ?, ?, ?, '');
        ]])
    self.delete_master_rr =
        assert(self.db:prepare[[
            DELETE FROM unified
                WHERE owner = ? AND type = ? AND rdata = ? AND source = ?
                    AND local_source = '';
        ]])
    self.insert_staging_rr =
        assert(self.db:prepare[[
            INSERT INTO staging (owner, type, rdata)
                VALUES (?, ?, ?);
        ]])

    -- Read out old local SOA and delete (will be recreated after db cleanup)
    local soarr = self:sql_query([[
        SELECT rdata FROM unified WHERE type == 'SOA' AND local_source = 'SOA';
    ]])
    self.soas[self.ip] = (#soarr > 0 and soarr[1][1] or 0) + 1
    self.db:exec([[DELETE FROM unified WHERE local_source = 'SOA';]])

    -- First, delete all non-local stuff
    self.db:exec([[DELETE FROM unified WHERE local_source = '';]])

    -- Preserve interactive trigger states CNAME and DNAME record types.
    if config.device_type == 'cx23r' then
        self.db:exec(
            [[DELETE FROM unified WHERE owner
                NOT LIKE '_interactive-trigger-%' AND
                type != 'DNAME' AND type != 'CNAME';]])
    else
        self.db:exec(
            [[DELETE FROM unified WHERE type != 'DNAME' AND type != 'CNAME';]])
    end

    -- Free up space if possible
    self.db:exec([[VACUUM;]])

    -- Add our hostname A record and recreate local SOA
    self:add_rr_int(self.hostname, 'A', self.ip, 'SOA')
    self:add_rr_int('', 'SOA', self.soas[self.ip], 'SOA')

    -- Grab the platform info
    self.platform_info = get_platform_info()
end

function NetworkDB:do_get_platform_info()
    self.platform_info = get_platform_info()
end

function NetworkDB:get_soa(ip)
    return self.soas[ip]
end

function NetworkDB:prepare_changes() --? private
    self:sql_query([[BEGIN TRANSACTION;]])
end

function NetworkDB:finish_changes() --? private
    local local_deltas = (self:sql_query([[
        SELECT count(*) FROM unified_deltas WHERE local_source <> '';
    ]]))[1][1] > 0
    if local_deltas then
        local soa = self.soas[self.ip] + 1
        self.soas[self.ip] = soa
        self:sql_query([[
            UPDATE unified SET rdata = ?
                WHERE type = 'SOA' AND local_source = 'SOA';
        ]], soa)
        self.cached_hadv_response = nil
    end
    local deltas = self:sql_query([[
        SELECT op, owner, type, rdata, source, local_source
            FROM unified_deltas;
    ]])
    local hdif_del, hdif_add = {}, {}
    if #deltas > 0 then
        local del, add = {}, {}
        for i = 1, #deltas do
            local r = deltas[i]
            local v = {r[2], r[3], r[4], r[5]}
            if r[1] == 'D' then
                del[#del + 1] = v
                if r[6] ~= '' then
                    hdif_del[#hdif_del + 1] = {r[2], r[3], r[4]}
                end
            elseif r[1] == 'A' then
                add[#add + 1] = v
                if r[6] ~= '' then
                    hdif_add[#hdif_add + 1] = {r[2], r[3], r[4]}
                end
            end
        end
        if #add > 0 or #del > 0 then
            self.network:changed_rrs(del, add)
        end
    end
    self:sql_query([[DELETE FROM unified_deltas;]])
    self:sql_query([[COMMIT;]])
    if local_deltas then
        self.network:local_deltas({del = hdif_del, add = hdif_add})
    end
end

function NetworkDB:rollback_changes() --? private
    self:sql_query([[ROLLBACK;]])
    self:sql_query([[DELETE FROM unified_deltas;]]) -- KLUDGE???
end

--! Generic SQL query.
function NetworkDB:sql_query(sql, ...)
    local stmt = self.stmt_cache[sql]
    if not stmt then
        stmt = assert(self.db:prepare(sql))
        self.stmt_cache[sql] = stmt
    end
    local rrs = {}
    stmt:reset()
    stmt:bind_values(...)
    for rr in stmt:rows() do
        rrs[#rrs + 1] = rr
    end
    return rrs
end

local function split(s, sep)
    sep = lpeg.P(sep)
    local elem = lpeg.C((1 - sep)^0)
    local p = lpeg.Ct(elem * (sep * elem)^0)
    return lpeg.match(p, s)
end

--! Resolves a network name to its canonical form.
function NetworkDB:resolve_name(name)
    local rrs = self:sql_query([[
        SELECT owner, rdata FROM unified
            WHERE type = 'CNAME' AND owner GLOB ?;
    ]], name)
    if #rrs > 0 then
        local rr = rrs[1]
        return self:resolve_hostname(rr[2])
    end
    local domain_parts = split(name, '.')
    local host_parts = { }
    table.insert(host_parts, table.remove(domain_parts, 1))
    while domain_parts[1] do
        local domain = table.concat(domain_parts, '.')
        local rrs = self:sql_query([[
            SELECT owner, rdata FROM unified
                WHERE type = 'DNAME' AND owner GLOB ?;
        ]], domain)
        if #rrs > 0 then
            local rr = rrs[1]
            local host = ''
            if #host_parts > 0 then
                host = table.concat(host_parts, '.') .. '.'
            end
            return self:resolve_hostname(host .. rr[2])
        end
        host_parts[#host_parts + 1] = table.remove(domain_parts, 1)
    end
    return name
end

--! Resolves a hostname to an address.
function NetworkDB:resolve_hostname(name)
    name = self:resolve_name(name)
    local rrs = self:sql_query([[
        SELECT owner, rdata FROM unified
            WHERE type = 'A' AND owner GLOB ?;
    ]], name)
    if #rrs > 0 then
        return rrs[1][2], rrs[1][1]
    end
    return nil
end

--! Resolves a service name to an address and port.
function NetworkDB:resolve_service(name)
    name = self:resolve_name(name)
    local rrs = self:sql_query([[
        SELECT rdata FROM unified
            WHERE type = 'SRV' AND owner GLOB ?;
    ]], name)
    for i = 1, #rrs do
        local rr = rrs[i]
        local port, host = rr[1]:match('^%d+ %d+ (%d+) (.*)')
        local ip, canonical, err = self:resolve_hostname(host)
        if not ip then return ip, canonical, err end
        return host, port, ip, canonical
    end
    return nil
end

--! Reverse-resolves an address to a hostname.
function NetworkDB:resolve_address(addr)
    addr = self.network.normalize_address(addr)
    local rrs = self:sql_query([[
        SELECT owner FROM unified
            WHERE type = 'A' AND rdata = ?;
    ]], addr)
    if #rrs > 0 then
        return rrs[1][1]
    end
    return nil
end

--! Queries the name database.
function NetworkDB:query_rr(owner, type, rdata, source)
    local rrs = { }
    self.select_master_rr:reset()
    self.select_master_rr:bind_values(owner, type, rdata, source)
    for rr in self.select_master_rr:rows() do
        rrs[#rrs + 1] = rr
    end
    return rrs
end

--! Queries the local name database.
function NetworkDB:query_local_rr(owner, type, rdata, source)
    local rrs = { }
    self.select_local_rr:reset()
    self.select_local_rr:bind_values(owner, type, rdata, source)
    for rr in self.select_local_rr:rows() do
        rrs[#rrs + 1] = rr
    end
    return rrs
end

--! Modifies the name database.
function NetworkDB:modify_rrs(remove_rrs, add_rrs, local_source)
    self:prepare_changes()
    for i = 1, #remove_rrs do
        local rr = remove_rrs[i]
        self.delete_local_rr:reset()
        self.delete_local_rr:bind_values(rr[1], rr[2], rr[3], local_source)
        repeat until self.delete_local_rr:step() ~= sqlite3.ROW
    end
    for i = 1, #add_rrs do
        local rr = add_rrs[i]
        self:add_rr_int(rr[1], rr[2], rr[3], local_source)
    end
    self:finish_changes()
end

function NetworkDB:remove_local_owner(owner)
    self:prepare_changes()
    self.delete_local_rr_owner:reset()
    self.delete_local_rr_owner:bind_values(owner)
    repeat until self.delete_local_rr_owner:step() ~= sqlite3.ROW
    self:finish_changes()
end

function NetworkDB:remove_local_otr(owner, type, rdata)
    self:prepare_changes()
    self.delete_local_rr_otr:reset()
    self.delete_local_rr_otr:bind_values(owner, type, rdata)
    repeat until self.delete_local_rr_otr:step() ~= sqlite3.ROW
    self:finish_changes()
end

function NetworkDB:add_rr_int(owner, type, rdata, local_source)
    self.insert_local_rr:reset()
    self.insert_local_rr:bind_values(owner, type, rdata, self.ip, local_source)
    repeat until self.insert_local_rr:step() ~= sqlite3.ROW
end

function NetworkDB:get_hadv_response()
    if not self.cached_hadv_response then
        local rrs = {}
        local stmt = self.local_hadv_response
        stmt:reset()
        for rr in stmt:rows() do
            rrs[#rrs + 1] = rr[1]
        end
        self.cached_hadv_response = table.concat(rrs)
    end
    return self.cached_hadv_response
end

function NetworkDB:update_host_rrs(host_ip, rrs)
    self:prepare_changes()
    local stmt = self.insert_staging_rr
    local new_soa
    for i = 1, #rrs do
        local rr = rrs[i]
        stmt:reset()
        stmt:bind_values(rr[1], rr[2], rr[3])
        repeat until stmt:step() ~= sqlite3.ROW
        if rr[2] == 'SOA' and rr[1] == '' then
            new_soa = tonumber(rr[3])
        end
    end
    if new_soa then
        -- Propagate only changes into unified table
        self:sql_query([[
            DELETE FROM unified
                WHERE oid IN (
                    SELECT u.oid
                    FROM unified AS u LEFT JOIN staging AS s
                    ON u.owner = s.owner
                        AND u.type = s.type
                        AND u.rdata = s.rdata
                    WHERE u.source = ? AND u.local_source = ''
                        AND s.owner IS NULL
                );
        ]], host_ip)
        self:sql_query([[
            INSERT INTO unified (owner, type, rdata, source, local_source)
                SELECT s.owner, s.type, s.rdata, ?, ''
                    FROM staging AS s LEFT JOIN unified AS u
                    ON u.owner = s.owner
                        AND u.type = s.type
                        AND u.rdata = s.rdata
                        AND u.source = ?
                        AND u.local_source = ''
                    WHERE u.owner IS NULL;
        ]], host_ip, host_ip)
        self:sql_query([[DELETE FROM staging;]])
        self.soas[host_ip] = new_soa
        self:finish_changes()
    else
        self:rollback_changes()
    end
    return new_soa
end

function NetworkDB:update_host_delta(host_ip, soa, delta)
    self:prepare_changes()
    local stmt = self.delete_master_rr
    for i = 1, #delta.del do
        local rr = delta.del[i]
        stmt:reset()
        stmt:bind_values(rr[1], rr[2], rr[3], host_ip)
        repeat until stmt:step() ~= sqlite3.ROW
    end
    stmt = self.insert_master_rr
    for i = 1, #delta.add do
        local rr = delta.add[i]
        stmt:reset()
        stmt:bind_values(rr[1], rr[2], rr[3], host_ip)
        repeat until stmt:step() ~= sqlite3.ROW
    end
    self.soas[host_ip] = soa
    self:finish_changes()
end

function NetworkDB:delete_host(host_ip)
    self:prepare_changes()
    self.soas[host_ip] = nil
    self:sql_query([[DELETE FROM unified WHERE source = ?;]], host_ip)
    self:finish_changes()
end

function NetworkDB:delete_local_source(local_source) --? private
    self:prepare_changes()
    -- Preserve interactive trigger states
    self:sql_query(
        [[DELETE FROM unified WHERE local_source = ?
            AND owner NOT LIKE '_interactive-trigger-%';]], local_source)
    self:finish_changes()
end

return NetworkDB
