require'middleclass'
local net = require'net'
local ev = require'ev'
local ffi = require'ptpv2-ffi/cdata'
local C = ffi.C
require'ptpv2-ffi/internal-time'

local ptpTimingInterface = class('ptpTimingInterface')

ffi.cdef[[
void *calloc(size_t nmemb, size_t size);

typedef struct timing_info {
    internal_time_t     xt;
    long long           xt_sec;
    long long           xt_nsec;
    long long           ptp_ctr;
    long long           ptp_rate;
    double              ptp_to_ps;
    long long           ptp_to_ns;
    long long           trillion;
    long long           x1000;
} timing_info_t;

typedef struct servo_info {
    double              dt;
    double              Kp;
    double              Ki;
    double              Kd;
    unsigned long long  count;
    double              integral;
    double              last_offset;
    int                 filter_countdown;
    double              output;
    double              error;
} servo_info_t;

typedef int64_t off_t;
typedef int     mode_t;

int open(const char *pathname, int flags, mode_t mode);
void *mmap(void *addr, size_t length, int prot, int flags, int fd, off_t
    offset);
char *strerror(int errnum);

enum {
    LOCAL_PROT_READ = 1,
    LOCAL_PROT_WRITE = 2,
};

enum {
    LOCAL_MAP_SHARED = 1,
};
]]

function ptpTimingInterface:initialize(hw_stamped)
    local dbus = require'dbus'
    local bus = dbus.bus_get('system')
    bus:hook_libev_default_loop()

    self.timing = assert(bus:auto_proxy_wait('com.somat.armada.Timing',
        '/com/somat/armada/Timing'))

--[[
    self.timing_info = ffi.cast('timing_info_t *',
        C.calloc(1, ffi.sizeof('timing_info_t')))

    self.ret_time = ffi.cast('internal_time_t *',
        C.calloc(1, ffi.sizeof('internal_time_t')))

    self.servo = ffi.cast('servo_info_t *', C.calloc(1,
        ffi.sizeof('servo_info_t')))
]]
    self.timing_info = {xt = {}}
    self.servo = {}

    -- Stuff used to read raw timestamp counters
    self.sys_rate = 40 * 1000000
    self.fd = C.open('/dev/mem', C.O_RDWR, 0)
    self.map = C.mmap(nil, 0x100000, C.LOCAL_PROT_READ, C.LOCAL_MAP_SHARED,
        self.fd, 0xe0000000)

    self.servo.dt = 1
    self.servo.Kp = 3.7
    self.servo.Ki = 1.0
    self.servo.Kd = 0.4
    self.hw_stamped = hw_stamped
    if not self.hw_stamped then
        self.servo.Kp = 3.7 / 1000
        self.servo.Ki = 1.0 / 10000
        self.servo.Kd = 0
    end
    self.servo.count = 0
    self.servo.last_offset = 0
    self.servo.filter_countdown = 20
    self.servo.error = 0
    self.servo.output = 0

    local _
    self.timing_info.xt.seconds, self.timing_info.xt.nanoseconds, _, _,
        self.timing_info.ptp_ctr, self.timing_info.ptp_rate =
        self.timing:GetClockRelationships()

    self.timing_info.xt_sec = tonumber(self.timing_info.xt.seconds)
    self.timing_info.xt_nsec = tonumber(self.timing_info.xt.nanoseconds)

    self.timing_info.ptp_to_ns = 1000000000 / self.timing_info.ptp_rate
    self.timing_info.trillion = 1000000000000
    self.timing_info.x1000 = 1000LL
    self.timing_info.ptp_to_ps =
        self.timing_info.trillion / self.timing_info.ptp_rate

    self.timing.TimeJumped:bind(
        function(xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate)
            self.timing_info.xt.seconds = xt_sec
            self.timing_info.xt.nanoseconds = xt_nsec
            self.timing_info.xt_sec = tonumber(self.timing_info.xt.seconds)
            self.timing_info.xt_nsec = tonumber(self.timing_info.xt.nanoseconds)
            self.timing_info.ptp_ctr = ptp_ctr
            self.timing_info.ptp_rate = ptp_rate
            self.timing_info.ptp_to_ns = 1000000000 / self.timing_info.ptp_rate
            self.timing_info.xt_sec = xt_sec
            self.timing_info.xt_nsec = xt_nsec
            self.timing_info.ptp_to_ps =
                self.timing_info.trillion / self.timing_info.ptp_rate
        end)
end


local dummy_time = InternalTime:new(0, 0)

-- FIXME:  This is only used with one step clocks.  So, it should probably be
-- made to work correctly at some point for those!
function ptpTimingInterface:get_time()
    return dummy_time
--[[
    local function read_32(offset)
        return ffi.cast('unsigned int *',
            ffi.cast('char *', self.map) + offset)[0]
    end

    -- Read the raw counters
    local rs = read_32(0x0061c)
    rs = rs * 4294967296
    rs = rs + read_32(0x0062c)

    -- Convert to and return time
    return InternalTime:new(self:raw_ts_to_time_int(rs))
]]
end

function ptpTimingInterface:raw_ts_to_time_float(rs)
    local ti = self.timing_info
    local ns = (tonumber(rs) - ti.ptp_ctr) * ti.ptp_to_ns
    local _s, _ = math.modf(ns / 1000000000)
    local s = tonumber(ti.xt_sec) + _s
    local _ns = ns % 1000000000
    local ns = tonumber(ti.xt_nsec) + _ns
    local _s, _ = math.modf(ns / 1000000000)
    s = s + _s
    ns = ns - (_s * 1000000000)
    return s, ns
end

function ptpTimingInterface:raw_ts_to_time_int(rs)
    local ti = self.timing_info
    local base_ctr_320 = rs - ti.ptp_ctr
    local base_ctr_ps = base_ctr_320 * ti.ptp_to_ps
    local base_wall_ps = ti.xt_sec * ti.trillion + ti.xt_nsec * ti.x1000
    local wall_ps = base_wall_ps + base_ctr_ps
    local s = wall_ps / ti.trillion
    local ns = wall_ps % ti.trillion
    return tonumber(s), tonumber(ns / ti.x1000)
end

function ptpTimingInterface:raw_ts_to_time_int_new(ts)
    local ti = self.timing_info
    -- Convert timestamp to nanoseconds
    local rs = (ffi.new('long long', ts.seconds) *
        ffi.new('long long', 1000000000)) + ffi.new('long long', ts.nanoseconds)
    local base_ctr_320 = rs - ti.ptp_ctr
    local base_ctr_ns = base_ctr_320 * ti.ptp_to_ns
    local base_wall_ns = ti.xt_sec * ffi.new('long long', 1000000000) +
        ti.xt_nsec
    local wall_ns = base_wall_ns + base_ctr_ns
    ts.seconds = wall_ns / 1000000000
    ts.nanoseconds = wall_ns % 1000000000
end

function ptpTimingInterface:raw_ts_to_time(rs)
    if type(rs) == 'cdata' then
        return self:raw_ts_to_time_int(rs)
    end
    return self:raw_ts_to_time_float(rs)
end

function ptpTimingInterface:to_internal_time(external)
    local int_max = 0x7fffffff
    if ffi.sizeof('int') == 8 then
        int_max = 0x7fffffffffffffffLL
    end
    if external.seconds.lsb <= int_max then
        return InternalTime:new(external.seconds.lsb, external.nanoseconds)
    end
end

function ptpTimingInterface:i64_to_internal_time(bigint)
    local v = (bigint.nanoseconds.msb * 4294967296) + bigint.nanoseconds.lsb
    local s, _ = math.modf(v / 1000000000)
    return InternalTime:new(s, v % 1000000000)
end

function ptpTimingInterface:PTPServo(offset, reset)
    local ti = self.timing_info
    if reset then
        self.servo.count = 0
        self.servo.integral = 0
        self.servo.last_offset = 0
        return
    end
    if self.servo.count == 0 then
        self.timing:AdjustClock(0, -offset, 'PTP')
        self.servo.integral = 0
    else
        self.servo.last_offset = self.servo.last_offset or offset
        local previous_error = self.servo.last_offset
        self.servo.error = offset
        self.servo.integral = self.servo.integral +
            (self.servo.error * self.servo.dt * self.servo.Ki)
        local derivative = (self.servo.error - previous_error) /
            self.servo.dt
        self.servo.output =
            (self.servo.Kp * self.servo.error) + self.servo.integral +
            (self.servo.Kd * derivative)
        self.timing:AdjustClockRate(-self.servo.output, 'PTP')
        self.servo.last_offset = offset
        if self.hw_stamped then
            if math.abs(self.servo.last_offset) < 125 then
                self.servo.Kp = 0.37 / 2
                self.servo.Ki = 0.10 / 2
                self.servo.Kd = 0.04 / 2
            elseif math.abs(self.servo.last_offset) < 250 then
                self.servo.Kp = 0.37
                self.servo.Ki = 0.10
                self.servo.Kd = 0.04
            else
                self.servo.Kp = 3.7
                self.servo.Ki = 1.0
                self.servo.Kd = 0.4
            end
        end
    end
    self.servo.count = self.servo.count + 1
end

return ptpTimingInterface
