local cjson = require 'cjson.safe'
local log = require 'log'
local sqlite3 = require 'lsqlite3'
local dbg = require 'dbg'
local armadb_client = require 'armadb_client'

-- Function used to exectute a database command
local function db_exec(db, fmt, ...)
    fmt = fmt:gsub('^[ ]*%|', '')
    fmt = fmt:gsub('\n[ ]*%|%+[ ]*', ' ')
    fmt = fmt:gsub('\n[ ]*%|', '\n')
    fmt = fmt:gsub('[ ]+$', '')
    local sql = string.format(fmt, ...)
    dbg('db_exec, sql=%s\n', sql)
    return db:exec(sql)
end

-- Function used to execute a database query
local function db_query(db, fmt, ...)
    fmt = fmt:gsub('^[ ]*%|', '')
    fmt = fmt:gsub('\n[ ]*%|%+[ ]*', ' ')
    fmt = fmt:gsub('\n[ ]*%|', '\n')
    fmt = fmt:gsub('[ ]+$', '')
    local sql = string.format(fmt, ...)
    local stmt = db:prepare(sql)
    if not stmt then return {} end
    local rrs = {}
    stmt:reset()
    for rr in stmt:rows() do
        rrs[#rrs + 1] = rr
    end
    return rrs
end

local results

-- Function used to change the setup file version in a setup
local function _migrate_version(setup, old_version, new_version)
    local s
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.id == 'summary' then
            if stanza.setup_file_version == old_version then
                stanza.setup_file_version = new_version
                results = (results or '\n') .. string.format(
                    "\t%s: change 'setup_file_version' '%s' -> '%s'\n",
                    stanza.name or stanza.id, old_version, new_version)
            else
                return
            end
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function used to change the name of a parameter in all setup stanzas
local function _rename_parameter(setup, old_name, new_name, category)
    local s
    category = category or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if category == 'all' or stanza.category == category then
            if stanza[old_name] ~= nil and stanza[new_name] == nil then
                results = (results or '\n') .. string.format(
                    "\t%s: rename parameter from '%s' to '%s'\n",
                    stanza.name or stanza.id, old_name, new_name)
                stanza[new_name] = stanza[old_name]
                stanza[old_name] = nil
            end
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to remove a parameter name from all setup stanzas
local function _delete_parameter(setup, parameter_name, category)
    local s
    category = category or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if category == 'all' or stanza.category == category then
            if stanza[parameter_name] ~= nil then
                results = (results or '\n') ..
                    string.format("\t%s: delete parameter '%s'\n",
                    stanza.name or stanza.id, parameter_name)
                stanza[parameter_name] = nil
            end
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to copy a parameter in each stanza
local function _copy_parameter(setup, src_param, dst_param, category)
    local s
    category = category or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if category == 'all' or stanza.category == category then
            if stanza[src_param] ~= nil and stanza[dst_param] == nil then
                results = (results or '\n') ..
                    string.format("\t%s: copy parameter '%s' to '%s'\n",
                    stanza.name or stanza.id, src_param, dst_param)
                stanza[dst_param] = stanza[src_param]
            end
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to copy a parameter in each stanza
local function _add_parameter(setup, param, value, category)
    local s
    category = category or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if category == 'all' or stanza.category == category then
            if stanza[param] == nil then
                results = (results or '\n') ..
                    string.format("\t%s: add parameter '%s'\n",
                    stanza.name or stanza.id, param)
            end
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function used to remove a parameter from one or all calibration stanza's
local function _delete_calibration_parameter(setup, parameter_name, purpose)
    local s
    purpose = purpose or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.calibration then
            local new_cal = {}
            for _, cal_stanza in ipairs(stanza.calibration) do
                if purpose == 'all' or
                    (cal_stanza.purpose and cal_stanza.purpose == purpose) then
                    if cal_stanza[parameter_name] ~= nil then
                        cal_stanza[parameter_name] = nil
                        results = (results or '\n') .. string.format(
                            "\t%s: delete parameter '%s' " ..
                            "from calibration purpose '%s'\n",
                            stanza.name or stanza.id,
                            parameter_name, cal_stanza.purpose)
                    end
                end
                new_cal[#new_cal + 1] = cal_stanza
            end
            stanza.calibration = new_cal
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to change a parameter name from one or all calibration stanza's
local function _rename_calibration_parameter(setup, old_name, new_name, purpose)
    local s
    purpose = purpose or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.calibration then
            local new_cal = {}
            for _, cal_stanza in ipairs(stanza.calibration) do
                if purpose == 'all' or
                    (cal_stanza.purpose and cal_stanza.purpose == purpose) then
                    if cal_stanza[old_name] ~= nil then
                        cal_stanza[new_name] = cal_stanza[old_name]
                        cal_stanza[new_name] = nil
                        results = (results or '\n') .. string.format(
                            "\t%s: rename parameter '%s' to " ..
                            "'%s' in calibration purpose '%s'\n",
                            stanza.name or stanza.id, old_name, new_name,
                            cal_stanza.purpose)
                    end
                end
                new_cal[#new_cal + 1] = cal_stanza
            end
            stanza.calibration = new_cal
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to add a parameter to one or all calibration stanza's
local function _add_calibration_parameter(setup, param, value, purpose)
    local s
    purpose = purpose or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.calibration then
            local new_cal = {}
            for _, cal_stanza in ipairs(stanza.calibration) do
                if purpose == 'all' or
                    (cal_stanza.purpose and cal_stanza.purpose == purpose) then
                    if cal_stanza[param] == nil then
                        cal_stanza[param] = value
                        results = (results or '\n') .. string.format(
                            "\t%s: add parameter '%s' " ..
                            "to calibration purpose '%s'\n",
                            stanza.name or stanza.id, param, stanza.purpose)
                    end
                end
                new_cal[#new_cal + 1] = cal_stanza
            end
            stanza.calibration = new_cal
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to copy a calibration parameter in each stanza.  The final
-- argument is optional and species a specific calibration step
local function _copy_calibration_parameter(setup, src_param, dst_param, purpose)
    local s
    purpose = purpose or 'all'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.calibration then
            local new_cal = {}
            for _, cal_stanza in ipairs(stanza.calibration) do
                if purpose == 'all' or
                    (cal_stanza.purpose and cal_stanza.purpose == purpose) then
                    if cal_stanza[dst_param] == nil and
                        cal_stanza[src_param] ~= nil then
                        cal_stanza[dst_param] = cal_stanza[src_param]
                        results = (results or '\n').. string.format(
                            "\t%s: copy parameter '%s' " ..
                            "to '%s' in calibration purpose '%s'\n",
                            stanza.name or stanza.id,
                            src_param, dst_param, stanza.purpose)
                    end
                end
                new_cal[#new_cal + 1] = cal_stanza
            end
            stanza.calibration = new_cal
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Sensor input functions

-- Function used to remove a parameter from one or all sensor input
local function _delete_sensor_parameter(setup, parameter_name, mode)
    local s
    mode = mode or 'any'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.sensor_input then
            local new_sensor = stanza.sensor_input
            if mode == 'any' or new_sensor.mode == mode then
                if new_sensor[parameter_name] ~= nil then
                    new_sensor[parameter_name] = nil
                    results = (results or '\n') .. string.format(
                        "\t%s: delete parameter '%s' " ..
                        "from sensor input mode '%s'\n",
                        stanza.name or stanza.id, parameter_name, mode)
                end
            end
            stanza.sensor_input = new_sensor
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to change a parameter name from one or all sensor_input
local function _rename_sensor_parameter(setup, old_name, new_name, mode)
    local s
    mode = mode or 'any'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.sensor_input then
            local new_sensor = stanza.sensor_input
            if mode == 'any' or new_sensor.mode == mode then
                if new_sensor[old_name] ~= nil and 
                    new_sensor[new_name] == nil then
                    new_sensor[new_name] = new_sensor[old_name]
                    new_sensor[old_name] = nil
                    results = (results or '\n') .. string.format(
                        "\t%s: rename parameter '%s' to " ..
                        "'%s' in sensor input mode '%s'\n",
                        stanza.name or stanza.id, old_name, new_name, mode)
                end
            end
            stanza.sensor_input = new_sensor
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to add a parameter to one or all sensor input
local function _add_sensor_parameter(setup, param, value, mode)
    local s
    mode = mode or 'any'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.sensor_input then
            local new_sensor = stanza.sensor_input
            if mode == 'any' or new_sensor.mode == mode then
                if new_sensor[param] == nil then
                    new_sensor[param] = value
                    results = (results or '\n') .. string.format(
                        "\t%s: add parameter '%s' " ..
                        "to sensor input mode '%s'\n",
                        stanza.name or stanza.id, param, mode)
                end
            end
            stanza.sensor_input = new_sensor
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to copy a sensor input parameter in each stanza.
local function _copy_sensor_parameter(setup, src_param, dst_param, mode)
    local s
    mode = mode or 'any'
    for _, stanza in ipairs(setup) do
        s = s or {}
        if stanza.sensor_input then
            local new_sensor = stanza.sensor_input
            if mode == 'any' or new_sensor.mode == mode then
                if new_sensor[old_name] ~= nil and
                    new_sensor[new_name] == nil then
                    new_sensor[new_name] = new_sensor[old_name]
                    results = (results or '\n') .. string.format(
                        "\t%s: copy parameter '%s' to " ..
                        "'%s' in sensor input mode '%s'\n",
                        stanza.name or stanza.id, old_name, new_name, mode)
                end
            end
            stanza.sensor_input = new_sensor
        end
        s[#s + 1] = stanza
    end
    return s
end

-- Function used to add unit_conversion stanza from 0.16 -> 0.17 migration
local function _add_units_conversion(setup)
    local s = {}
    for _, stanza in ipairs(setup) do
        if stanza.category and stanza.sensor_input and stanza.calibration then
           if stanza.category:find('hbm_mx1609') or
                (stanza.category:find('hbm_mx1615') and
                (stanza.sensor_input.mode:find('pt') or
                stanza.sensor_input.mode == 'strain_gage')) then
                local found
                local new_cal = {}
                for _, cal_stanza in ipairs(stanza.calibration) do
                    if cal_stanza.purpose == 'units_conversion' then
                        found = true
                    end
                    new_cal[#new_cal + 1] = cal_stanza
                end
                if not found then
                    results = (results or '\n') .. string.format(
                        "\t%s: add units_conversion calibration\n",
                        stanza.name or stanza.id)
                    new_cal[#new_cal + 1] = {
                        purpose = 'units_conversion',
                        applied = true,
                        units = stanza.output_units,
                        slope = 1,
                        intercept = 0,
                    }
                end
                results = (results or '\n') .. string.format(
                    "\t%s: add 'units_conversion_supported'\n",
                    stanza.name or stanza.id)
                stanza.units_conversion_supported = true
                stanza.calibration = new_cal
            end
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function used to migrate setups from v0.14 to v0.15
local function migrate_014_to_015(setup)
    -- Per page 8 of rev 15 SXR documentation - remove 'calibration_date'
    local new_setup = _delete_parameter(setup, 'calibration_date')
    if not new_setup then return setup end

    -- Per page 14 of rev 15 SXR documentation - Add 'rezero_target' to
    -- rezero calibration step.
    new_setup =
        _add_calibration_parameter(new_setup, 'rezero_target', 0, 'rezero')
    if not new_setup then return setup end

    -- Per page 15 of rev 15 SXR documentation - Change 'kfactor' of strain
    -- gage sensor input to 'gage_factor'
    new_setup =
        _rename_sensor_parameter(new_setup, 'kfactor', 'gage_factor',
            'strain_gage')
    if not new_setup then return setup end

    -- Per page 28 of rev 15 SXR documentation - Remove parameter
    -- 'brigde_gage_factor'
    new_setup = _delete_parameter(new_setup, 'bridge_gage_factor')
    if not new_setup then return setup end

    -- Per page 29 of rev 15 SXR documentation - Remove parameter
    -- 'brigde_bridge_factor'
    new_setup = _delete_parameter(new_setup, 'bridge_bridge_factor')
    if not new_setup then return setup end

    -- Per page 40 of rev 15 SXR documentation - Add parameter
    -- 'max_output_rate' to digital outputs
    new_setup = _add_parameter(new_setup,
        'max_output_rate', 1, 'sink/hbm_cx23r/digital_output')
    if not new_setup then return setup end

    -- Per page 41 of rev 15 SXR documentation - Add parameter
    -- 'transducer_power' to digital outputs
    new_setup = _add_parameter(new_setup,
        'transducer_power', 'off', 'sink/hbm_cx23r/digital_output')
    if not new_setup then return setup end
    return new_setup
end


-- Function used to migrate setups from v0.15 to v0.16
local function migrate_015_to_016(setup)
    -- Per page 11 of rev 16 SXR documentation - rename parameter
    -- 'phsycial_units' to 'output_units'
    local new_setup = _rename_parameter(setup, 'physical_units', 'output_units')
    if not new_setup then return setup end

    -- Per page 11 of rev 16 SXR documentation - rename parameter
    -- add parameter 'source_physical_units' (which is a copy of parameter
    -- 'electrical_units')
    new_setup =
        _copy_parameter(new_setup, 'electrical_units', 'source_physical_units')
    if not new_setup then return setup end

    -- Per page 15 of rev 16 SXR documentation - change value of parameter
    -- 'resistor' in sensor inputs who's mode is 'shunt'
    new_setup = _delete_sensor_parameter(new_setup, 'resistor', 'shunt')
    if not new_setup then return setup end
    new_setup = _add_sensor_parameter(new_setup, 'resistor', 100, 'shunt')
    if not new_setup then return setup end

    -- Per page 40 of rev 16 SXR documentation - remove parameter
    -- 'max_output_rate' from digital output channels
    new_setup = _delete_parameter(new_setup, 'max_output_rate',
        'sink/hbm_cx23r/digital_output')
    if not new_setup then return setup end

    -- Per page 47 of rev 16 SXR documentation - add unit conversion
    -- calibration step for thermocouple and strain gage sensor input types
    new_setup = _add_units_conversion(new_setup)
    if not new_setup then return setup end
    return new_setup
end


-- Function used to migrate setups from v0.16 to v0.17
local function migrate_016_to_017(setup)
    -- Per page 11 of rev 17 SXR documentation - remove parameter
    -- 'units_conversion_supported'
    local new_setup = _delete_parameter(setup, 'units_conversion_supported')
    if not new_setup then return setup end

    -- Per page 28 of rev 17 SXR documentation - add sensor input parameter
    -- 'regulation' for 'bridge' sensors.
    new_setup = _add_sensor_parameter(new_setup, 'regulation', true, 'bridge')
    if not new_setup then return setup end

    return new_setup
end


-- Function to add parameter bridge_lead_wire_resistance to bridge sensors that
-- don't have the parameter already.  This is for 0.17 -> 0.18 migration.
local function _add_bridge_lead_wire_resistance(setup)
    local s = {}
    for _, stanza in ipairs(setup) do
        if stanza.category and stanza.sensor_input then
            if stanza.category:find('hbm_mx1615') then
                if stanza.sensor_input.mode == 'bridge' then
                    stanza.bridge_lead_wire_resistance =
                        stanza.bridge_lead_wire_resistance or 0
                end
            end
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function to add data modes ordering parameters to the setup summary stanza
local function _add_data_modes_order_params(setup)
    local s = {}
    for _, stanza in ipairs(setup) do
        if stanza.category and stanza.category == 'summary' then
	    stanza.data_modes_sort_field = 'sort_order'
	    stanza.data_modes_sort_direction = 'ASC'
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function to add data_type parameter to computed channels
local function _add_data_type_param_to_computed_channels(setup)
    -- create a lookup table by channel name
    local channelsTable = {}
    for _, stanza in ipairs(setup) do
        if stanza.name and not string.match(stanza.category, '^sink/') then
            channelsTable[stanza.name] = stanza
        end
    end

    local s = {}
    for _, stanza in ipairs(setup) do
        if stanza.category and string.match(stanza.category, '^filter/') then
            -- for time base shift, up/down sampler, the data_type is inherited from the input_channel
            if string.match(stanza.category, 'filter/time_base_shifter') or string.match(stanza.category, 'filter/up_sampler') or string.match(stanza.category, 'filter/down_sampler') then
                local inputChannel
                inputChannel = stanza.input_channels[1]
                if (inputChannel) then
                    local inheritedChannel
                    inheritedChannel = channelsTable[inputChannel]
                    if inheritedChannel then
                        stanza.data_type = inheritedChannel.data_type
                    end
                end
            else
	        stanza.data_type = 'sequential_float64'	-- default filter channel data type for all others
            end
        end
        s[#s + 1] = stanza
    end
    return s
end


-- Function used to migrate setups from v0.17 to v0.18
local function migrate_017_to_018(setup)
    -- Per page 28 and 29 of rev 18 SXR documentation - add parameter
    -- 'bridge_lead_wire_resistance' to bridge channels
    local new_setup = _add_bridge_lead_wire_resistance(setup)
    if not new_setup then return setup end

    -- Nothing to do except adjust version number
    return new_setup
end


-- Function used to migrate setups from v0.18 to v0.20
local function migrate_018_to_020(setup)
    -- Nothing to do except adjust version number
    return setup
end


-- Function used to migrate setups from v0.20 to v0.21
local function migrate_020_to_021(setup)
    -- Add data modes ordering parameters
    local new_setup = _add_data_modes_order_params(setup)
    if not new_setup then return setup end

    return new_setup
end


-- Function used to migrate setups from v0.21 to v0.22
local function migrate_021_to_022(setup)
    -- Add data_type to filter channels
    local new_setup = _add_data_type_param_to_computed_channels(setup)
    if not new_setup then return setup end

    return new_setup
end


-- Function used to migrate setups from v0.22 to v1.22.
local function migrate_022_to_122(setup)
    -- Nothing to do except adjust version number
    return setup
end


-- Function used to migrate setups from v1.22 to v1.23.
local function migrate_122_to_123(setup)
    -- Nothing to do except adjust version number
    return setup
end

-- Table used to migrate setups from their current version to the last version
-- in the table.
local migration_table = {
    {from = 'v0.14', to = 'v0.15', migrate = migrate_014_to_015},
    {from = 'v0.15', to = 'v0.16', migrate = migrate_015_to_016},
    {from = 'v0.16', to = 'v0.17', migrate = migrate_016_to_017},
    {from = 'v0.17', to = 'v0.18', migrate = migrate_017_to_018},
    -- NOTE:  Version 0.19 never saw the light of day!
    {from = 'v0.18', to = 'v0.20', migrate = migrate_018_to_020},
    {from = 'v0.20', to = 'v0.21', migrate = migrate_020_to_021},
    {from = 'v0.21', to = 'v0.22', migrate = migrate_021_to_022},
    {from = 'v0.22', to = 'v1.22', migrate = migrate_022_to_122},
    -- NOTE:  Version 1.23 is the baseline for all migration strategies in the future
    {from = 'v1.22', to = 'v1.23', migrate = migrate_122_to_123},
}

-- Function used to return a table of active setups
local function get_active_setups(db)
    local setups
    local s = db_query(db, [==[
        |SELECT db_name FROM tsd_setups WHERE deleted = '' AND
        |     id = 'summary' and NOT content LIKE '%%database%%';
        ]==])
    for _, ss in ipairs(s) do
        local fmt = string.format(
            'SELECT db_name, content, id, added FROM tsd_setups WHERE '..
            '    deleted = "" AND db_name = "%s" ORDER by db_name;', ss[1])
        local r = db_query(db, fmt)
        for _, rr in ipairs(r) do
            setups = setups or {}
            if rr[1] ~= '' then
                setups[rr[1]] = setups[rr[1]] or {}
                local s = setups[rr[1]]
                s[#s + 1] = cjson.decode(rr[2])
            end
        end
    end
    return setups
end

-- Function used to show the active setups
local function show_setups(setups)
    for db_name, setup in pairs(setups) do
        if db_name ~= '' then
            for _, stanza in ipairs(setup) do
                if stanza.id == 'summary' then
                    print(string.format("Name:  %s, Version %s",
                        stanza.test_name, stanza.setup_file_version))
                end
--                print(cjson.encode(stanza))
            end
        end
    end
end

-- Function used to migrate setups from one version to the next.  This can take
-- a setup from the base version all the way to the maximum version supported.
local function migrate_setups(setups)
    for _, entry in ipairs(migration_table) do
        for db_name, setup in pairs(setups) do
            if db_name ~= '' then
                results = nil
                for _, stanza in ipairs(setup) do
                    if stanza.setup_file_version == entry.from then
                        results = (results or '\n') .. string.format(
                            "Migrating setup '%s' from version %s " ..
                            "to version %s\n",
                            stanza.test_name, stanza.setup_file_version,
                            entry.to)
                        if entry.migrate then
                            setups[db_name] = entry.migrate(setup)
                        end
                    end
                end
                -- The last item we deal with is migration of
                -- the setup file version
                local new_setup = _migrate_version(setup, entry.from, entry.to)
                if new_setup then setups[db_name] = new_setup end
                if results then log:notice(results) end
            end
        end
    end
end

-- Function used to migrate a setup from one version to another.  This is
-- primarily used by the API server when importing setups.
local function migrate_setup(db_name, setup)
    results = nil
    for _, entry in ipairs(migration_table) do
        for _, stanza in ipairs(setup) do
            if stanza.setup_file_version == entry.from then
                results = (results or '\n') .. string.format(
                    "Migrating setup '%s' from version %s to version %s\n",
                    stanza.test_name, stanza.setup_file_version, entry.to)
                if entry.migrate then
                    setup = entry.migrate(setup)
                end
            end
        end
        local new_setup = _migrate_version(setup, entry.from, entry.to)
        if new_setup then setup = new_setup end
    end
    if results then log:notice(results) end
    return setup, results
end

-- Function used to get the setup file version from a setup
local function get_version(setup)
    for _, stanza in ipairs(setup) do
        if stanza.id and stanza.id == 'summary' then
            return stanza.setup_file_version
        end
    end
end

-- Function used to update a setup
local function update_setup(db_name, setup_old, setup_new)
    local old_version = get_version(setup_old)
    local new_version = get_version(setup_new)
    if not old_version or not new_version then
        log:log(log.ERR, "Unable to get old/new version")
        return
    end

    -- If the old an new version are the same, we are done
    if old_version == new_version then
        return true
    end

    -- Get a connection to the API server
    local client = armadb_client:new({host='127.0.0.1'})
    if not client then
        log:log(log.ERR, "Unable to connect to API server")
        return
    end
    client:set({
        query = {
            authorization_key = 'locally_authorized',
            author = 'admin',
        }
    })

    -- Generate the JSON to send in the request
    local o = {}
    for _, v in ipairs(setup_old) do
        if v.id then
            o[v.id] = v
        end
    end
    local n = {}
    for _, v in ipairs(setup_new) do
        if v.id then
            n[v.id] = v
        end
    end
    local json = cjson.encode({
        description = string.format("Migrated from %s to %s",
            old_version, new_version),
        old = o,
        new = n,
    })

    -- Send the request and get a response
    local res, _ = client:request(
        'POST',
        string.format('/api/tracked_db/setups/%s/_delta', db_name),
        {data=json})
    if res and res.ok ~= nil then
        if res.ok ~= true then
            log:log(log.ERR, string.format("Error updating setup:  %s",
                res.error or "Unknown"))
            return
        end
    elseif v then
        log:log(log.ERR, string.format("Error updating setup:  %s", v))
    end
    return true
end


-- This is where the work is done
local function do_setup_migration()
    local db = sqlite3.open('/var/lib/armada/setup.db')
    local setups = get_active_setups(db)
    local save_setups = get_active_setups(db)
    db:close()
    local ok, err = pcall(migrate_setups, setups)
    if not ok then
        log:log(log.ERR, string.format(
            'Setup database content migration failed\n%s', err))
    else
        -- Now update all of the setups
        for db_name, old_setup in pairs(save_setups) do
            if update_setup(db_name, old_setup, setups[db_name]) ~= true then
                log:log(log.ERR,
                    'Setup database content migration failed during update')
                os.exit(1)
            end
        end

        log:notice('Setup database content migration complete')
--    show_setups(setups)
    end
    os.exit(0)
end


-- This function returns true if ALL of the active setups in the database are
-- compatible with this version of the system.  This basically means that all
-- active setups are at the top setup file version defined by the last
-- migration step in the migration table.
local function is_compatible()
    local db = sqlite3.open('/var/lib/armada/setup.db')
    local setups = get_active_setups(db)
    db:close()
    if not setups then return true end
    local latest_version = migration_table[#migration_table].to
    -- Now check the versions of each of the active setups
    for db_name, setup in pairs(setups) do
        if get_version(setup) ~= latest_version then
            return false
        end
    end

    -- If we got here, all of the setup versions are where they need to be
    return true
end


-- This function is used to determine if the setups are migrateable to the
-- version on this system.
local function is_migrateable()
    local db = sqlite3.open('/var/lib/armada/setup.db')
    local setups = get_active_setups(db)
    db:close()
    if not setups then return true end
    local migrateable = 0
    local num_setups = 0
    local latest_version = migration_table[#migration_table].to
    for db_name, setup in pairs(setups) do
        num_setups = num_setups + 1
        local setup_file_version = get_version(setup)
        if setup_file_version == latest_version then
            migrateable = migrateable + 1
        else
            for _, ms in ipairs(migration_table) do
                if setup_file_version == ms.from then
                    migrateable = migrateable + 1
                end
            end
        end
    end

    -- If all of the active setups are migrateable, return success otherwise
    -- return false
    if migrateable == num_setups then return true end
    return false
end

-- This function is used to retrieve the setup file version expected on this
-- system.
local function get_expected_setup_version()
    return migration_table[#migration_table].to
end

-- This function is used to retieve a table of information
local function get_setups_info()
    return {
        expected_setup_file_version = get_expected_setup_version(),
        setups_compatible = is_compatible(),
        setups_migrateable = is_migrateable(),
    }
end

local r = {
    do_migrate = do_setup_migration,
    get_info = get_setups_info,
    migrate_setup = migrate_setup,
}

return r
