require 'strict'
local bit = require 'bit'
local endian = require 'endian'
local ffi = require 'ffi'

require 'middleclass'

local C = ffi.C

local function dbg(fmt, ...)
    io.stdout:write(string.format(fmt, ...))
end

local function hexify(s)
    return string.gsub(s, "(.)", function(c)
        return string.format(" %02X", string.byte(c))
    end)
end

ffi.cdef[[
    enum {
        O_RDONLY = 0,
    };

    int open(const char *pathname, int flags, int mode);
    size_t read(int fd, const void *buf, size_t count);
    int shutdown(int sockfd, int how);
    int close(int fd);
    int64_t lseek64(int fd, int64_t offset, int whence);
    uint32_t crc32(uint32_t crc, const uint8_t *buf, int len);

    typedef struct __attribute((packed)) {
        uint32_t size;
        uint32_t group;
        uint32_t magic;
    } sie_block_header;

    typedef struct __attribute((packed)) {
        uint32_t crc;
        uint32_t size;
    } sie_block_trailer;

    typedef struct {
        int fd;
        uint32_t pos;
        uint32_t max_payload;
        sie_block_trailer *trailer;
        sie_block_header block;
    } sie_buffer;

    typedef struct __attribute((packed)) {
        uint64_t offset;
        uint32_t group;
    } sie_index_entry;

    /* For sanity checking index stuff */
    typedef struct {
        uint64_t offset;
        uint32_t size;
        uint32_t group;
    } sie_index_check_entry;
]]

local zlib = ffi.load('z')

local sie_overhead = 20
local sie_magic = 0x51EDA7A0

local m = class('sie_sanity')

function m:err(fmt, ...)
    local msg = string.format(fmt, ...)
    local hdr = self.block_header
    local extra = string.format('hdr(pos=%d=0x%X, size=0x%.08X, group=0x%.08X, magic=0x%.08X)\n', self.file_pos, self.file_pos, hdr.size, hdr.group, hdr.magic)
    -- local hex = hexify(ffi.string(hdr, 12))
    local pos = self.file_pos - 8
    pos = (pos < 0) and 0 or pos
    pos = bit.band(pos, bit.bnot(0x0F))
    local buflen = 64
    local buf = ffi.new('uint8_t[?]', buflen)
    C.lseek64(self.fd, pos, 0)
    local got = tonumber(C.read(self.fd, buf, buflen))
dbg('err(%s), C.read(%d, ..., %d)=%s\n', fmt, self.fd, buflen, got)

    local t = {}
    for i = 0, got - 1 do
        if bit.band(i, 0x0F) == 0 then
            t[#t + 1] = string.format("\n%.08X:", pos + i)
        end
        t[#t + 1] = string.format(" %.02X", buf[i])
    end
    local hex = table.concat(t, '') .. '\n'

    error(msg .. '\n' .. extra .. '\n' .. hex)
end

function m:initialize(args)
dbg('new, file_name=%s\n', args.file_name)
    self.file_name = args.file_name
    self.fd = C.open(self.file_name, C.O_RDONLY, 0)
    assert(self.fd >= 0)

    self.block_header = ffi.new('sie_block_header')
    self.block_trailer = ffi.new('sie_block_trailer')
    self.block_count = 0
    self.file_pos = 0

    self.read_check = {}
    self.index_check = {}
    self.xml = {}
end

function m:read(description, buf, size)
    local got = C.read(self.fd, buf, size)
    if got ~= size then
        if got == 0 then -- special case for EOF
            return 0
        end
        self:err('Error reading %s, wanted=0x%X, got=0x%X\n',
            description, size, got)
    end
    return got
end

function m:verify(cond, fmt, ...)
    if not cond then
        self:err(fmt, ...)
    end
end

function m:read_block()
    local got
    local hdr = self.block_header
    local tlr = self.block_trailer

    if self:read('block header', hdr, ffi.sizeof('sie_block_header')) == 0 then
        return nil
    end

    local crc = zlib.crc32(0, nil, 0)
    crc = zlib.crc32(crc, ffi.cast('uint8_t *', hdr), ffi.sizeof(hdr))

    hdr.size = endian.ntohl(hdr.size)
    hdr.group = endian.ntohl(hdr.group)
    hdr.magic = endian.ntohl(hdr.magic)

    local payload_size = hdr.size - sie_overhead

    self:verify(payload_size < 1000000000,
        'payload size implausible: 0x%X', payload_size)

    local payload = ffi.new('uint8_t[?]', payload_size)

    self:read('payload', payload, payload_size)
    self:read('block trailer', tlr, ffi.sizeof('sie_block_trailer'))

    crc = zlib.crc32(crc, payload, payload_size)

    tlr.crc = endian.ntohl(tlr.crc)
    tlr.size = endian.ntohl(tlr.size)

    self:verify(hdr.size == tlr.size, 'Header size 0x%x NOT trailer size 0x%x',
        hdr.size, tlr.size)
    self:verify(hdr.magic == sie_magic,
        'Header magic wrong: 0x%.08x', hdr.magic)

    self:verify(tlr.crc == crc,
        'Trailer CRC wrong: 0x%.08x, should be 0x%.08x',
        tlr.crc, 0, crc, 0)

    dbg('pos=0x%.08x size=0x%.08x group=0x%.08x magic=0x%.08x crc=0x%.08x BLOCK\n', self.file_pos, hdr.size, hdr.group, hdr.magic, tlr.crc)

    self:add_ice(self.read_check, self.file_pos, hdr.size, hdr.group)

    if hdr.group == 0 then
        -- XML
        self.xml[#self.xml + 1] = ffi.string(payload, payload_size)
    elseif hdr.group == 1 then
        -- Index block
--dbg('INDEX_DUMP %s\n', hexify(ffi.string(payload, payload_size)))
        local idx = ffi.cast('sie_index_entry *', payload)
        local count = payload_size / ffi.sizeof('sie_index_entry')
        for i = 0, count - 1 do
dbg('pos=0x%.08x group=0x%.08x INDEX 0x%.08x [%s]\n', tonumber(endian.ntohll(idx[i].offset)), tonumber(endian.ntohl(idx[i].group)), self.file_pos, hexify(ffi.string(idx + i, 12)))
            local offset = tonumber(endian.ntohll(idx[i].offset))
            local size = i < count - 1
                and endian.ntohll(idx[i + 1].offset) - offset or 0
            local group = endian.ntohl(idx[i].group)
            self:add_ice(self.index_check, offset, size, group)
        end
        self:add_ice(self.index_check, self.file_pos, hdr.size, hdr.group)
    end

    self.block_count = self.block_count + 1
    self.file_pos = self.file_pos + hdr.size

    return true
end

function m:add_ice(t, offset, size, group)
    t[#t + 1] = ffi.new('sie_index_check_entry', {offset, size, group})
end

function m:check_index()
    --table.sort(self.index_check, function(a, b) return a.offset < b.offset end)

    self:add_ice(self.index_check, self.file_pos, 0, -1) -- For last block len
    for i = 1, #self.index_check - 2 do
        local ie = self.index_check[i]
        if ie.size == 0 then
            ie.size = self.index_check[i + 1].offset - ie.offset
        end
    end
    self.index_check[#self.index_check] = nil -- For last block len

    local function dic(ice)
        return string.format('pos=0x%.08x size=0x%.08x group=0x%.08x',
            tonumber(ice.offset), tonumber(ice.size), tonumber(ice.group))
    end

    local ri, ii = 1, 1
    
    while ri <= #self.read_check or ii <= #self.index_check do
        local re = self.read_check[ri]
        local ie = self.index_check[ii]
        if not re then
            dbg('%s BADNESS index_check MISSING in read_check\n', dic(ie))
            ii = ii + 1
        elseif not ie then
            dbg('%s SUBOPTIMAL real_check MISSING in index_check\n', dic(re))
            ri = ri + 1
        elseif re and ie and re.offset == ie.offset then
            -- Cool, matching stuff
            local pos = tonumber(re.offset)
            if re.size ~= ie.size then
                local msg = ie.size == 0 and 'SUBOPTIMAL' or 'BADNESS'
                dbg('%s %s size 0x%.08x read / 0x%.08x indexed\n',
                    dic(re), msg, tonumber(re.size), tonumber(ie.size))
            end
            if re.group ~= ie.group then
                dbg('%s BADNESS group 0x%.08x read / 0x%.08x indexed\n',
                    dic(re), tonumber(re.group), tonumber(re.size))
            end
            if re.size == ie.size and re.group == ie.group then
                dbg('%s OK\n', dic(re))
            end
            ri = ri + 1
            ii = ii + 1
        elseif not ie or re.offset < ie.offset then
            dbg('%s SUBOPTIMAL real_check MISSING in index_check\n', dic(re))
            ri = ri + 1
        elseif not re or re.offset > ie.offset then
            dbg('%s BADNESS index_check MISSING in read_check\n', dic(ie))
            ii = ii + 1
        else
            error("This can't happen")
        end
    end
end

function m:emit_xml_file(xml_file_name)
    self.xml[#self.xml + 1] = '\n</sie>\n'
    if not xml_file_name then
        xml_file_name = self.file_name:gsub('%.[sS][iI][eE]$', '') .. '.xml'
    end
    local fh = assert(io.open(xml_file_name, 'w'))
    fh:write(table.concat(self.xml))
    assert(fh:close())
    self.xml_file_name = xml_file_name
    return xml_file_name
end

function m:jing()
    local armada_root = os.getenv('ARMADA_ROOT')
    local schema_file = armada_root .. '/src/armada/testing/SIE.rnc'
    local cmd = string.format("jing -c '%s' '%s'",
        schema_file, self.xml_file_name)
--cmd = '/bin/true'
    local got = os.execute(cmd)
    dbg('%s RETURNED %s\n', cmd, got)
end

function m:close()
    assert(C.close(self.fd) == 0)
end

return m
