require 'strict'
require 'middleclass'
local cjson = require 'cjson.safe'
local uuid = require 'uuid'
local config = require 'system_config'

local sys_notify = class('sys_notify')

-- Buttons is an array of
--[[
{
    text = string, -- Text to display in the button
    uri = string,  -- URI to use when the button is depressed
}
]]

local reboot_msg = (config.max_test_restarts == 0) and
[[  This error would normally result in the system automatically being reset but that is disabled because the 'maximum test restarts' preference is set to 0 (zero).]] or
[[  The system will be rebooted.]]

    

function sys_notify:initialize(title, buttons)
    self.bus = require('dbus').bus_get_private('system')
    self.bus:hook_libev_default_loop()
    self.network = self.bus:auto_proxy('com.somat.armada.Network',
        '/com/somat/armada/Network')
    if not self.network then return end
    self.buttons = {{
            text = 'OK',
            uri = '/api/system/acknowledge_notify',
        }}
    for _, button in ipairs(buttons or {}) do
        self.buttons[#self.buttons + 1] = button
    end
    self.title = title or "System Notification Message"
end

function sys_notify:_send_notification(notification)
    if not notification or type(notification) ~= 'table' or
        not self.network then return end

    self.my_notification = notification

    local do_reboot = notification.reboot
    local id = uuid.generate()
    self.owner = '_system-notification-' .. id

    self.my_notification.buttons = self.buttons
    self.my_notification.title = self.title
    self.my_notification.id = id
    self.my_notification = cjson.encode(self.my_notification)

    -- Notifications are persistent until explicitly removed via
    -- acknowledgement from an API call
    self.network:ModifyRRs({}, {{self.owner, 'TXT', self.my_notification}})

    if do_reboot then
        -- This is what triggers the system to be rebooted (if enabled)
        self.network:ModifyRRs({{self.owner, 'TXT', self.my_notification}},
            {})
    end

    -- If dont_close is set then just return.  This is only used by the test
    -- restart mechanism.  It does the close and unref itself.
    if self.dont_close then return end

    -- Close the DBUS connection
    self.bus:close()
    self.bus:unref()
    self.network = nil
end

function sys_notify:clear_notification()
    self.network:ModifyRRs({{self.owner, 'TXT', self.my_notification}}, {})
end

local code_map = {
    ['This is some bs message'] = 1000,
}

function sys_notify:critical_error(msg, reboot)
    self:_send_notification({
        type = 'error',
        severity = 'critical',
        msg = msg .. (reboot and reboot_msg or ''),
        code = code_map[msg],
        reboot = reboot,
    })
end

function sys_notify:error(msg, reboot, noled)
    self:_send_notification({
        type = 'error',
        severity = noled and 'normal' or 'error',
        msg = msg .. (reboot and reboot_msg or ''),
        code = code_map[msg],
        reboot = reboot,
    })
end

function sys_notify:warning(msg)
    self:_send_notification({
        type = 'warning',
        severity = 'none',
        msg = msg,
        code = code_map[msg],
    })
end

function sys_notify:info(msg)
    self:_send_notification({
        type = 'information',
        severity = 'none',
        msg = msg,
        code = code_map[msg],
    })
end

function sys_notify:notify(msg, hidden, announce_remove)
    self:_send_notification({
        type = (hidden and '_' or '') .. 'notice',
        severity = 'none',
        msg = msg,
        code = code_map[msg],
        announce_remove = announce_remove,
    })
end

return sys_notify
