require 'strict'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'
local net = require 'net'
local membuf = require 'membuf'
local sample_clock = require 'sample_clock'
local sample_types = require 'sample_types'
local message_can = require 'message_can'
local Frame = require 'test_engine/frame'
local Pipe = require 'test_engine/pipe'
local ev = require 'ev'
local C = ffi.C
local notify = require 'sys_notify'

ffi.cdef[[
    char *strerror(int errnum);
]]

local ffi_copy = ffi.copy
local ffi_cast = ffi.cast

local Frame_delete = Frame.delete

local Pipe_flush_frames = Pipe.flush_frames
local Pipe_get_frame = Pipe.get_frame

local ctor_cache

local mt = {}

function mt:proxy_reader()
    -- We just read the data and throw it away!
    local ar = C.read(self.fd, self.junk_buffer, 8192)
    if ar < 0 then
        self.read_errno = ffi.errno()
        if self.read_errno == C.EAGAIN then
            self.read_errno = 0
            return
        end
        self.read_ret = ar
        self.failed = 1
        C.shutdown(self.fd, C.SHUT_RDWR)
        C.close(self.fd)
        error(string.format("Error on connection to channel %s:  %s",
            ffi.string(self.channel),
            ffi.string(C.strerror(self.read_errno))))
        return
    elseif ar == 0 then
        self.read_errno = C.ECONNRESET
        self.read_ret = ar
        self.failed = 1
        C.shutdown(self.fd, C.SHUT_RDWR)
        C.close(self.fd)
        error(string.format(
            "Channel %s has unexpectedly disconnected from the test",
            ffi.string(self.channel)))
        return
    end
end

function mt:run()
    if self.in_pipe ~= nil then
        if self.fd <= 0 or self.latched ~= 0 then
            Pipe_flush_frames(self.in_pipe)
            return
        end
        local f = Pipe_get_frame(self.in_pipe)
        if f ~= nil then
            local ptr = ffi_cast('double *', f.buf)
            for i = 0, f.count - 1 do
                if ptr[i] ~= self.last_state then
                    local state = ptr[i] ~= 0 and self.on or self.off
                    local cj_enc = cjson.encode({
                        state = tonumber(state),
                    })
                    local amt_written = C.write(self.fd, cj_enc, #cj_enc)
                    if amt_written ~= #cj_enc then
                        local e = 'Short write'
                        if amt_written < 0 then
                            e = ffi.string(C.strerror(ffi.errno()))
                        end
                        error(string.format(
                            "Can not write to digital out '%s':  %s",
                            ffi.string(self.channel), e))
                    end
                    self.last_state = ptr[i]
                    if state == self.on and self.latched_mode ~= 0 then
                        self.latched = 1
                        break
                    end
                end
            end
            Frame_delete(f)
        end
    end
end

function mt:setup()
end

function mt:finalize()
end

function mt:__gc()
    if self.channel ~= nil then C.free(self.channel) end
end

local function proxy_ctor()
    if not ctor_cache then
        ffi.cdef([[
            struct proxy_sink {
                struct pipe     *in_pipe;
                int             i;
                int             fd;
                int             failed;
                int             read_errno;
                ssize_t         read_ret;
                char            *channel;
                char            junk_buffer[8192];
                double          last_state;
                int             latched_mode;
                int             latched;
                int             on;
                int             off;
            };
        ]])
        ctor_cache = ffi.metatype('struct proxy_sink', {__index = mt})
    end
    return ctor_cache
end

local m = {}

function m:new(loop, md, host, port, ip)
    local ctor = proxy_ctor()
    local pn = ctor()
    pn.channel = C.malloc(md.name:len())
    ffi_copy(pn.channel, md.name)
    pn.fd = 0

    -- Set whether or not we latch upon transition to true
    if md.output_mode:find('_latched') then
        pn.latched_mode = 1
    else
        pn.latched_mode = 0
    end

    -- Set level to use for "on" and "off"
    if md.output_mode:find('^clear') then
        pn.on = 0
        pn.off = 1
    else
        pn.on = 1
        pn.off = 0
    end

    -- If the connection is remote, use IP otherwise use UNIX domain socket
    local sock = net.socket{
        family = C.AF_UNIX,
        type = C.SOCK_STREAM,
        flags = 0}

    -- Function called when the connection to a channel proxy completes
    local function handle_connect_complete(s)
        if not s then return end
        pn.fd = s.sock
        local w = ev.IO.new(function(l, i, e) pn:proxy_reader() end,
            s.sock, ev.READ)
        self.reader_watcher = w
        w:start(loop)
        local out_md = cjson.encode(md)
        local msg = string.format(
            "POST /channel_sink_stream HTTP/1.0\r\n" ..
            "Content-length: %d\r\n\r\n%s", #out_md, out_md)
        local written = C.write(s.sock, msg, #msg)
        if written ~= #msg then
            error(string.format("Unable to send channel setup request for %s",
                md.name))
        end
    end

    -- Function called when UNIX domain socket connect finishes
    local function handle_unix_connect_complete(s)
        if s then
            return handle_connect_complete(s)
        end
        -- UNIX domain connect failed, try IP domain
        sock:close()
        net.socket{ip = ip, port = port}:connect_nb(
            loop, handle_connect_complete)
    end

    -- Setup for connect via UNIX domain socket
    local sa = ffi.new('struct sockaddr_un[1]')
    sa[0].sun_family = C.AF_UNIX
    local name = md.connector:match('^.*%@(.+)$')
    ffi.copy(sa[0].sun_path, '/tmp/proxy/' .. name)

    sock.addrs = ffi.new('struct addrinfo')
    sock.addrs.ai_flags = 0
    sock.addrs.ai_family = C.AF_UNIX
    sock.addrs.ai_socktype = C.SOCK_STREAM
    sock.addrs.ai_protocol = 0
    sock.addrs.ai_addrlen = ffi.sizeof('struct sockaddr_un')
    sock.addrs.ai_addr = ffi.cast('struct sockaddr *', sa)
    sock.addrs.ai_next = nil

    sock:connect_nb(loop, handle_unix_connect_complete)

    return pn, 0
end

return m
