require 'strict'

-- histogram.lua - Implements a multidimensional histogram algorithm

local math_floor = math.floor
local math_huge = math.huge
local math_min = math.min
local math_max = math.max

local m = {}

-- Executes an algorithm which increments the count of the proper
-- histogram bin based on the input.
-- Parameter:
--     'hist'  table containing information for a histogram
--     'data'  array containing the multi-dimensional
--             data point to be histogrammed
--     'len'   number of data point sets in 'data'
function m.histogram(hist, data, len)
    local dim
    local bin            -- index for the current histogram dimension
    local dat_in = 0     -- index into input data
    local mat_index = 0  -- final index into bin matrix

    -- determine the correct index into the bin matrix
    while dat_in < len do
        -- calculate matrix index
        for n = 0, hist.nbr_dims - 1 do
            dim = hist.dims[n]
            -- determine where in this dimension the data point falls
            bin = math_floor((data[dat_in] - dim.base) / dim.bin_size)
            dat_in = dat_in + 1
            -- check for out of bounds, in which case overflow bins are used
            bin = math_max(0, bin)
            bin = math_min(dim.total_dimension_size - 1, bin)
            -- The following lines execute the next steps in calculating
            -- the target bin's offset into the histogram's bin matrix.
            -- That matrix is represented as linear array, so these
            -- calculations result in:
            --     Index = ((((((x0 * d1) + x1) * d2) + x2) * d3) + x3) ...
            -- where:
            --     x0, x1, ... refer to bin indices and
            --     d0, d1, ... refer to dimension sizes
            mat_index = mat_index + bin
            if n < hist.nbr_dims - 1 then
                -- point to histogram information for next dimension
                dim = hist.dims[n + 1]
                -- track the offset into the histogram
                mat_index = mat_index * dim.total_dimension_size
            end
        end

        -- increment the targeted bin (don't if already pegged out)
        if hist.binmatrix[mat_index] < math_huge then
            hist.binmatrix[mat_index] = hist.binmatrix[mat_index] + 1
        end
        mat_index = 0
    end
end

return m

