require 'strict'


-- pv_slice_algo.lua - Implements the Peak Valley Processing Algorithm
-- for peak valley slice data modes.  Note that there is close coupling
-- between the "chan_info" tables expected by these functions and the
-- the tables as defined by the datamode code using this file.

local m = {}

-- Used by state "initial".  Saves in the info for each "channel to
-- output" the channel's current input sample value as either a peak
-- or valley candidate.  These values will be used when a
-- peak or valley is detected unless overwritten by subsequent
-- candidate values in "initial" state.
-- Parameters:
--     'is_peak' true if a set of peak candidates is to be saved,
--         false if a set of valley candidates is to be saved
--     'mc' index in 'chans' of the master channel associated with
--         this candidate
--     'index' index to the current sample for each input frame
--     'ts' timestamp for the samples currently being collected
--     'chans'  array of tables containing information for
--         input channels
--     'nbr_output_chans'  number of channels in 'chans' which
--         need to contribute a sample to a slice
function m.first_candidate_slice(is_peak, mc, index, ts,
                                 chans, nbr_output_chans)
    if is_peak then
        if chans[mc].first_peak_candidate == nil then
            chans[mc].first_peak_candidate = {}
            chans[mc].first_peak_candidate.vals = {}
        end
        chans[mc].first_peak_candidate.ts = ts
        for c = 1, nbr_output_chans do
            chans[mc].first_peak_candidate.vals[c] = chans[c].invals[index]
        end
    else
        if chans[mc].first_valley_candidate == nil then
            chans[mc].first_valley_candidate = {}
            chans[mc].first_valley_candidate.vals = {}
        end
        chans[mc].first_valley_candidate.ts = ts
        for c = 1, nbr_output_chans do
            chans[mc].first_valley_candidate.vals[c] = chans[c].invals[index]
        end
    end
end

-- Saves in the info for each "channel to output" the channel's
-- current input sample value.  These values will be used when a
-- peak or valley is detected unless overwritten by subsequent
-- candidate values.
-- Parameters:
--     'mc' index in 'chans' of the master channel associated with
--         this candidate
--     'index' index to the current sample for each input frame
--     'ts' timestamp for the samples currently being collected
--     'chans'  array of tables containing information for
--         input channels
--     'nbr_output_chans'  number of channels in 'chans' which
--         need to contribute a sample to a slice
function m.save_slice_candidate(mc, index, ts, chans, nbr_output_chans)
    if chans[mc].candidate == nil then
        chans[mc].candidate = {}
        chans[mc].candidate.vals = {}
    end
    chans[mc].candidate.ts = ts
    for c = 1, nbr_output_chans do
        chans[mc].candidate.vals[c] = chans[c].invals[index]
    end
end

-- Puts in given array a slice consisting of the first peak or valley
-- candidate for the specified master channel.  Note that this
-- function assumes that only one (if any) of first_peak_candidate and
-- first_valley_candidate will be non-nil.  Both the master's "first"
-- candidate slices are then set to nil (even though one is already
-- nil).
-- stored.
-- Parameters:
--     'slices'  array to which slice is to be appended
--     'info'  table of information for the master channel which
--         has detected a peak or valley
function m.take_first_slice(slices, info)
    if info.first_peak_candidate ~= nil then
        slices[#slices + 1] = info.first_peak_candidate
    elseif info.first_valley_candidate ~= nil then
        slices[#slices + 1] = info.first_valley_candidate
    end
    info.first_peak_candidate = nil
    info.first_valley_candidate = nil
end

-- Puts in given array the current candidate slice for the specified
-- master channel.  The master's candidate reference is then set
-- to nil.
-- Parameters:
--     'slices' array to which slice is to be appended
--     'info'  table of information for the master channel which
--         has detected a peak or valley
function m.take_slice(slices, info)
    if info.candidate ~= nil then
        slices[#slices + 1] = info.candidate
        info.candidate = nil
    end
end

-- NOTE:  This algorithm is tailored for the Peak-Valley-type
--        datamodes which can use multiple "master" channels.
-- Executes the Peak Valley Processing algorithm on the given
-- input frames, producing slices of sample values to output
-- (one slice for each input channel).
-- Parameters:
--     'in_count'  the number of samples in each input frame
--     'chans'  array of tables containing information for
--         input channels
--     'nbr_master_chans'  number of "master" channels in 'chans' -
--         these channels will be first in the array
--     'nbr_output_chans'  number of channels in 'chans' which
--         need slices - these channels will be first in the array
--         (and master channels will be first among these)
--     'hysteresis'  change needed before peak/valley determined
--     'in_ts'  the timestamp of the current input frames
-- Returns: table of slice arrays, one for each master channel
function m.run_pv_slice_algo(in_count, chans, nbr_master_chans,
             nbr_output_chans, hysteresis, in_ts)

    local slice_table = {}

    for c = 1, nbr_master_chans do
        local slices = {}
        slice_table[c] = slices
        local input_info = chans[c]
        local iv = input_info.invals

        -- NOTE:  For states "peak_search" and "valley_search"
        --        comparisons of the current input sample (iv[index])
        --        to the existing peak or valley values use >= or <=
        --        instead of just > or <, so that the candidate slices
        --        saved are kept as close to the current timestamp as
        --        possible.  This is not done with the "initial" state
        --        to keep the first sample results the same as for the
        --        eDAQs.
        for index = 0, in_count - 1 do
            if input_info.state == "initial" then
                if iv[index] < input_info.valley then
                    input_info.valley = iv[index]
                    m.first_candidate_slice(
                        false, c, index, in_ts + index,
                        chans, nbr_output_chans)
                elseif iv[index] - input_info.valley > hysteresis then
                    -- valley found, save slice
                    chans[c].first_peak_candidate = nil
                    m.take_first_slice(slices, input_info)
                    -- set up search for next peak
                    input_info.state = "peak_search"
                    input_info.peak = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                end

                if input_info.state == "initial" and
                   iv[index] > input_info.peak then
                    input_info.peak = iv[index]
                    m.first_candidate_slice(
                        true, c, index, in_ts + index,
                        chans, nbr_output_chans)
                elseif input_info.peak - iv[index] > hysteresis then
                    -- peak found, save slice
                    chans[c].first_valley_candidate = nil
                    m.take_first_slice(slices, input_info)
                    -- set up search for next valley
                    input_info.state = "valley_search"
                    input_info.valley = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                end
            elseif input_info.state == "valley_search" then
                -- look for a valley
                if iv[index] <= input_info.valley then
                    input_info.valley = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                elseif iv[index] - input_info.valley > hysteresis then
                    -- valley found, save slice (if not a duplicate)
                    m.take_slice(slices, input_info)
                    input_info.state = "peak_search"
                    input_info.peak = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                end
            elseif input_info.state == "peak_search" then
                -- look for a peak
                if iv[index] >= input_info.peak then
                    input_info.peak = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                elseif input_info.peak - iv[index] > hysteresis then
                    -- peak found, save slice (if not a duplicate)
                    m.take_slice(slices, input_info)
                    input_info.state = "valley_search"
                    input_info.valley = iv[index]
                    m.save_slice_candidate(c, index, in_ts + index,
                                           chans, nbr_output_chans)
                end
            else
                assert(false, 'Invalid PV datamode state for input ' ..
                    tostring(input_info.name) .. '(state = ' ..
                    tostring(input_info.state) .. ')')
            end
        end
    end

    return slice_table
end

-- Find the oldest candidate slice among all the master channels.
-- Parameters:
--     'chans'  array of tables containing information for
--         input channels
--     'nbr_master_chans'  number of "master" channels in 'chans' -
--         these channels will be first in the array
-- Returns: first the oldest candidate slice (nil if none exist),
--     and then the index into 'chans' of the master channel
--     associated with that candidate
function m.get_oldest_candidate(chans, nbr_master_chans)
    local oldest = nil
    local mc = nil

    for c = 1, nbr_master_chans do
        if chans[c].state == "initial" then
            if chans[c].first_valley_candidate ~= nil then
                if oldest == nil then
                    oldest = chans[c].first_valley_candidate
                    mc = c
                elseif chans[c].first_valley_candidate ~= nil then
                    if oldest.ts > chans[c].first_valley_candidate.ts then
                        oldest = chans[c].first_valley_candidate
                        mc = c
                    end
                end
            end
            if chans[c].first_peak_candidate ~= nil then
                if oldest == nil then
                    oldest = chans[c].first_peak_candidate
                    mc = c
                else
                    if oldest.ts > chans[c].first_peak_candidate.ts then
                        oldest = chans[c].first_peak_candidate
                        mc = c
                    end
                end
            end
        else
            if oldest == nil then
                oldest = chans[c].candidate
                mc = c
            elseif chans[c].candidate ~= nil then
                if oldest.ts > chans[c].candidate.ts then
                    oldest = chans[c].candidate
                    mc = c
                end
            end
        end
    end

    return oldest, mc
end

return m

