require 'strict'

-- anomaly_detect.lua - Implements the Anomaly Detect computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local IDX_FLATLINE  = 0
local IDX_DRIFT     = 1
local IDX_LIMITS    = 2
local IDX_KURTOSIS  = 3

local BIT_FLATLINE  = bit.lshift(1, IDX_FLATLINE)
local BIT_DRIFT     = bit.lshift(1, IDX_DRIFT)
local BIT_LIMITS    = bit.lshift(1, IDX_LIMITS)
local BIT_KURTOSIS  = bit.lshift(1, IDX_KURTOSIS)

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    self.of.hdr.ts = self.tot_count
    self.of.ts = math.floor(self.tot_count)
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
        self.out_vals = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

function mt:calculate_bit_mask()
    local mean = 0
    self.mask = 0
    if self.flatline then
        if self.w_max - self.w_min < self.flatline_gate then
            self.mask = bit.bor(self.mask, BIT_FLATLINE)
        end
    end
    if self.drift or self.kurtosis then
        mean = self.w_mean_sum / self.factor
    end
    if self.drift then
        if not self.have_first_w_mean then
            self.first_w_mean = mean
            self.have_first_w_mean = true
        else
            if math.abs(mean - self.first_w_mean) > self.drift_gate then
                self.mask = bit.bor(self.mask, BIT_DRIFT)
            end
        end
    end
    if self.limits and self.limit_exceeded then
        self.mask = bit.bor(self.mask, BIT_LIMITS)
    end
    if self.kurtosis then
        local diff, m2, kurtosis = nil, nil, nil
        local sum_m2, sum_m4 = 0, 0
        for kk = 0,self.factor-1 do
            diff = self.fbuf[kk] - mean
            m2 = diff * diff
            sum_m2 = sum_m2 + m2
            sum_m4 = sum_m4 + m2*m2
        end
        if sum_m2 > 0 then
            kurtosis = (self.factor * sum_m4) / (sum_m2 * sum_m2)
        else
            kurtosis = 0
        end
        if kurtosis > self.kurtosis_max then
            self.mask = bit.bor(self.mask, BIT_KURTOSIS)
        end
    end
end

-- Track window based params
function mt:track_params(size, inp_index, in_vals, fbuf_offset)
    for w = 0, size-1 do
        if self.flatline then
            if in_vals[inp_index+w] > self.w_max then
                self.w_max = in_vals[inp_index+w]
            elseif in_vals[inp_index+w] < self.w_min then
                self.w_min = in_vals[inp_index+w]
            end
        end
        if self.drift or self.kurtosis then
            self.w_mean_sum = self.w_mean_sum + in_vals[inp_index+w]
        end
        if self.limits then
            if in_vals[inp_index+w] > self.limit_max or
               in_vals[inp_index+w] < self.limit_min then
               self.limit_exceeded = true
            end
        end
        if self.kurtosis then
            self.fbuf[fbuf_offset+w] = in_vals[inp_index+w]
        end
    end
end

function mt:new_window_analysis_check(in_vals, inp_index)
    if self.w_res == self.factor then
        if self.flatline then
            self.w_min, self.w_max = nil, nil
            self.w_min = in_vals[inp_index]
            self.w_max = in_vals[inp_index]
        end
        if self.drift or self.kurtosis then
            self.w_mean_sum = 0
        end
        if self.limits then
            self.limit_exceeded = false
        end
    end
end

-- Iterate through all of input buffer, unless the size needed
-- to complete a window analysis is less than the input buffer size
function mt:analysis_size_and_offset(f, inp_index)
    local size = 0
    if f.count - inp_index > self.w_res then
        size = self.w_res
    else
        size = f.count - inp_index
    end
    local fbuf_offset = self.factor - self.w_res
    assert(size > 0, "Size is "..size)
    return size, fbuf_offset
end

--Finished reading input frame
function mt:check_window_end()
    if self.w_res == 0 then
        self.w_res = self.factor
        self:calculate_bit_mask()
        self.out_vals[self.out_index] = self.mask
        self.out_index = self.out_index + 1
        if self.out_index == self.bufsize then
            self:flush_output_frame()
        end
    end
end

function mt:process_in_frame(f)
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
    local inp_index = 0 -- Index to input frame data
    while inp_index < f.count do
        self:new_window_analysis_check(in_vals, inp_index)
        local size, fbuf_offset = self:analysis_size_and_offset(f, inp_index)
        self:track_params(size, inp_index, in_vals, fbuf_offset)
        inp_index = inp_index + size
        self.w_res = self.w_res - size
        self:check_window_end()
    end
end

-- Processes incoming frames, produces outgoing frames.
-- This is broken into many smaller functions due to the way the
--   LuaJIT compiler handles edge-cases.  To be ported to C in the future.
function mt:run()
    while Pipe_has_frames(self.in_pipe) do
        local f = Pipe_get_frame(self.in_pipe) -- input frame
        assert(f.count > 0, "Input frame had "..f.count.." elements")
        self:process_in_frame(f)
        Frame_delete(f)
        f = nil
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local anom_d = setmetatable({}, {__index = mt})
    anom_d.sie_collect_idx = first_sie_index
    utils.set_common_meta(anom_d, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Anomaly Detect computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical ~= true, "Input channel must not be a logical "
        .."channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    anom_d.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(anom_d.pipe_rate and anom_d.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(anom_d.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    anom_d.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(anom_d.sample_rate and anom_d.sample_rate > 0,
        "Invalid sample rate: " .. tostring(anom_d.sample_rate))
    anom_d.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(anom_d.out_frame_sample_size and anom_d.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(anom_d.out_frame_sample_size))

    --Module specific parameters
    anom_d.factor = tonumber(stanza.factor) --size of window analysis in samples
    assert(anom_d.factor and anom_d.factor%1 == 0 and anom_d.factor > 1,
        "Invalid factor: "..tostring(stanza.factor)..
        ", must be an integer greater than 1.")
    --IFDEF flatline, drift, limits, kurtosis
    anom_d.flatline = stanza.flatline
    if anom_d.flatline then
        anom_d.flatline_gate = tonumber(stanza.flatline_gate)
        assert(anom_d.flatline_gate and anom_d.flatline_gate >= 0,
            "Invalid flatline gate: "..tostring(stanza.flatline_gate)..
            ", must be a number >= 0.")
        anom_d.w_min = 0
        anom_d.w_max = 0
    end
    anom_d.drift = stanza.drift
    if anom_d.drift then
        anom_d.drift_gate = tonumber(stanza.drift_gate)
        assert(anom_d.drift_gate and anom_d.drift_gate >= 0,
            "Invalid drift gate: "..tostring(stanza.drift_gate)..
            ", must be a number >= 0.")
        anom_d.have_first_w_mean = false
        anom_d.w_mean_sum = 0
    end
    anom_d.limits = stanza.limits
    if anom_d.limits then
        anom_d.limit_min = tonumber(stanza.limit_min)
        assert(anom_d.limit_min, "Invalid limit_min: "..
            tostring(stanza.limit_min)..", must be a number")
        anom_d.limit_max = tonumber(stanza.limit_max)
        assert(anom_d.limit_max, "Invalid limit_max: "..
            tostring(stanza.limit_max)..", must be a number")
        assert(anom_d.limit_max >= anom_d.limit_min,
            "Invalid limit range, must have limit_max >= limit_min.")
        anom_d.limit_exceeded = false
    end
    anom_d.kurtosis = stanza.kurtosis
    if anom_d.kurtosis then
        anom_d.kurtosis_max = tonumber(stanza.kurtosis_max)
        assert(anom_d.kurtosis_max and anom_d.kurtosis_max >= 0,
            "Invalid kurtosis max: "..tostring(stanza.kurtosis_max)..
            ", must be a number >= 0.")
        anom_d.w_mean_sum = 0
        anom_d.fbuf = {}
    end

    -- create input/output pipe structures
    anom_d.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(anom_d.in_pipe and anom_d.in_pipe ~= nil,
        "failed to create input pipe")
    anom_d.pipe_list = Pipe_new_pipe_list()
    assert(anom_d.pipe_list and anom_d.pipe_list ~= nil,
        "failed to create (output) pipe list")

    --local variables needed not set from stanza.
    anom_d.tot_count = 0
    anom_d.bufsize = frame.bufsize(anom_d.sample_rate, anom_d.pipe_rate)
    -- output frame
    anom_d.of = Frame_create(anom_d.bufsize, anom_d.out_frame_sample_size)
    -- "pointer" to output frame data
    anom_d.out_vals = ffi_cast(Utils_frame_data_type(
        anom_d.out_frame_sample_size), anom_d.of.buf)
    anom_d.out_index = 0

    anom_d.mask = 0 -- the output bitmask
    anom_d.w_res = anom_d.factor -- # of samples needed to complete window

    return anom_d, 1
end

return m
