require 'strict'

-- integrator.lua - Implements the Integrator computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local dioconst = require 'dioconst'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sample_clock = require 'sample_clock'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_ceil = math.ceil
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list

-- per issue #2439, SomatXR Integrator needs to support the same
-- triggering options as the eDAQ Integrator, for at least all real
-- (i.e., floating point) input channels.  Summary of the functionality
-- required (copied from issue #2439):
--
-- Trigger / Reset Options - A (logical) trigger channel can be
-- specified for either or both of the following two purposes.
--
-- 1) Integrate Only When True - Check this option to have the
--    integration suppressed when the trigger channel is FALSE.
--
-- 2) Enable Triggered Reset - Check this option to have the Integrator
--    reset (back to the specified Initial Value) whenever the trigger
--    channel satisfies the selected Reset Mode.
--
-- There are three trigger reset modes available.
--
--     a) When True - Reset whenever the trigger channel is TRUE.
--
--     b) On False-True Edge - Reset when the trigger channel
--        transitions from FALSE to TRUE.
--
--     c) After True-False Edge - Reset on the sample after the trigger
--        channel transitions from TRUE to FALSE. (If the Integrate Only
--        When True option is not selected, then the output sample after
--        the True to False edge is the sum of the Initial Value and the
--        scaled input sample.)
--
--     Note - When both the Integrate Only When True and the Enable
--     Triggered Reset options are selected, the Reset Mode is limited
--     to After True-False Edge.
--
-- Reset on Sum Exceeds - Check this option to have the Integrator
--    reset to the Initial Value whenever the absolute value of the
--    Integrator sum exceeds the (positive) value specified in the
--    associated edit field.
--
-- Note - If the Reset on Sum Exceeds option is selected then none of
-- the other Trigger / Reset Options can be used.


local mt = {}

function mt:dump()
    dbg('out_value=%lld, triggered=%s', self.out_value, self.triggered)
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) or
       (self.trig_pipe ~= nil and not Pipe_has_frames(self.trig_pipe)) then
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local in_vals      -- "pointer" to current input frame's buf
    local of  -- current output frame
    local out_vals     -- "pointer" to current output frame's buf
    local tf = nil
    if self.trig_pipe ~= nil then
        tf = Pipe_get_frame(self.trig_pipe)
    end
    local f = Pipe_get_frame(self.in_pipe)
    local process_input = true
    while process_input do
        -- create frame for output
        of, in_vals, out_vals =
            utils.create_output_frame(f.maxsize, f,
                self.tot_count, self.out_frame_sample_size)
        -- fill output frame
        local of_val_index = 0
        local pre_reset = false
        local post_reset = false
        local integrate = true
        local prev_value
        local integrated_value
        local trig_vals
        if tf ~= nil then
            trig_vals = ffi_cast(
                tf.element_size == 4 and 'float *' or 'double *', tf.buf)
        end
        for in_buf_index = 0, f.count - 1 do
            prev_value = self.out_value
            integrated_value =
                prev_value + in_vals[in_buf_index] * self.factor
            -- Reset On Sum Exceeds trumps other trigger/reset options
            if self.reset_when_sum_exceeds then
                if math.abs(integrated_value) > self.sum_maximum_value then
                    pre_reset = true
                end
            elseif tf ~= nil then
                local prev_triggered = self.triggered
                self.triggered = (trig_vals[in_buf_index] ~= 0) and
                    true or false
                if self.integrate_only_when_true then
                    integrate = self.triggered
                end
                if self.enable_triggered_reset then
                    if self.reset_mode == 'when_true' then
                        pre_reset = self.triggered
                    elseif self.reset_mode == 'on_false_to_true_edge' then
                        pre_reset = not prev_triggered and self.triggered
                    elseif self.reset_mode == 'after_true_to_false_edge' then
                        post_reset = prev_triggered and not self.triggered
                    end
                end
            end

            -- set output value
            if pre_reset then
                self.out_value = self.initial_value
                pre_reset = false
            elseif integrate then
                self.out_value = integrated_value
                integrate = true
            else
                self.out_value = prev_value
            end
            out_vals[of_val_index] = self.out_value
            of_val_index = of_val_index + 1
            if post_reset then
                post_reset = false
                self.out_value = self.initial_value
            end
        end

        of.count = f.count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end

        -- Delete frames that have been processed and get another set.
        -- Make sure we only get frames when all of the pipes have frames
        Frame_delete(f)
        f = nil
        if self.trig_pipe ~= nil then
            Frame_delete(tf)
            tf = nil
            if Pipe_has_frames(self.in_pipe) and
                Pipe_has_frames(self.trig_pipe) then
                f = Pipe_get_frame(self.in_pipe)
                tf = Pipe_get_frame(self.trig_pipe)
            end
            process_input = f ~= nil and tf ~= nil
        else
            f = Pipe_get_frame(self.in_pipe)
            process_input = f ~= nil
        end
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    local integrator = setmetatable({}, {__index = mt})
    integrator.stanza = stanza
    integrator.sie_collect_idx = first_sie_index
    -- validate needed metadata
    assert(integrator.stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = integrator.stanza.input_channels or {}
    assert(#input_channels == 1, "Integrator computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    integrator.sample_rate =
        rational:new(stanza.sample_rate):tonumber()
    assert(integrator.sample_rate and integrator.sample_rate > 0,
        "Invalid sample rate: " .. stanza.sample_rate)
    --   output frame element size
    assert(integrator.stanza.data_type, "Missing data type")
    integrator.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(integrator.out_frame_sample_size and
           integrator.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(integrator.out_frame_sample_size)))
    --   scale factor
    integrator.factor = tonumber(integrator.stanza.scale_factor) or 1
    --   initial sample value
    integrator.initial_value = integrator.stanza.initial_value
    integrator.initial_value = tonumber(integrator.initial_value) or 0
    --   integration mode
    integrator.reset_when_sum_exceeds =
        utils.xlate_meta_boolean(integrator.stanza.reset_when_sum_exceeds)
    if integrator.reset_when_sum_exceeds then
        integrator.sum_maximum_value = integrator.stanza.sum_maximum_value
        integrator.sum_maximum_value = tonumber(integrator.sum_maximum_value)
        assert(integrator.sum_maximum_value and
            integrator.sum_maximum_value >= 0,
        "INVALID integrator \"sum maximum value\": " ..
        tostring(integrator.stanza.sum_maximum_value))
    else  -- other modes allowed only when not in
          --     "reset when sum exceeds" mode
          -- optional trigger channel 
        local trig_channel = integrator.stanza.trigger_channel
        if trig_channel ~= nil then
            local trig_node = node_list[trig_channel]
            assert(trig_node, "Node for Trigger Channel not found")
            integrator.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
            assert(integrator.trig_pipe and integrator.trig_pipe ~= nil,
                "failed to create trigger pipe")
            integrator.enable_triggered_reset =
                utils.xlate_meta_boolean(
                    integrator.stanza.enable_triggered_reset)
            if integrator.enable_triggered_reset then
                -- get reset mode
                integrator.reset_mode =
                    integrator.stanza.triggered_reset_mode
                assert(integrator.reset_mode == 'when_true' or
                    integrator.reset_mode == 'on_false_to_true_edge' or
                    integrator.reset_mode == 'after_true_to_false_edge',
                    string.format("Invalid triggered reset mode: %s",
                    tostring(integrator.stanza.triggered_reset_mode)))
            else
                assert(not integrator.reset_mode,
                    "triggered reset mode not allowed unless Enable " ..
                    "Triggered Reset is true")
            end
            integrator.integrate_only_when_true =
                utils.xlate_meta_boolean(
                    integrator.stanza.integrate_only_when_true)
            -- The following restriction is required by issue #2439,
            -- if not by the "SomatXR Test Setup File Format"
            -- document.
            assert(not(integrator.integrate_only_when_true and
                       integrator.enable_triggered_reset) or
                   integrator.reset_mode == 'after_true_to_false_edge',
                   "only After True-False Edge reset mode " ..
                   "allowed when both Integrate Only When True " ..
                   "and Enable Triggered Reset selected")
        else
            assert(not integrator.stanza.enable_triggered_reset,
                "triggered reset not allowed without a " ..
                "trigger channel")
            assert(not integrator.stanza.integrate_only_when_true,
                "Integrate Only When True not allowed without a " ..
                "trigger channel")
        end
    end
    
    -- create input/output pipe structures
    integrator.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(integrator.in_pipe and integrator.in_pipe ~= nil,
       "failed to create input pipe")
    integrator.pipe_list = Pipe_new_pipe_list()
    assert(integrator.pipe_list and integrator.pipe_list ~= nil,
        "failed to create (output) pipe list")


    -- stores most recently output sample value
    integrator.out_value = integrator.initial_value
    -- current triggered state
    integrator.triggered = false

    -- used for frames' time stamps (just a sample count for
    -- this module)
    integrator.tot_count = 0

    return integrator, 1
end

return m
