require 'strict'

-- peak_valley.lua - Implements the Peak Valley data mode.
--
local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local log = require 'log'
local pv = require 'test_engine/modules/common/pv_algo'
local notify = require 'sys_notify'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = info.name .. '@' .. self.name
        info.stanza.somat_input_channel = info.stanza.name
        -- the following line makes the schema info reflect the
        -- input channel's data bit size (?should this be different?)
        self.stanza.nbr_data_bits = 
            tonumber(info.data_type:match('^.-(%d+)$'))
        assert(self.stanza.nbr_data_bits and self.stanza.nbr_data_bits > 0,
            "Missing data bit size")
        local schema_name = sie_schema.get_dm_schema_name(self.stanza.category)
        local schema_info = sie_schema.get_schema_info(schema_name, self.stanza)
        schema_info.somat_datamode_type = 'peak_valley'
        schema_info.dims[1]['core:label'] = "Peak Valley Count"
        schema_info.dims[1]['core:units'] = nil 
        schema_info.dims[2]['core:label'] = info.stanza.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, 8192,
            ffi_sizeof(info.data_type), 8, p)
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Processes incoming frames, putting outgoing frames in
-- their associated output pipe.
function mt:run()
    if self.trigger_module then
        -- trigger node feeds this node's input channels
        self.trigger_module:run()
    end

    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- Process frame for each input channel, setting input frame
        -- pointers, creating output frames as needed (with output
        -- frame pointers).
        for i = 1, #self.chan_info do
            local ch_info = self.chan_info[i]
            local infr = ch_info.in_frame
            if not ch_info.sample_type then
                ch_info.element_size = infr.element_size
                ch_info.sample_type =
                    utils.frame_data_type(ch_info.element_size)
            end
            ch_info.invals = ffi_cast(ch_info.sample_type, infr.buf)

            ch_info.pv_array = {} -- accumulates pv info
            -- find pv's (peaks and valleys)
            pv.run_pv_algo(count, ch_info,
                           self.hysteresis, ch_info.pv_array)
            if #ch_info.pv_array > 0 then
                local outfr = ch_info.out_frame
                if outfr == nil then
                    outfr = Frame_create(infr.maxsize,
                        ch_info.element_size)
                    ch_info.out_frame = outfr
                    outfr.hdr.ts = self.tot_count
                    ch_info.outvals =
                        ffi_cast(ch_info.sample_type, outfr.buf)
                    ch_info.out_index = 0
                end
                -- fill output frame
                local out_vals = ch_info.outvals
                local oi = ch_info.out_index
                for pi = 1, #ch_info.pv_array do
                    out_vals[oi] = ch_info.pv_array[pi]
                    oi = oi + 1
                end
                ch_info.out_index = oi
                outfr.count = oi
                -- put output frame in output pipe(s)
                if not utils.put_frame(
                        outfr, ch_info.pipe_list) then
                    Frame_delete(outfr)
                end
                ch_info.out_frame = nil
            end
            ch_info.last_input_value =
                ch_info.invals[count - 1]
        end

        self.tot_count = self.tot_count + count

        -- don't need the processed input frames any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

-- Finishes processing of outgoing frames at end of test.
-- NOTE:  Any last data to be written to the SIE file must
--        be written here, since no module's run() function
--        will be called again!
-- Parameter:
--     'sie_writer'  SIE writer
function mt:finalize(sie_writer)
    -- return immediately if there isn't enough information
    -- to process a "last value"
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        if not info.last_input_value or
           not info.element_size then
            return
        end
    end

    local framed_last_val = false

    for i = 1, #self.chan_info do
        local last_val
        local info = self.chan_info[i]
        if info.state == "initial" then
            last_val = info.last_input_value
        elseif info.state == "valley_search" then
            if info.valley == info.last_input_value or
               info.last_input_value - info.valley < self.hysteresis then
                last_val = info.valley
            end
        else  -- info.state == "peak_search"
            if info.peak == info.last_input_value or
               info.peak - info.last_input_value < self.hysteresis then
                last_val = info.peak
            end
        end
        -- output any last peak or valley value
        if last_val ~= nil then
            framed_last_val = true
            local outfr = Frame_create(1, info.element_size)
            local outvals =
                ffi_cast(utils.frame_data_type(info.element_size),
                         outfr.buf)
            outfr.hdr.ts = self.tot_count
            outvals[0] = last_val
            outfr.count = 1
            if not utils.put_frame(outfr, info.pipe_list) then
                Frame_delete(outfr)
            end
        end
    end
    if framed_last_val then
        -- we must ensure frames get written to file, since
        -- sie_writer's run() function is not called again
        sie_writer:flush_these(self.sie_collect_idx,
            self.nbr_input_channels, true)
        self.tot_count = self.tot_count + 1
    end
end

local m = {}

-- Creates a new instantiation of this datamode type.
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local dm_obj = setmetatable({}, {__index = mt})
    dm_obj.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    dm_obj.sie_collect_idx = first_sie_index
    dm_obj.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Peak Valley data mode " ..
        "must have at least input channel")
    --     node name
    dm_obj.name = stanza.name
    assert(dm_obj.name, "Missing node name")
    --     sample rate
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        if not dm_obj.sample_rate then
            dm_obj.sample_rate = node_meta.sample_rate
        else
            assert(dm_obj.sample_rate == node_meta.sample_rate,
                tostring(input_channels[i]) .. " sample rate " ..
                "does not match")
        end
    end
    dm_obj.sample_rate = rational:new(dm_obj.sample_rate):tonumber()
    assert(dm_obj.sample_rate and dm_obj.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(dm_obj.sample_rate)))
    dm_obj.hysteresis = rational:new(stanza.pv_hysteresis):tonumber()
    assert(type(dm_obj.hysteresis) == 'number', "Invalid pv_hysteresis")
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    --     data_type
    --     (create input channel info objects to hold data_type
    --      and other channel-specific values)
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        dm_obj.chan_info[i] =
            {name = input_channels[i],
             stanza = node_meta,
             data_type = node_meta.data_type,
             element_size = nil,
             sample_type = nil,
             state = "initial",
             valley = math.huge,
             peak = -(math.huge - 1),
             last_input_value = nil}
    end
    -- create input pipe structures
    if trigger_mode == "always_on" then
        --     create non-trigger input pipe structures
        local input_node
        local node_meta
        for i = 1, #input_channels do
            input_node = node_list[input_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            dm_obj.chan_info[i].in_pipe =
                 Pipe_attach_pipe(input_node.pipe_list)
            assert(dm_obj.chan_info[i].in_pipe and
                dm_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            dm_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(dm_obj.chan_info[i].pipe_list and
                dm_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
    else
        --     create trigger module instance
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            local sn = notify:new(__'Test Error')
            local msg = string.format(__'Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name)
            sn:error(msg)
            log:log(log.ERR, msg)
        end
        local trigger_module =
            mod:new(input_channels, trig_chan, node_list, meta_list)
        if not trigger_module then
            local sn = notify:new(__'Test Error')
            local msg = string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name)
            sn:error(msg)
            log:log(log.ERR, msg)
        end
        assert(#trigger_module.chan_info ==
               #input_channels, "Number of trigger output pipes " ..
               "does not equal number of input channels")
        --     create input pipe structures using
        --     trigger module output pipes as this instance's
        --     input pipes
        for i = 1, #input_channels do
            dm_obj.chan_info[i].in_pipe =
                Pipe_attach_pipe(trigger_module.chan_info[i].pipe_list)
            assert(dm_obj.chan_info[i].in_pipe and
                dm_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            dm_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(dm_obj.chan_info[i].pipe_list and
                dm_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        dm_obj.trigger_module = trigger_module
    end
 
    -- sample count for this module
    dm_obj.tot_count = 0
    -- used by finalize()
    dm_obj.nbr_input_channels = #input_channels

    return dm_obj, dm_obj.nbr_input_channels -- 1 sie buf per input chan
end

return m
