require 'strict'

-- rainflow.lua - Implements the Rainflow data mode.
--
-- This module comprises three main submodules.  The first processes a
-- channel's input frames to produce Peak Valley data.  The results are
-- given to an algorithm which uses it to produce Rainflow data.  Those
-- results are given to the last submodule to produce a histogram.  All
-- this is coordinated by the main module.

local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C 
local Pipe = require 'test_engine/pipe'
local Frame = require 'test_engine/frame'
local hg = require 'test_engine/modules/common/histogram'
local pv = require 'test_engine/modules/common/pv_algo'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local Pipe_attach_pipe = Pipe.attach_pipe
local Pipe_get_frame = Pipe.get_frame
local Pipe_has_frames = Pipe.has_frames
local Frame_delete = Frame.delete

local math_floor = math.floor
local math_min = math.min
local math_max = math.max

local ctor_cache = {}

-- mh stores histogram objects information
local mh = {}
-- mr stores rainflow algorithm information
local mr = {}
-- mpv stores peak valley objects information
local mpv = {}

local config = require 'system_config'
local libte = ffi.load(config.test_engine_library)
local libteFlushFrames = libte.SIEBuffer_flush_frames
local libteFlushThese = libte.SIEWriter_flush_these
local libteFlush = libte.SIEBuffer_flush
local libteAmountWritten = libte.SIEWriter_amount_written
local libteAppend = libte.SIEBuffer_append
local libteEnq = libte.SIEBuffer_enq


--=== Histograms Object's definition

-- local "constants"
local HIST_TYPE_RANGE_ONLY = 1
local HIST_TYPE_RANGE_MEAN = 2
local HIST_TYPE_TO_FROM    = 3

-- function mh:run_evenly_divided_sequential(in_buf, count)
-- function mh:run_evenly_divided_message(in_buf, count)

-- Computes ranges and histograms the results.
-- Parameters:
--   hist       -  individual histogram object
--   rain_data  -  array containing rainflow data
function mh:run_range_only(hist, rain_data)
    local bin
    local data = {}
    local dlen = 0

    for i = 1, #rain_data, 2 do
        data[dlen] = math.abs(rain_data[i+1] - rain_data[i])
        dlen = dlen + 1
    end

    if dlen > 0 then
        hg.histogram(hist, data, dlen)
    end
end

-- Computes range/mean pairs and histograms the results.
-- Parameters:
--   hist    -  individual histogram object
--   rain_data  -  array containing rainflow data
function mh:run_range_mean(hist, rain_data)
    local bin
    local data = {}  -- contains range-mean data in sequential pairs

    for i = 1, #rain_data, 2 do
        data[i - 1] = math.abs(rain_data[i+1] - rain_data[i])
        data[i] = (rain_data[i+1] + rain_data[i]) / 2
    end

    if #data > 0 then
        hg.histogram(hist, data, #data)
    end
end

-- Computes to/from pairs and histograms the results.
-- Parameters:
--   hist    -  individual histogram object
--   rain_data  -  array containing rainflow data
function mh:run_to_from(hist, rain_data)
    local bin
    local data = {}  -- contains range-mean data in sequential pairs

    for i = 1, #rain_data do
        data[i - 1] = rain_data[i]
    end

    if #data > 0 then
        hg.histogram(hist, data, #data)
    end
end

-- Places the given arrays of rainflow data in the appropriate
-- bins of their associated histograms.
-- Parameters:
--   rain_arrays  -  array of arrays containing rainflow data,
--                     one array per histogram
function mh:run(rain_arrays)
    for i = 0, self.num_hist - 1 do
        local hist = self.hist[i]
        if hist.histogram_code == HIST_TYPE_RANGE_ONLY then
            self:run_range_only(hist, rain_arrays[i + 1])
        elseif hist.histogram_code == HIST_TYPE_RANGE_MEAN  then
            self:run_range_mean(hist, rain_arrays[i + 1])
        elseif hist.histogram_code == HIST_TYPE_TO_FROM then
            self:run_to_from(hist, rain_arrays[i + 1])
        end
    end
end

-- At the end of the test, writes to the SIE file the bins data
-- for the histograms.
-- Parameter:  sie_writer  -  SIE file writer object
function mh:finalize(sie_writer)
    local num = self.num_hist - 1
    local bin_bits = 64
    local hdr = ffi.new('uint32_t[3]')
    for i = 0, num do
        local hist = self.hist[i]
        local buf = sie_writer.bufs[hist.sie_idx]
        hdr[0] = hist.nbr_dims 
        hdr[1] = hist.total_dimension_size
        if hist.nbr_dims > 1 then
            hdr[2] = hist.total_dimension_size
            libteEnq(buf, hdr, ffi_sizeof('uint32_t') * 3)
        else
            libteEnq(buf, hdr, ffi_sizeof('uint32_t') * 2)
        end
        libteEnq(buf, hist.binmatrix, hist.matrix_size * (bin_bits / 8))
        libteFlush(buf, 1)
    end
end

-- Writes a histogram's bounds information to the SIE file using
-- sie_writer's "scratch" sie buf.
-- Parameter:
--   sw    -  SIE file writer object
--   hist  -  array of histogram tables
function mh:write_bounds(sw, dim_size, bounds)
        local bounds_bits = 64  -- all bounds use 64 bits (for now!)
        sw:new_buffer(sw.scratch_buf_idx, nil,
            (dim_size + 1) * bounds_bits / 8, bounds_bits / 8, 0)
        local bbuf = sw.bufs[sw.scratch_buf_idx]
        local ptr = ffi.cast('const char *', bounds)
        libteAppend(bbuf, ptr, (dim_size + 1) * bounds_bits / 8)
        libteFlush(bbuf, 0)
    return bbuf.group
end

-- Create all of the SIE buffers needed by this sink
function mh:setup()
end

-- provides definition of a histogram table
local function meta_ctor(t)
    if not ctor_cache[t] then
        local hdm_type = string.format('struct %s_rflow_hdm', t)
        local spec = string.gsub([[
            struct rflow_dimension {
                double          lo_bound;
                double          hi_bound;
                double          base;
                double          bin_size;
                int             dimension_size;
                int             total_dimension_size;
                double          *bounds;
            };
            struct $t_rflow_hist {
                int             nbr_dims;
                int             dimension_size;
                int             total_dimension_size;
                int             histogram_code;
                int             matrix_size;
                struct rflow_dimension *dims;
                $t              *binmatrix;
                /* For evenly_divided special case */
                double          prev_val;
                int             sie_idx;
                char            first_time;
                char            *name;
                char            *description;
                char            *hist_name;
                char            *category;
                char            *units;
                double          sample_rate;
                //struct pipe     *in_pipe;
            };
            struct $t_rflow_hdm {
                int             num_hist;
                int             final_called;
                struct $t_rflow_hist  *hist;
            };
        ]], '%$(%w+)', {t = t})
        ffi.cdef(spec)
        ctor_cache[t] = ffi.metatype(hdm_type, {__index = mh})
    end
    return ctor_cache[t]
end

--=== Peak Valley object's definition

function mpv:setup()
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mpv:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Computes the peaks and valleys for the input channels, one
-- full set of incoming frames at a time.
function mpv:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return nil
    end

    -- initialize set of arrays which will accumulate pv info
    local pv_arrays = {}
    for i = 1, #self.chan_info do
        pv_arrays[i] = {}
    end


    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- Process frame for each input channel, setting input frame
        -- pointers, creating output frames as needed (with output
        -- frame pointers).
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            if not self.chan_info[i].sample_type then
                self.chan_info[i].sample_type =
                    utils.frame_data_type(infr.element_size)
            end
            self.chan_info[i].invals = ffi_cast(self.chan_info[i].sample_type, infr.buf)

            -- find pv's (peaks and valleys)
            pv.run_pv_algo(count, self.chan_info[i], self.hysteresis, pv_arrays[i])
            self.chan_info[i].last_input_value =
                self.chan_info[i].invals[count - 1]
        end

        -- don't need the processed input frames any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames(self.chan_info)
    end

    return pv_arrays
end

-- Finishes processing of outgoing frames at end of test.
-- Parameter:
--     'sie_writer'  SIE writer
function mpv:finalize()
    local pv_arrays = {}
    -- return immediately if there isn't enough information
    -- to process a "last value"
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        if not info.last_input_value then
            return
        end
    end

    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        pv_arrays[i] = {} -- accumulates pv info
        if info.state == "initial" then
            pv_arrays[i][1] = info.last_input_value
        elseif info.state == "valley_search" then
            if info.valley == info.last_input_value or
               info.last_input_value - info.valley < self.hysteresis then
                pv_arrays[i][1] = info.valley
            end
        else  -- info.state == "peak_search"
            if info.peak == info.last_input_value or
               info.peak - info.last_input_value < self.hysteresis then
                pv_arrays[i][1] = info.peak
            end
        end
    end

    return pv_arrays
end

--=== Rainflow algorithm object's definition

function mr:setup()
    -- nada for now
end

-- Converts incoming peak-valley data into rainflow data.
-- Parameter:  pv_arrays - array of peak-valley data arrays, one array
--     per channel
-- Returns:  array of rainflow data arrays, one array per input channel
function mr:run(pv_arrays)
    local pva
    local stack
    local stack_len
    local ringup
    local prev1
    local prev2
    local nextpoint
    local rain_arrays = {}
    local ra

    for ch_nbr = 1, #pv_arrays do
        ra = {}
        rain_arrays[ch_nbr] = ra
        pva = pv_arrays[ch_nbr]
        stack = self.stacks[ch_nbr]
        stack_len = self.stack_lens[ch_nbr]
        ringup = self.ringups[ch_nbr]
        for i = 1, #pva do
            nextpoint = pva[i]
            stack_len = stack_len + 1
            stack[stack_len] = nextpoint

            while (stack_len - ringup >= 4) do
                -- load the previous point
                prev1 = stack[stack_len - 1]
                -- load the previous-previous point
                prev2 = stack[stack_len - 2]

                -- compare ranges
                if math.abs(nextpoint - prev1) >= math.abs(prev1 - prev2) then
                    -- load the points for histogramming
                    ra[#ra + 1] = prev2
                    ra[#ra + 1] = prev1
                    -- collapse the stack
                    stack[stack_len - 2] = nextpoint
                    stack_len = stack_len - 2
                else
                    break
                end
            end

            -- if exactly 3 points lie in the stack after the ringup,
            -- check for an additional ringup.
            if stack_len - ringup == 3 then
                -- load the previous point
                prev1 = stack[stack_len - 1]
                -- load the previous-previous point
                prev2 = stack[stack_len - 2]
                -- if a ringup occurred, move the current ringup pointer,
                -- which always points to the extreme Peak or Valley
                --  stack value.
                if math.abs(nextpoint - prev1) > math.abs(prev1 - prev2) then
                    ringup = ringup + 1
                end
            end
        end
        self.stack_lens[ch_nbr] = stack_len
        self.ringups[ch_nbr] = ringup
    end

    return rain_arrays
end

-- Called at the end of a test.  Closes the cycles remaining on
-- the rainflow stack.
function mr:close_cycles()
    local rain_arrays = {}
    local ra

    -- one loop iteration for each input channel
    -- (i.e., for each histogram)
    for ch_nbr = 1, #self.stacks do
        local restack = {}
        local histogram_code = self.histogram_code[ch_nbr]
        local stack = self.stacks[ch_nbr]
        local stack_len = self.stack_lens[ch_nbr]
        local start = 1

        ra = {}
        rain_arrays[ch_nbr] = ra

        if stack_len > 0 then
            -- Find start point
            local max = stack[1]
            for i = 2, stack_len do
                if max < stack[i] then
                    max = stack[i]
                    start = i
                end
            end

            -- Create a new stack that is sized to stack_len + 1
            -- and is reordered to start at “start” computed above.
            -- The last point in the new stack is set equal to the first
            -- point in the new stack.
            local tmplen = 0
            for i = start, stack_len do
                tmplen = tmplen + 1
                restack[tmplen] = stack[i]
            end
            for i = 1, start - 1 do
                tmplen = tmplen + 1
                restack[tmplen] = stack[i]
            end
            stack_len = tmplen + 1
            restack[stack_len] = restack[1]

            -- Make sure the restack sequence is a PV sequence.
            -- NOTE: This is required because the extra point added
            -- to restack (i.e., the max value) was not processed
            -- through the PV picker algorithm. About 50% of the time,
            -- the point before this last one will be discarded!   
            local peak = 1
            local istart = 1
            for i = 1, stack_len - 1 do
                if peak * (restack[i] - restack[i + 1]) > 0 then
                    restack[istart] = restack[i]
                    istart = istart + 1
                    peak = -peak
                else
                    restack[istart] = restack[i + 1]
                end
            end
            restack[istart] = restack[stack_len]
            stack_len = istart

            -- Run a special rainflow counting algorithm on
            -- the closed cycles to generate the final data
            -- to be returned for histogramming
            local n = 1
            local this_cycle
            local last_cycle
            for i = 1, stack_len do
                restack[n] = restack[i]
                while n > 2 do
                    this_cycle = math.abs(restack[n] - restack[n - 1])
                    last_cycle = math.abs(restack[n - 1] - restack[n - 2])
                    if this_cycle < last_cycle then break end

                    -- store last_cycle points for later histogramming
                    ra[#ra + 1] = restack[n - 2]
                    ra[#ra + 1] = restack[n - 1]

                    -- reduce the stack
                    n = n - 2
                    restack[n] = restack[n + 2]
                end
                n = n + 1
            end
        end
    end

    return rain_arrays
end

-- Closes the remaining rainflow stack, and writes unclosed rainflow
-- cycles to the SIE file.  Content is written for each channel
-- consisting of "unclosed cycles" data converted to ASCII formatted
-- floating point and embedded in XML.  The SIE format causes this XML
-- to be merged into the channel's other XML.  This function then
-- returns closed rainflow data for histogramming.
-- Parameters:
--    'sie_writer'  SIE writer
-- Returns:  array of closed cycle rainflow data arrays, one array
--    per input channel
function mr:finalize(sie_writer)
    local rain_arrays = self:close_cycles()
    for ch_nbr = 1, #self.stacks do
        sie_utils.write_unclosed_cycles(
             sie_writer, self.unclosed_tab[ch_nbr],
             self.stacks[ch_nbr], self.stack_lens[ch_nbr])
    end
    return rain_arrays
end

--=== Rainflow data mode object's definition

require 'middleclass'

local m = class('rainflow')

-- Creates the SIE buffers (pipes and XML) required.
-- There will be multiple channel descriptions per input channel,
-- and a generic histogram lookup table description if it doesn't
-- already exist.
function m:sie_xml(sie_writer, dm_tab)
    for i = 0, self.num_hist - 1 do
        local hist = self.hdm.hist[i]

        -- get channel id for the appropriate generic histogram
        -- lookup table channel, writing it if necessary
        local bin_tab = {}
        bin_tab._base_id =
            sie_utils.ensure_sie_bounds_channel(sie_writer, 64)
        bin_tab._dm = ffi.string(hist.hist_name)
        bin_tab._name = ffi.string(hist.name)
        -- write "bin" min/max values and the histogram "bin"
        -- channels which will use them
        local index_chan_id = {}
        -- for dimension 0
        --     write bounds information to the SIE file
        -- [the following call prevents the bounds from being written
        --  twice - both before and after the dim bounds channel
        --  DON'T KNOW WHY, so FIXME]
        sie_writer:xml()
        bin_tab._sie_group =
            self.hdm:write_bounds(sie_writer, hist.total_dimension_size,
                                  hist.dims[0].bounds)
        --     write the dim 0 channel which will use the bounds info
        bin_tab._histdim = 0
        index_chan_id[0] =
            sie_utils.write_bin_channel(sie_writer, bin_tab)
        if hist.nbr_dims > 1 then
            -- for dimension 1
            --     write bounds information to the SIE file
            bin_tab._sie_group =
                self.hdm:write_bounds(sie_writer, hist.total_dimension_size,
                                      hist.dims[1].bounds)
            --     write the dim 1 channel which will use the bounds info
            bin_tab._histdim = 1
            index_chan_id[1] =
                sie_utils.write_bin_channel(sie_writer, bin_tab)
        end
        -- write XML for the histogram's SIE channel
        local sch_tab = {}
        sch_tab.description = ffi.string(hist.description) or ""
        sch_tab.datamode_name = ffi.string(hist.hist_name)
        sch_tab.sie_name = ffi.string(hist.name) .. '@' ..
            sch_tab.datamode_name
        sch_tab.somat_input_channel = ffi.string(hist.name)
        sch_tab.category = ffi.string(hist.category)
        sch_tab.data_type = dm_tab.data_type
        sch_tab.datamode_type = dm_tab.datamode_type
        sch_tab.sample_rate = hist.sample_rate
        sch_tab.nbr_data_bits = 64
        sch_tab.name = ffi_string(hist.name)
        sch_tab.units = ffi_string(hist.units)
        sch_tab.dim1_name =
            (self.histogram_code == HIST_TYPE_RANGE_ONLY or
             self.histogram_code == HIST_TYPE_RANGE_MEAN) and
            ("Range " .. sch_tab.name) or ("From " .. sch_tab.name)
        sch_tab.dim2_name =
            (self.histogram_code == HIST_TYPE_RANGE_MEAN and
             ("Mean " .. sch_tab.name)) or
            (self.histogram_code == HIST_TYPE_TO_FROM and
             ("To " .. sch_tab.name)) or "Cycles"
        sch_tab.nbr_dims = hist.nbr_dims
        sch_tab.low_bound = {}
        sch_tab.high_bound = {}
        sch_tab.low_bound[0] = hist.dims[0].lo_bound
        sch_tab.high_bound[0] = hist.dims[0].hi_bound
        if hist.nbr_dims > 1 then
            sch_tab.low_bound[1] = hist.dims[1].lo_bound
            sch_tab.high_bound[1] = hist.dims[1].hi_bound
        end
        local schema_name = sie_schema.get_dm_schema_name(sch_tab.category)
        local schema_info = sie_schema.get_schema_info(schema_name, sch_tab)
        -- assign index channel ids and dim labels
        schema_info.dims[2]._index_ch = index_chan_id[0]
        schema_info.dims[3]._index_ch = index_chan_id[0]
        if hist.nbr_dims > 1 then
            schema_info.dims[4]._index_ch = index_chan_id[1]
            schema_info.dims[5]._index_ch = index_chan_id[1]
        end
        local st_tab = {}
        st_tab.sie_name = sch_tab.sie_name
        st_tab.category = sch_tab.category
        st_tab.data_type = dm_tab.data_type
        st_tab.datamode_name = sch_tab.datamode_name
        st_tab.connector = ""
        st_tab.somat_input_channel = sch_tab.somat_input_channel
        st_tab.description = sch_tab.description
        st_tab.sample_rate = sch_tab.sample_rate
        local md = sie_utils.write_sie_channel(sie_writer, schema_info,
            st_tab)
        -- Create the SIE buffer and attach the pipe to it
        local bin_bits = 64
        sie_writer:new_buffer(hist.sie_idx, md._sie_group,
            hist.matrix_size * bin_bits / 8, bin_bits / 8,
            ffi_sizeof('uint64_t') * (hist.nbr_dims + 1))
        -- Give rainflow object the channel's id (for use when
        -- it writes the unclosed cycles data to the channel
        -- at the end of the test).
        self.rain_obj.unclosed_tab[i + 1] =
            {_test_id = self.test_id, _ch_id = md._sie_ch}

        -- Ensure there is file space to write the channel data
        -- at end of test.
        sie_writer:reserve_space(
            ((hist.matrix_size + 2) * (bin_bits / 8)) +   -- bin space
            (ffi_sizeof('uint32_t') * 2) + -- header space
            (ffi_sizeof('uint32_t') * 5), -- SIE overheader
            function() self:finalize(sie_writer) end)
    end
end

function m:setup(sie_writer, is_restart)
    self:sie_xml(sie_writer, self.dm_tab)
    self.hdm:setup()
    self.rain_obj:setup()
    self.pv_obj:setup()
end

function m:run()
    if self.trigger_module then self.trigger_module:run() end
    -- pv_obj processes input frames
    local pv_arrays = self.pv_obj:run()
    if pv_arrays then
        local rain_arrays = self.rain_obj:run(pv_arrays)
        self.hdm:run(rain_arrays)
    end
end

function m:finalize(sie_writer)
    local rain_arrays
    if self.final_called ~= 0 then return end
    self.final_called = 1
    -- process final input values
    if self.trigger_module then self.trigger_module:run() end
    local pv_arrays = self.pv_obj:finalize()
    if pv_arrays then
        rain_arrays = self.rain_obj:run(pv_arrays)
        self.hdm:run(rain_arrays)
    end
    -- get final rainflow data (from cycle closing algorithm)
    -- and histogram it
    rain_arrays = self.rain_obj:finalize(sie_writer)
    self.hdm:run(rain_arrays)
    -- write histogram
    self.hdm:finalize(sie_writer)
end

function m:initialize(stanza, node_list, sie_idx, stanza_list)
    local bin_type = stanza.bin_type or 'uint64_t'
    local ctor = meta_ctor(bin_type)

    --== initialize module-level items
    self.test_id = 0
    self.final_called = 0
    local input_channels = stanza.input_channels
    self.sie_buffers_needed = #input_channels
    self.nbr_input_channels = #input_channels
    -- If triggered then instantiate a new trigger module
    self.trigger_module = stanza.trigger_mode ~= 'always_on' and
        require('test_engine/modules/trigger/' .. stanza.trigger_mode):new(
            stanza.input_channels, stanza.trigger_channel, node_list, stanza_list)

    --== initialize peak-valley support

    self.pv_obj = setmetatable({}, {__index = mpv})
    local hysteresis = rational:new(stanza.pv_hysteresis):tonumber()
    assert(type(hysteresis) == 'number', "Invalid pv_hysteresis")
    self.pv_obj.chan_info = {}
    for i = 1, #input_channels do
        local ic = input_channels[i]
        local ic_meta = stanza_list[ic]
        local ic_node = node_list[ic]
        assert(ic_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        assert(ic_meta.data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        self.pv_obj.hysteresis = hysteresis
        self.pv_obj.chan_info[i] =
            {name = ic,
             stanza = ic_meta,
             data_type = ic_meta.data_type,
             sample_rate = rational(ic_meta.sample_rate):tonumber(),
             sample_type = nil,
             state = "initial",
             nbr_input_channels = self.nbr_input_channels,
             valley = math.huge,
             peak = -(math.huge - 1),
             last_input_value = nil}

        if not self.trigger_module then
            -- Not triggered...attach directly to input channel
            self.pv_obj.chan_info[i].in_pipe = Pipe_attach_pipe(ic_node.pipe_list)
        else
            -- Triggered...attach to gated channel
            self.pv_obj.chan_info[i].in_pipe =
                Pipe_attach_pipe(self.trigger_module.chan_info[i].pipe_list)
        end
    end

    --== initialize histogram support

    self.num_hist = #input_channels
    self.dm_tab = {}
    self.dm_tab.data_type = ""  -- data_type irrelevant for histogram
    self.histogram_type = stanza.histogram_type
    assert(self.histogram_type == "histogram_range_only" or
           self.histogram_type == "histogram_range_mean" or
           self.histogram_type == "histogram_to_from",
        "Invalid histogram type")
    local nbr_dims
    if self.histogram_type == "histogram_range_only" then
        self.dm_tab.datamode_type = "rainflow/range_only"
        nbr_dims = 1
        self.histogram_code = HIST_TYPE_RANGE_ONLY
    elseif self.histogram_type == "histogram_range_mean" then
        self.dm_tab.datamode_type = "rainflow/range_mean"
        nbr_dims = 2
        self.histogram_code = HIST_TYPE_RANGE_MEAN
    else
        self.dm_tab.datamode_type = "rainflow/from_to"
        nbr_dims = 2
        self.histogram_code = HIST_TYPE_TO_FROM
    end

    self.hdm = ctor()  -- named "hdm" for "histogram data mode"
    -- all dimensions have the same number of bins (add 2 extra bins
    -- for underflow and overflow)
    self.hdm.num_hist = self.num_hist
    self.hdm.hist = ffi_cast(string.format('struct %s_rflow_hist *', bin_type),
        C.calloc(self.hdm.num_hist,
        ffi_sizeof(string.format('struct %s_rflow_hist', bin_type))))
    assert(self.hdm.hist ~= 0, "Allocation failed: self.hdm.hist")
    local objs = self.hdm.hist

    for i = 0, self.num_hist - 1 do
        local hb = stanza.histogram_bins[i + 1]
        local obj = objs[i]
        local ic = stanza.input_channels[i + 1]
        local chan_meta = stanza_list[ic]

        assert(hb.type == "evenly_divided",
                "Only 'evenly_divided' is supported")
        obj.first_time = 1
        obj.prev_val = 0
        obj.sie_idx = sie_idx + i
        obj.sample_rate = rational(chan_meta.sample_rate):tonumber()
        obj.nbr_dims = nbr_dims
        local units = chan_meta.output_units or ''
        obj.units = ffi_cast('char *', C.calloc(units:len() + 1,
                ffi_sizeof('char')))
        assert(obj.units ~= 0, "Allocation failed: obj.units")
        ffi_copy(obj.units, units)
        -- allocate space for histogram's bin matrix 
        obj.dimension_size = stanza.number_of_bins
        obj.total_dimension_size = obj.dimension_size + 2
        obj.matrix_size = obj.total_dimension_size ^ obj.nbr_dims
        obj.binmatrix = ffi_cast(string.format('%s *', bin_type),
                C.calloc(obj.matrix_size, ffi_sizeof(bin_type)))
        assert(obj.binmatrix ~= 0, "Allocation failed: obj.binmatrix")
        -- initialize dimensions information
        obj.dims = ffi_cast('struct rflow_dimension *',
                C.calloc(obj.nbr_dims, ffi_sizeof('struct rflow_dimension')))
        assert(obj.dims ~= 0, "Allocation failed: obj.dims")
        for j = 0, obj.nbr_dims - 1 do  -- nbr_dims == 1 or 2 for now
            obj.dims[j].dimension_size = obj.dimension_size
            obj.dims[j].total_dimension_size = obj.total_dimension_size
            obj.dims[j].bin_size =
                    (hb.high_bound - hb.low_bound) / obj.dims[j].dimension_size
            obj.dims[j].bounds = ffi_cast('double *',
                C.calloc(obj.total_dimension_size + 1, ffi_sizeof('double')))
            assert(obj.dims[j].bounds ~= 0,
                "Allocation failed: obj.dims[" .. tostring(j) .. "].bounds")
            obj.histogram_code = self.histogram_code
            -- all bins are to be evenly divided
            if j == 0 and (self.histogram_code == HIST_TYPE_RANGE_ONLY or
                           self.histogram_code == HIST_TYPE_RANGE_MEAN) then
                obj.dims[j].lo_bound = 0
                obj.dims[j].hi_bound = hb.high_bound - hb.low_bound
            else
                obj.dims[j].lo_bound = hb.low_bound
                obj.dims[j].hi_bound = hb.high_bound
            end
            obj.dims[j].base = obj.dims[j].lo_bound - obj.dims[j].bin_size
            obj.dims[j].bounds[0] = -1/0
            obj.dims[j].bounds[obj.total_dimension_size] = 1/0
            for k = 0, obj.total_dimension_size - 2 do
                obj.dims[j].bounds[k + 1] = obj.dims[j].lo_bound
                    + k * obj.dims[j].bin_size
            end
        end
        -- initialize remaining histogram information
        local desc = chan_meta.description or ic or
            'No channel description'
        obj.description = ffi_cast('char *', C.calloc(desc:len() + 1,
            ffi_sizeof('char')))
        assert(obj.description ~= 0, "Allocation failed: obj.description")
        ffi_copy(obj.description, desc)
        obj.name = ffi_cast('char *',
            C.calloc(ic:len() + 1, ffi_sizeof('char')))
        assert(obj.name ~= 0, "Allocation failed: obj.name")
        ffi_copy(obj.name, ic)
        obj.hist_name = ffi_cast('char *',
            C.calloc(stanza.name:len() + 1, ffi_sizeof('char')))
        assert(obj.hist_name ~= 0, "Allocation failed: obj.hist_name")
        ffi_copy(obj.hist_name, stanza.name)
        obj.category = ffi_cast('char *',
            C.calloc(stanza.category:len() + 1, ffi_sizeof('char')))
        assert(obj.category ~= 0, "Allocation failed: obj.category")
        ffi_copy(obj.category, stanza.category)
    end

    --== initialize rainflow support

    self.rain_obj = setmetatable({}, {__index = mr})
    self.rain_obj.stacks = {}
    self.rain_obj.stack_lens = {}
    self.rain_obj.ringups = {}
    self.rain_obj.histogram_code = {}
    self.rain_obj.unclosed_tab = {}
    for i = 1, #input_channels do
        self.rain_obj.stacks[i] = {}
        self.rain_obj.stack_lens[i] = 0
        self.rain_obj.ringups[i] = 0
        self.rain_obj.histogram_code[i] = self.histogram_code
    end

    return self, #input_channels
end

return m
