require 'strict'
local ffi = require 'ffi'

-- sie_utils.lua - Utility functions for the test engine to use in
--     setting up sie file XML.

local cjson = require 'cjson'
local rational = require 'rational'
local sample_clock = require 'sample_clock'
local stempls = require 'test_engine/sie_templates'
local sie_schema = require 'test_engine/sie_schema'

-- "constants"
local NL = '\n'
local INDENT_LEV_1 = ' '

-- Stores channel decoder ids.
local decoder_ids = {}

-- Stores info tables for defining base channels.
local ch_base_tab = {}

-- Stores info tables for defining base channels.
local histo_generics = {}

-- Builds a table with the information needed for a base channel.
-- NOTE that id fields are not set in the returned table,
--      specifically _sie_ch and _sie_decoder.
-- ALSO NOTE that table entries with keys WITHOUT a prepended '_'
--      will be written as channel tag elements!
-- Params:
--     'schema_info' contains information about the channel's schema.
-- Returns: table containing base channel info, followed by the
--     info's signature (the latter being a string built from the info)
local function base_chan_info(schema_info)
    local base_md = {}
    local sig = {}
    base_md._sie_name = "base channel"
    table.insert(sig, "_sie_name = " .. tostring(base_md._sie_name))
    base_md['somat:datamode_type'] = schema_info.somat_datamode_type
    table.insert(sig, "somat:datamode_type = " ..
        tostring(base_md['somat:datamode_type']))
    base_md['somat:data_bits'] = schema_info.bits
    table.insert(sig, "somat:data_bits = " ..
        tostring(base_md['somat:data_bits']))
    base_md['somat:data_format'] = schema_info.somat_data_format
    table.insert(sig, "somat:data_format = " ..
        tostring(base_md['somat:data_format']))
    base_md['core:description'] = ""
    table.insert(sig, "core:description = " ..
        tostring(base_md['core:description']))
    base_md['core:schema'] = schema_info.sie_core_schema
    table.insert(sig, "core:schema = " ..
        tostring(base_md['core:schema']))
    base_md['somat:version'] = "1.0"
    table.insert(sig, "somat:version = " ..
        tostring(base_md['somat:version']))

    return base_md, table.concat(sig, ';')
end

-- The information needed to write the SIE XML for a specific
-- test channel comprises a subset of the values from the channel's
-- setup stanza.  This function provides that subset.
-- NOTE that id fields are not set in the returned table,
--      specifically _sie_ch, _sie_group, and _sie_decoder.
--      However, _sie_group is set to 'true' if a group id
--      is needed, or to 'false' if a group id is not needed.
-- ALSO NOTE that table entries with keys WITHOUT a prepended '_'
--      will be written as channel tag elements!
-- Params:
--     'stanza' contains setup info for this channel
-- Returns: table containing info for this channel
local function test_chan_info(schema_info, stanza)
    local ch_md = {}
    ch_md._sie_test = stanza._sie_test or 0
    ch_md._sie_name = stanza.sie_name
    if stanza.private_chan then
        ch_md._sie_group = false
    else
        ch_md._sie_group = true
    end
    ch_md.data_type = stanza.data_type
    ch_md['somat:data_format']    = schema_info.somat_data_format
    ch_md['somat:message_format'] = stanza.message_format
    ch_md['core:description']     = stanza.description
    ch_md['somat:input_channel']  = stanza.somat_input_channel
    ch_md['core:sample_rate']     = rational:new(stanza.sample_rate):tonumbers()
    ch_md['somat:datamode_name']  = stanza.datamode_name or 'dm'
    ch_md['somat:datamode_type']  = schema_info.somat_datamode_type
    ch_md['somat:connector']      = stanza.connector
    ch_md['somat:data_bits']      = schema_info.bits
    ch_md['core:schema']          = schema_info.sie_core_schema

    -- xform scale & offset are used for all but histograms
    -- (for histograms xform index is used, which is set elsewhere)
    if not stanza.category:find('^sink/sie/time_at_level_1d') then
        if ch_md['core:sample_rate'] and ch_md['core:sample_rate'] > 0 then
            ch_md._scale = schema_info.xform_scale or
                           (1.0 / ch_md['core:sample_rate'])
            ch_md._offset = stanza.xform_offset or 0
        else
            ch_md._scale = 1.0 / sample_clock.master_rate
            ch_md._offset = 0
        end
    end

    return ch_md
end

-- Puts base channel's id in local table, using base channel's
-- signature as the key.
-- Params:
--     'sig' signature of the base channel
--     'id' base channel's id
local function ch_base_tab_put(sig, id)
    ch_base_tab[sig] = id
end

-- Finds the id of the base channel which has the specified signature.
-- Params:  'sig' is the signature of the desired base channel
-- Returns: id of the base channel with the specified signature,
--     otherwise nil
local function ch_base_tab_get(sig)
    return ch_base_tab[sig]
end

-- Module table.
local m = {}

--  Values for sie buf objects' dm_type field.
--  These indicate sie bufs which are used for and need
--  special processing for a specific data mode.
--    For synchronous datamodes which are distinguished
--    only by having discontinuous timestamps:
m.DISCON_DATAMODE = 0x40
--    For a Block History Datamode:
m.BH_DATAMODE = 0x42

-- Converts the input into a rational number.
-- (This function was copied from sie_writer.lua)
function m.get_rational(x)
    local slash = (type(x) == 'string') and x:find('/')
    if slash then
        local num = tonumber(x:sub(1, slash - 1))
        local denom = tonumber(x:sub(slash + 1))
        return num / denom, num, denom
    end
    local num = tonumber(x)
    return num, num, 1
end

-- Writes XML to define a SIE decoder's loop.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'k'  is the key value for the set of loop attributes,
--          which indicates what type of loop is specified
--     'v'  is the the set of loop attributes to be used
--     'indent'  is the level of indentation of the XML
function m.write_decoder_loop(sw, k, v, indent)
    -- write loop start
    if k:find('^bare') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_bare_loop .. NL, v.head)
    elseif k:find('^short') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_short_loop .. NL, v.head)
    elseif k:find('^long') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_long_loop .. NL, v.head)
    end
    if v.inner then
        -- write inner loop
        m.write_decoder_loop(sw, k, v.inner, indent+1)
    end
    -- write interior reads/sets/sample
    for _, r in ipairs(v.reads or {}) do
        if r.bits then
            sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_read .. NL, r)
        else
            sw:xml(INDENT_LEV_1:rep(indent+1) ..
                   stempls.dec_read_octets .. NL, r)
        end
        if r.post_seek then
            sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_seek .. NL,
                   r.post_seek)
        end
    end
    if v.sample then
        sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_sample .. NL)
    end
    for _, r in ipairs(v.sets or {}) do
        sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_set .. NL, r)
    end
    -- write loop end
    sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_loop_end .. NL)
end

-- Writes XML to define a SIE decoder.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'schema_info' contains information about the somat
--         schema with which this decoder will be associated,
--         and which is needed to write the XML
-- Returns:  id for the newly written decoder
function m.write_decoder(sw, schema_info)
    local decoder_id = sw:next_id('decoder')
    sw:xml(stempls.dec_start .. NL, {_sie_decoder = decoder_id})
    for _, v in ipairs(schema_info.nonrep_reads or {}) do
        sw:xml(INDENT_LEV_1 .. stempls.dec_read .. NL, v)
    end
    for _, v in ipairs(schema_info.nonrep_sets or {}) do
        sw:xml(INDENT_LEV_1 .. stempls.dec_set .. NL, v)
    end
    for k, v in pairs(schema_info.loops or {}) do
        m.write_decoder_loop(sw, k, v, 1)
    end
    sw:xml(stempls.dec_end .. NL)
    return decoder_id
end

-- Writes the info in the given table to the SIE file as 
-- the XML for a channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains the remaining info to be written
--     'dims' contains info about the channel's "dim" elements
--         (NOTE that this may be nil, for base channels)
function m.write_chan_xml(sw, md, dims)
    if md._priv_type then
        -- write a base channel
        sw:xml(stempls.base_ch .. NL, md)
    elseif md._sie_group then
        -- write a specific test channel
        sw:xml(stempls.test_ch .. NL, md)
    else
        -- write a specific private test channel
        sw:xml(stempls.priv_test_ch .. NL, md)
    end
    -- write <tag> elements
    -- NOTE:  Only keys WITHOUT prepended '_' will be written
    --        as channel tag elements!
    for k, v in pairs(md) do
        if k:sub(1, 1) ~= '_' then
            sw:xml(INDENT_LEV_1 .. stempls.ch_tag .. NL, {k=k, v=v})
        end
    end
    if dims then
        -- write <dim> elements
        for i, dim in ipairs(dims) do
            sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, dim)
            -- NOTE:  Only keys WITHOUT prepended '_' will be written
            --        as dim tag elements!
            for k, v in pairs(dim) do
                if k:sub(1, 1) ~= '_' then
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.ch_tag .. NL,
                        {k=k, v=v})
                end
            end
            sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL, dim, md)
            if dim._use_scale then
                if dim._use_index then
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_xindex .. NL,
                        {_index_ch = dim._index_ch,
                         _index_dim = dim._index_dim})
                else
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_scale .. NL,
                        {_scale = md._scale,
                         _offset = md._offset})
                end
            end
            -- finish writing <dim> element
            sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)
        end
    end
    -- finish writing channel element
    sw:xml(stempls.ch_end .. NL)
end

-- Writes the SIE XML for a channel's SIE base channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains information about the base channel
--         NOTE that this table's id fields will be modified
--     'decoder_type' indicates the signature of the decoder
--         to be used
function m.write_channel_base(sw, md, decoder_type)
    md._decoder_type = decoder_type
    md._priv_type = "base"
    m.write_chan_xml(sw, md)
end

-- Writes the SIE XML for a specific test channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains information about the channel
--     'schema_info' contains information about the
--         channel's schema.
function m.write_channel(sw, md, schema_info)
    m.write_chan_xml(sw, md, schema_info.dims)
end

-- Writes XML to the SIE file for a channel.  This includes
-- writing the decoder and/or base channel XML if needed.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'schema_info' contains info about a channel's SIE schema
--     'stanza' contains setup info for a channel
-- Returns:  table containing this specific channel's info
function m.write_sie_channel(sw, schema_info, stanza)
    assert(sw, "Missing sw (sie_writer)")
    assert(schema_info, "Missing SIE schema info")
    assert(stanza, "Missing channel setup info")

    -- get from the schema info (most of) the values needed to
    -- write the XML for a base channel
    local base_md
    local base_sig
    -- information for base channel
    base_md, base_sig = base_chan_info(schema_info)
    -- information for this specific channel
    local chan_md = test_chan_info(schema_info, stanza)
    -- need decoder id
    local decoder_id = decoder_ids[schema_info.decoder_type]
    if not decoder_id then
        -- decoder doesn't already exist, so write XML for it
        decoder_id = m.write_decoder(sw, schema_info)
        decoder_ids[schema_info.decoder_type] = decoder_id
    end
--~     -- get base channel id
--~     local base_id = ch_base_tab_get(base_sig)
--~     if not base_id then
--~         base_id = sw:next_id('ch')
--~         -- base channel doesn't already exist, so write XML for it
--~         ch_base_tab_put(base_sig, base_id)
--~         -- set ids for the base channel
--~         base_md._sie_decoder = decoder_id
--~         base_md._sie_ch = base_id
--~         m.write_channel_base(sw, base_md, schema_info.decoder_type)
--~     end
    chan_md._sie_decoder = decoder_id
--~     chan_md._base_id = base_id
    chan_md._sie_ch = sw:next_id('ch')
    if chan_md._sie_group then
        chan_md._sie_group = sw:next_id('group')
    end
    m.write_channel(sw, chan_md, schema_info)
    return chan_md
end

-- Checks for the existence of the min/max lookup table base channel
-- needed for a histogram.  Writes it if it doesn't exist (includes
-- writing decoder XML).
-- Parameters:
--     'sw' is the sie_writer object to use
--     'bounds_bits' size in bits of bounds min/max datatype
-- Returns:  table containing this specific channel's info
function m.ensure_sie_bounds_channel(sw, bounds_bits)
    local bounds_type = '_histogram_bin_bounds' .. bounds_bits
    local ch_id = histo_generics[bounds_type]
    if not ch_id then
        -- write decoder
        local decoder_info = sie_schema.get_hist_decoder_schema(bounds_bits)
        local decoder_id = m.write_decoder(sw, decoder_info)

        -- write channel (start XML on new line)
        ch_id = sw:next_id('ch')
        sw:xml(NL .. stempls.generic_bin_ch .. NL, {_sie_ch = ch_id})
        sw:xml(INDENT_LEV_1 .. stempls.ch_tag .. NL, {k = "core:description",
            v = "Generic histogram bin min/max values lookup table"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "0"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "0"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "1"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "1"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "2"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "2"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(stempls.ch_end .. NL)
        sw:xml()
        histo_generics[bounds_type] = ch_id
    end
    return ch_id
end

-- Writes the XML to describe a channel which is used to decode
-- one of the bin bounds dimensions for a histogram.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' data needed for the XML template
-- Returns:  a new channel id created for this bin channel
function m.write_bin_channel(sw, md)
    md._sie_ch = sw:next_id('ch')
    sw:xml(stempls.priv_bin_ch .. NL, md)
    sw:xml()
    return md._sie_ch
end

-- Writes the XML for a Rainflow datamode's unclosed cycles data.
-- This occurs at the end of a test.  The cycles data is converted
-- to ASCII floating point format before writing.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' data needed for the XML template
--     'cycles' array containing unclosed cycles data, in raw 64 bit format
--     'len' number of elements in 'cycles' array
-- Returns:  a new channel id created for this bin channel
function m.write_unclosed_cycles(sw, md, cycles, len)
    local asciiCycles = {}

    -- convert cycles data
    for i = 1, len do
        asciiCycles[#asciiCycles + 1] = string.format("%f", cycles[i])
    end
    -- write XML with converted data
    md._unclosed_cycles = table.concat(asciiCycles, " ")
    sw:xml(NL .. stempls.unclosed_cycles .. NL, md)
end

return m

