local cjson = require 'cjson'
local xpath = require "xpath"
local lom = require "lxp.lom"
local log = require "log"
local xlsx = require 'xlsx_read'

local m = {}

local summaryKeys = {["Title"] = "test_title", ["Author"] = "test_author", ["Summary"] = "test_summary",
                     ["Sample rate domain"] = "sample_rate_domain", ["Test engine frame rate (Hz)"] = "pipe_rate",
                     ["Test start delay (sec)"] = "test_start_delay", ["Live updates always on"] = "live_updates",
                     ["Live updates message channels"] = "live_updates", 
                     ["Live updates computed channels"] ="live_updates",["Live updates refresh rate"] = "live_updates",
                     ["Test run mode"] = "test_run_mode", ["Cycles to run"] = "test_run_max_cycles"}

local summaryOrder = {["Title"] = 0, ["Author"] = 1, ["Summary"] = 2, ["Sample rate domain"] = 3,
                      ["Test engine frame rate (Hz)"] = 4, ["Test start delay (sec)"] = 5,
                     ["Live updates always on"] = 6, ["Live updates message channels"] = 7, 
                     ["Live updates computed channels"] = 8, ["Live updates refresh rate"]=9,
                     ["Test run mode"] = 10, ["Cycles to run"] = 11}

local channelKeys = {["Name"] = "name", ["Description"] = "description", ["Measurand"] = "measurand",
                     ["Collect"] = "collect", 
                     ["Filter Frequency"] = "digital_filter_attenuation_frequency",
                     ["Transducer power"] = "transducer_power", ["Bridge Ohms"] = "bridge_resistance", 
                     ["Leadwire Ohms"] = "bridge_lead_wire_resistance",["Electrical units"] = "electrical_units",
                     ["Units"] = "output_units", ["Range min"] = "range_min", ["Range max"] = "range_max"}

local channelOrder = {["name"] = 4,["description"] = 5,["measurand"] = 6,["collect"] = 7,
                     ["digital_filter_attenuation_frequency"] = 11, 
                     ["transducer_power"] = 15,["bridge_resistance"] = 19,["bridge_lead_wire_resistance"] = 20, 
                     ["electrical_units"] = 33, ["output_units"] = 34, ["range_min"] = 49, ["range_max"] = 50}

local otherCategories = {["Module type"] = 0, ["Module SN"] = 1, ["Connector"] = 2, ["Sample rate"] = 9, 
                         ["Filter type"]=10,["Speed mode"] = 12, ["Input mode"] = 14, ["Threshold mode"] = 16,
                         ["Low threshold (mV)"] = 17, 
                         ["High threshold (mV)"] = 18,["Excitation mode"] = 21, ["Excitation range"] = 22,
                         ["Max electrical"] = 23, ["Quadrature"] = 24,
                         ["Encoder index reset"] = 25, ["Encoder index divisor"] = 26, ["SSI coding"] = 27, 
                         ["SSI clock"] = 28, ["SSI bits"] = 29, ["Thermocouple type"] = 30, ["Scaling mode"] = 32,
                         ["Electrical units"] = 33, ["Electrical 1"] = 35, ["Electrical 2"] = 37, ["Physical 1"] = 36,
                         ["Physical 2"] = 38, ["Electrical zero"] = 39, ["Electrical span"] = 40, 
                         ["Physical span"] = 41, ["Slope"] = 42, ["Intercept"] = 43, ["Strain gage factor"] = 44,
                         ["Bridge factor"] = 45, ["Shunt resistor (kOhm)"] = 46, ["Shunt polarity"] = 47,
                         ["% Dev"] = 48, ["Prerun zero mode"] = 51, ["Prerun zero offset"] = 52, 
                         ["Prerun zero target"] = 53, ["Chart type"] = 55, ["Decimals"] = 56}

local function setDefaults(summaryTable)
    summaryTable.id = "summary"
    summaryTable.data_storage_format = "SIE"
    summaryTable.delete_old_data_files_on_test_initialization = false
    summaryTable.sort_order = 0
    summaryTable.setup_file_version = "v1.23"
    if summaryTable.sample_rate_domain == "decimal" then
        summaryTable.sample_rate_correction_factor = "1.0"
    else 
        summaryTable.sample_rate_correction_factor = '8388608/8386425'
    end
    summaryTable.channel_sort_direction = "ASC"
    summaryTable.channel_sort_direction_calibration = "ASC"
    summaryTable.channel_sort_direction_computed_channels = "ASC"
    summaryTable.channel_sort_field = "sort_order"
    summaryTable.channel_sort_field_calibration = "sort_order"
    summaryTable.channel_sort_field_computed_channels = "sort_order"
    summaryTable.data_modes_sort_direction = "ASC"
    summaryTable.data_modes_sort_field = "sort_order"
    if summaryTable.test_run_mode == "cyclic" then
        if not summaryTable.test_run_max_cycles then
            summaryTable.test_run_max_cycles = 0
        end
    end
    return summaryTable
end

local function grabSummary(worksheet)
    local summaryTable = {}
    local updatesTable = {}
    local previousK = 0
    for k,v in pairs(worksheet) do --every row
        if k ~= 1 then             --except the first
            for k2,v2 in pairs(v) do-- every cell in row
                if worksheet[1][k2] == "Summary" then
                    if not summaryTable["test_summary"] then
                        summaryTable["test_summary"] = v2
                    elseif k - previousK > 1 then
                        summaryTable["test_summary"] = summaryTable["test_summary"] .. "\n\n" .. v2
                    else
                        summaryTable["test_summary"] = summaryTable["test_summary"] .. "\n" .. v2
                    end
                elseif worksheet[1][k2] == "Sample rate domain" then
                    if v2 == "Decimal" then
                        summaryTable["sample_rate_domain"] = "decimal"
                    else
                        summaryTable["sample_rate_domain"] = "hbm_legacy"
                    end
                elseif worksheet[1][k2] == "Test start delay (sec)" then
                    summaryTable["test_start_delay"] = tonumber(v2) + 3
                elseif worksheet[1][k2] == "Live updates always on" then
                    updatesTable.always_on = v2
                elseif worksheet[1][k2] == "Live updates message channels" then
                    updatesTable.include_msg = v2
                elseif worksheet[1][k2] == "Live updates computed channels" then
                    updatesTable.include_computed = v2
                elseif worksheet[1][k2] == "Live updates refresh rate" then
                    updatesTable.update_rate = tonumber(v2)
                elseif worksheet[1][k2] == "Test run mode" then
                    summaryTable[summaryKeys[worksheet[1][k2]]] = string.lower(v2)
                elseif worksheet[1][k2] == "Cycles to run" then 
                    summaryTable[summaryKeys[worksheet[1][k2]]] = tonumber(v2)
                else
                    summaryTable[summaryKeys[worksheet[1][k2]]] = v2
                end
            end
            previousK = k
        end
    end
    summaryTable.live_updates = updatesTable
    summaryTable = setDefaults(summaryTable)
    return summaryTable
end

local function inferModuleCategory(channel, connectorType)
    if string.match(connectorType, "cx") then
        channel.category = "source/hbm_cx23r/dio_input"
    elseif string.match(connectorType, "dx") then
        channel.category = "source/hbm_dx23r/dio_input"
    elseif string.match(connectorType, "mx1601") then
        channel.category = "source/hbm_mx1601b"
    elseif string.match(connectorType, "mx1609") then
        channel.category = "source/hbm_mx1609b"
    elseif string.match(connectorType, "mx1615") then
        channel.category = "source/hbm_mx1615b"
    elseif string.match(connectorType, "mx840") then
        channel.category = "source/hbm_mx840b"
    elseif string.match(connectorType, "mx411") then
        channel.category = "source/hbm_mx411b"
    end
end

local function parseConnector(channel, category, value) 
    if not channel.connectorTable then
        channel.connectorTable = {}
    end
    if category == "Module type"  then
        channel.connectorTable.type = value
        inferModuleCategory(channel, string.lower(value))
    elseif category == "Module SN" then
        channel.connectorTable.SN = value
    elseif category == "Connector" then
        channel.connectorTable.connector = value
    end    
    if channel.connectorTable.type and channel.connectorTable.SN and channel.connectorTable.connector then
        if channel.category == "source/hbm_cx23r/dio_input" or channel.category == "source/hbm_dx23r/dio_input" then
            channel.connector = channel.connectorTable.connector .. "." .. channel.connectorTable.type.. "-" .. channel.connectorTable.SN
        else
            channel.connector = "@" .. channel.connectorTable.connector .. "." .. channel.connectorTable.type.. "-" .. channel.connectorTable.SN
        end
        channel.connectorTable = nil
    end
end

local function setSampleRate(channel, value, isClassic)
    channel.sample_rate_legacy_uncorrected = value
    if isClassic then
        channel.sample_rate_legacy_uncorrected = tostring(value)
        value = value * 8388608
        channel.sample_rate = tostring(value) .. "/8386425"
    else
        channel.sample_rate_legacy_uncorrected = tostring(value)
        channel.sample_rate = value
    end
end

local function setInputMode(channel, value)
    if value == "Thermocouple" then
        channel.sensor_input.mode = "thermocouple"
    elseif value == "60V" then
        channel.sensor_input.mode = "volt"
        channel.sensor_input.max = 60
        channel.sensor_input.min = -60
    elseif value == "10V" then 
        channel.sensor_input.mode = "volt"
        channel.sensor_input.max = 10
        channel.sensor_input.min = -10
    elseif value == "100mV" then
        channel.sensor_input.mode = "volt"
        channel.sensor_input.max = .1 
        channel.sensor_input.min = -.1
    elseif value == "IEPE" then
        channel.sensor_input.mode = "iepe"
    elseif value == "4-20mA" then
        channel.sensor_input.mode = "4-20ma"
    elseif value == "Bridge full 6 Wire" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "full"
        channel.sensor_input.wires = 6
        channel.sensor_input.regulation = true
    elseif value == "Bridge half 5 Wire" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "half"
        channel.sensor_input.wires = 5
        channel.sensor_input.regulation = true
    elseif value == "Bridge quarter 4 Wire" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "quarter"
        channel.sensor_input.wires = 4
        channel.sensor_input.regulation = true
    elseif value == "Bridge quarter 3 Wire" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "quarter"
        channel.sensor_input.wires = 3
        channel.sensor_input.regulation = true
    elseif value == "Bridge full 4 Wire (no sense lines)" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "full"
        channel.sensor_input.wires = 4
        channel.sensor_input.regulation = false
    elseif value == "Bridge half 3 Wire (no sense lines)" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "half"
        channel.sensor_input.wires = 3
        channel.sensor_input.regulation = false
    elseif value == "Bridge quarter 3 Wire (no sense lines)" then
        channel.sensor_input.mode = "bridge"
        channel.sensor_input.type = "quarter"
        channel.sensor_input.wires = 3
        channel.sensor_input.regulation = false
    elseif value == "Potentiometer" then
        channel.sensor_input.mode = "potentiometer"
    elseif value == "PT100" then
        channel.sensor_input.mode = "pt100"
    elseif value == "PT1000" then
        channel.sensor_input.mode = "pt1000"
    elseif value == "Resistor" then
        channel.sensor_input.mode = "resistor"
    elseif value == "Inductive bridge full 6 Wire" then
        channel.sensor_input.mode = "inductive_bridge"
        channel.sensor_input.type = "full"
        channel.sensor_input.wires = 6
    elseif value == "Inductive bridge half 5 Wire" then
        channel.sensor_input.mode = "inductive_bridge"
        channel.sensor_input.type = "half"
        channel.sensor_input.wires = 5
    elseif value == "LVDT" then
        channel.sensor_input.mode = "lvdt"
    elseif value == "Pulse frequency" then
        channel.sensor_input.mode = "pulse_frequency"
    elseif value == "Encoder frequency" then
        channel.sensor_input.mode = "encoder_frequency"
    elseif value == "Pulse counter" then
        channel.sensor_input.mode = "pulse_counter"
    elseif value == "Encoder counter" then
        channel.sensor_input.mode = "encoder_counter"
    elseif value == "SSI encoder counter" then
        channel.sensor_input.mode = "ssi_encoder_counter"
    end
end

local function setSensorInput(channel, category, value)
    if not channel.sensor_input then channel.sensor_input = {} end
    if category == "Input mode" then
        setInputMode(channel, value)
    elseif category == "Max electrical" then
        channel.sensor_input.max = value
        channel.sensor_input.min = -1 * tonumber(value)
    elseif category == "Quadrature" then
        channel.sensor_input.quadrature = value
    elseif category == "Encoder index reset" then
        channel.sensor_input.encoder_index_reset = value
    elseif category == "Encoder index divisor" then
        channel.sensor_input.encoder_index_divisor = value
    elseif category == "SSI coding" then
        value = value:gsub("^%u", string.lower)
        channel.sensor_input.ssi_coding = value
    elseif category == "SSI clock" then
        local _
        _, _, value = value:find("^(.+) kHz")
        channel.sensor_input.ssi_clock = value
    elseif category == "SSI bits" then
        channel.sensor_input.ssi_bits = value
    elseif category == "Thermocouple type" then
        channel.sensor_input.type = string.match(value, "Type (.+)$"):lower()
    end
end

local function setThreshold(channel, category, value)
    if not channel.threshold then channel.threshold = {} end
    if category == "Low threshold (mV)" then
        channel.threshold.min = value
    elseif category == "High threshold (mV)" then
        channel.threshold.max = value
    elseif category == "Threshold mode" then
        channel.threshold.mode = value
    end
    if not channel.threshold.mode then
        channel.threshold.mode = "static"
    end
end

local function setCalibrationMode(channel, value)
    if value == "Defined slope intercept" then
        channel.calibration[1].mode = "defined_slope_intercept"
    elseif value == "Defined zero span" then
        channel.calibration[1].mode = "defined_zero_span"
    elseif value == "Defined two point" then
        channel.calibration[1].mode = "defined_two_point"
    elseif value == "Experimental two point" then
        channel.calibration[1].mode = "experimental_two_point"
    elseif value == "Shunt" then
        channel.calibration[1].mode = "shunt"
    elseif value == "Strain gage" then
        channel.calibration[1].mode = "strain_gage"
    elseif value == "Internal" then
        channel.calibration[1].mode = "internal"
    else
        channel.calibration[1].mode = value
    end
end

local function setCalibration(channel, category, value)
    if not channel.calibration then
        channel.calibration = {}
        table.insert(channel.calibration, {})
        channel.calibration[1].purpose = "sensor_scaling"
    end
    if category == "Scaling mode" then
        setCalibrationMode(channel, value)
--        channel.calibration[1].mode = value
    elseif category == "Electrical 1" then
        if not channel.calibration[1].electrical then channel.calibration[1].electrical = {} end
        channel.calibration[1].electrical[1] = value
    elseif category == "Electrical 2" then
        if not channel.calibration[1].electrical then channel.calibration[1].electrical = {} end
        channel.calibration[1].electrical[2] = value
    elseif category == "Physical 1" then
        if not channel.calibration[1].physical then channel.calibration[1].physical = {} end
        channel.calibration[1].physical[1] = value
    elseif category == "Physical 2" then
        if not channel.calibration[1].physical then channel.calibration[1].physical = {} end
        channel.calibration[1].physical[2] = value
    elseif category == "Electrical zero" then
        channel.calibration[1].electrical_zero = value
    elseif category == "Electrical span" then
        channel.calibration[1].electrical_span = value
    elseif category == "Physical span" then
        channel.calibration[1].physical_span = value
    elseif category == "Slope" then
        channel.calibration[1].slope = value
    elseif category == "Intercept" then
        channel.calibration[1].intercept = value
    elseif category == "Strain gage factor" then    
        channel.calibration[1].gage_factor = value
    elseif category == "Bridge factor" then
        channel.calibration[1].bridge_factor = value
    elseif category == "Shunt resistor (kOhm)" then
        channel.calibration[1].resistor = "100k_ohm"
    elseif category == "Shunt polarity" then
        if value == "Up scale" then channel.calibration[1].polarity = "up_scale" 
        else channel.calibration[1].polarity = "down_scale" end
    elseif category == "% Dev" then
        channel.calibration[1].deviation_from_ideal = value
    end
end

local function setPreZero(channel, category, value)
    if not channel.prerun_balance then channel.prerun_balance = {} end
    if category == "Prerun zero mode" then
        if value == "Interactive only" then
            channel.prerun_balance.mode = "interactive_only"
        elseif value == "Not applicable" then
            channel.prerun_balance.mode = "not_applicable"
        end
    elseif category == "Prerun zero offset" then
        channel.prerun_balance.offset = value
    elseif category == "Prerun zero target" then
        channel.prerun_balance.target_value = value
    end
end

local function setFilterType(channel, excelCategory, value)
    if value == "None" then
        channel.digital_filter_type = "none"
    elseif value == "Linear phase (FIR)" then
        channel.digital_filter_type = "linear_phase_fir"
    elseif value == "Butterworth (FIR)" then
        channel.digital_filter_type = "butterworth_fir"
    elseif value == "Butterworth (IIR)" then
        channel.digital_filter_type = "butterworth_iir"
    elseif value == "Bessel (IIR)" then
        channel.digital_filter_type = "bessel_iir"
    end
end

local function setChartType(channel, value)
    if value == "Analog Meter" then
        channel.chart_type = "gauge"
    elseif value == "Digital Meter" then
        channel.chart_type = "digital"
    elseif value == "Spectrum" then
        channel.chart_type = "spectrum"
    end
end

local function setExcitationMode(channel, value)
    if value == "DC" then
        channel.excitation_mode = "dc"
    elseif value == "AC 1200Hz square" then
        channel.excitation_mode = "ac_1200_hz_square"
    elseif value == "AC 4800Hz sine" then
        channel.excitation_mode = "ac_4800_hz_sine"
    end
end

local function setChannelTrait(channel, excelCategory, value , isClassic)
    if excelCategory == "Module type" or excelCategory == "Module SN" or excelCategory == "Connector" then 
        parseConnector(channel, excelCategory, value)
    elseif excelCategory == "Sample rate" then
        setSampleRate(channel, value, isClassic)
    elseif excelCategory == "Input mode" or excelCategory == "Max electrical" or excelCategory == "Quadrature" or 
        excelCategory == "Encoder index reset" or excelCategory == "Encoder index divisor"  or excelCategory == 
        "SSI coding" or excelCategory == "SSI clock" or excelCategory == "SSI bits" or 
        excelCategory == "Thermocouple type" then
        setSensorInput(channel, excelCategory, value)
    elseif excelCategory == "Threshold mode" or excelCategory == "Low threshold (mV)" or excelCategory == "High threshold (mV)" then
        setThreshold(channel, excelCategory, value) 
    elseif excelCategory == "Scaling mode" or excelCategory == "Electrical 1" or excelCategory == "Electrical 2"
        or excelCategory == "Physical 1" or excelCategory == "Physical 2" or excelCategory == "Electrical zero"
        or excelCategory == "Electrical span" or excelCategory == "Physical span" or excelCategory == "Slope"
        or excelCategory == "Intercept" or excelCategory == "Strain gage factor" or excelCategory == "Bridge factor"
        or excelCategory == "Shunt resistor (kOhm)" or excelCategory == "Shunt polarity" or excelCategory == "% Dev"
        then
            setCalibration(channel, excelCategory, value)
    elseif excelCategory == "Prerun zero mode" or excelCategory == "Prerun zero offset" or 
           excelCategory == "Prerun zero target"then
        setPreZero(channel, excelCategory, value)
    elseif excelCategory == "Filter type" then
        setFilterType(channel, excelCategory, value)
    elseif excelCategory == "Decimals" then
        if value ~= "Auto" then
            channel.decimals = string.lower(value)
        else channel.decimals = "auto"
        end
    elseif excelCategory == "Excitation mode" then
        setExcitationMode(channel, value)
    elseif excelCategory == "Excitation range" then
        channel.excitation_range = string.lower(value)
    elseif excelCategory == "Speed mode" then
        channel.speed_mode = string.lower(value)
    elseif excelCategory == "Chart type" then
        if value ~= "Strip Chart" then
            setChartType(channel, value)
        end
    end
    return
end

function validateAndDefault(channel)
    if channel.prerun_balance and not channel.prerun_balance.target_value then
        channel.prerun_balance.target_value = 0
    end
    if channel.prerun_balance and not channel.prerun_balance.offset then
        channel.prerun_balance.offset = 0
    end
    if channel.category == "source/hbm_mx1615b" 
    and channel.sensor_input.mode == "volt" 
    and channel.sensor_input.max == 60 then
        channel.sensor_input.max = 15
        channel.sensor_input.min = -15
    end
    if channel.category == "source/hbm_mx1615b" 
    or channel.category == "source/hbm_mx1601b" 
    or channel.category == "source/hbm_mx1609b" 
    or channel.category == "source/hbm_mx840b" 
    or channel.category == "source/hbm_mx411b" then
        channel.data_type = "sequential_float32"
    else
        channel.data_type = "sequential_float64"
    end
    if channel.category == "source/hbm_mx1609b"
    or (channel.sensor_input and (channel.sensor_input.mode == "pt100"
    or channel.sensor_input.mode == "pt1000")) then
        channel.source_physical_units = "℃"
    elseif channel.sensor_input and channel.sensor_input.mode == "strain_gage" then
        channel.source_physical_units = "µm/m"
    else
        channel.source_physical_units = channel.output_units
    end
end

local function grabChannels(worksheet, test, isClassic)
    for k,v in pairs(worksheet) do --every row
        if k ~= 1 then             --except the first
            local channel = {}
            for k2,v2 in pairs(v) do-- every cell in row except row 1
                local excelCategory = worksheet[1][k2]
                if channelKeys[excelCategory] then
                    channel[channelKeys[excelCategory]] = v2
                else
                    setChannelTrait(channel, excelCategory, v2, isClassic)   
                end
            end
            channel.sort_order = k
            validateAndDefault(channel)
            table.insert(test,channel)
        end
    end
    return test
end

local function writeSxr(testTable)
    io.output("/tmp/xlsxoutput.sxr")
    io.write(cjson.encode(testTable))
    io.flush()
end

local function setWidth(column, widths, value)
    if widths then
        value = tostring(value)
        if widths[column] < string.len(value) then widths[column] = string.len(value) end
    end
end

local function setWidths(widths, sheet)
    for k,v in pairs(widths) do
        sheet:set_column(k, k, v + 5)
    end
end

local function writeSummaryColumn(w, column, value)
    local i = 1
    local longText = ""
    local cellText
    local indexStart, indexEnd
    value = value .. "\n"
    for line in string.gmatch(value, "(.-)\n") do
        w:write(i, column, line)
        i = i + 1
        cellText = line
        indexStart, indexEnd = string.find(value, line)
        if string.len(line)>string.len(longText) then longText = line end
    end
--    if not indexStart then 
--        indexStart = 1
--        cellText = value
--    end --find the last instance of the string incase a line is repeated 
--    for i=indexStart+1, string.len(value) do
--      indexStart, indexEnd = string.find(value, cellText, i)
--    if not indexStart then indexStart = i - 1 break end 
--    end
--    local lastCell = string.sub(value, indexStart) --because lua doesn't accept | in patterns
--    if lastCell and string.len(lastCell) > string.len(longText) then longText = lastCell end
--    w:write(i,column, lastCell)
    return longText
end

local function writeSummary(w, t, colors)
    for k,v in pairs(summaryKeys) do
        w:write(0, summaryOrder[k], k, colors.grey)
        if k == "Sample rate domain" then
            if t[v] == "decimal" then
                w:write(1,summaryOrder[k], "Decimal")
                t[v] = "Decimal"
            else
                w:write(1,summaryOrder[k], "Classic")
                t[v] = "Classic"
            end
        elseif k == "Summary" and t[v] then
            t[v] = writeSummaryColumn(w, summaryOrder[k], t[v]) --return for length purposes
        elseif k == "Test start delay (sec)" then
            if t[v] then
                w:write(1, summaryOrder[k], tonumber(t[v]) - 3)
            else
                w:write(1, summaryOrder[k], tonumber(0))
            end
        elseif k == "Live updates always on" then
            w:write(1, summaryOrder[k], t[v].always_on)
        elseif k == "Live updates message channels" then
                w:write(1, summaryOrder[k], t[v].include_msg)
        elseif k == "Live updates computed channels" then
                w:write(1, summaryOrder[k],t[v].include_computed)
        elseif k == "Live updates refresh rate" then
            w:write(1, summaryOrder[k], tonumber(t[v].update_rate))
        elseif k == "Test run mode" then
            local capped = ""
            if t[v] then
                capped = t[v]:sub(1,1):upper()..t[v]:sub(2)
            else
                capped = "Normal"
            end
            w:write(1, summaryOrder[k], capped)
        elseif k == "Cycles to run" then
            w:write(1, summaryOrder[k], tonumber(t[v]))
        else
            if t[v] then
                w:write(1, summaryOrder[k], t[v])
            end
        end
        if string.len(k) > string.len(tostring(t[v])) then 
            w:set_column(summaryOrder[k], summaryOrder[k], string.len(k) + 5)
        else 
            w:set_column(summaryOrder[k], summaryOrder[k], string.len(tostring(t[v])) + 5)
        end
    end
end

local function writeColumnHeadings(sheet, colors, widths)
    for k,v in pairs(channelKeys) do
        if channelOrder[v] then
            if channelOrder[v] >= 0 and channelOrder[v] < 8 then
                sheet:write(0, channelOrder[v], k, colors.grey)
            elseif channelOrder[v] == 9 then
                sheet:write(0, channelOrder[v], k, colors.purple)
            elseif channelOrder[v] > 9 and channelOrder[v] < 13 then
                sheet:write(0, channelOrder[v], k, colors.lightPurple)
            elseif channelOrder[v] == 14 then
                sheet:write(0, channelOrder[v], k, colors.brown)
            elseif channelOrder[v] > 14 and channelOrder[v] < 31 then
                sheet:write(0, channelOrder[v], k, colors.lightBrown)
            elseif channelOrder[v] == 32 then
                sheet:write(0, channelOrder[v], k, colors.green)
            elseif channelOrder[v] > 32 and channelOrder[v] < 54 then
                sheet:write(0, channelOrder[v], k, colors.lightGreen)
            elseif channelOrder[v] == 55 then
                sheet:write(0, channelOrder[v], k, colors.blue)
            elseif channelOrder[v] == 56 then
                sheet:write(0, channelOrder[v], k, colors.lightBlue)
            end
            setWidth(channelOrder[v], widths, k)
        end
    end
    for k,v in pairs(otherCategories) do
        if v >= 0 and v < 8 then
            sheet:write(0, v, k, colors.grey)
        elseif v == 9 then
            sheet:write(0, v, k, colors.purple)
        elseif v > 9 and v < 13 then
            sheet:write(0, v, k, colors.lightPurple)
        elseif v == 14 then
            sheet:write(0, v, k, colors.brown)
        elseif v > 14 and v < 31 then
            sheet:write(0, v, k, colors.lightBrown)
        elseif v == 32 then
            sheet:write(0, v, k, colors.green)
        elseif v > 32 and v < 54 then
            sheet:write(0, v, k, colors.lightGreen)
        elseif v == 55 then
            sheet:write(0, v, k, colors.blue)
        elseif v == 56 then
            sheet:write(0, v, k, colors.lightBlue)
        end
        setWidth(v, widths, k)
    end
end

local function writeDeviceName(sheet, value, row, widths)
    local connector, deviceType, SN = string.match(value, "(.*@.*)%.(.+)-(.*)")
    if string.match(connector, "^@.*") then
        connector = string.sub(connector, 2)
    end
    sheet:write(row, 0, deviceType)
    setWidth(0, widths, deviceType)
    sheet:write(row, 1, SN)
    setWidth(1, widths, SN)
    sheet:write(row, 2, connector)
    setWidth(2, widths, connector)
    
end

local function writeSensorInput(sheet, value, row, widths)
    local mode, max, bridgeType, wires, regulation, tmpType 
    tmpType = nil
    for k,v in pairs(value) do
        if k == "mode" then
            mode = v
        elseif k == "max" then
            max = tonumber(v)
        elseif k == "quadrature" then
            sheet:write(row, 24, v)
            setWidth(24, widths, v)
        elseif k == "encoder_index_reset" then
            sheet:write(row, 25, v)
            setWidth(25, widths, v)
        elseif k == "encoder_index_divisor" then
            sheet:write(row, 26, v)
            setWidth(26, widths, v)
        elseif k == "type" then
            if mode then
                if mode == "thermocouple" then
                    sheet:write(row, 30, "Type "..v:upper())
                    setWidth(30, widths, v)
                else
                    bridgeType = v
                end
            else
                tmpType = v
            end
        elseif k == "wires" then
            wires = tonumber(v)
        elseif k == "regulation" then
            regulation = v
        elseif k == "ssi_coding" then
            v = tostring(v):gsub("^%l", string.upper) --capitalize first letter
            sheet:write(row,27,v)
            setWidth(27, widths, v)
        elseif k == "ssi_clock" then
            v = tostring(v) .. ' kHz'
            sheet:write(row,28,v)
            setWidth(28, widths, v)
        elseif k == "ssi_bits" then
            sheet:write(row,29,tonumber(v))
            setWidth(29, widths, v)
        end
    end
    if mode then
        if mode ~= "volt" then
            sheet:write(row, 23, max)
            setWidth(23, widths, max)
        end
        if mode == "thermocouple" then
            sheet:write(row,14, "Thermocouple")
            setWidth(14, widths, "Thermocouple")
            if tmpType then 
                sheet:write(row, 30, tmpType)
                setWidth(30, width, tmpType)
            end
        elseif mode == "iepe" then
            sheet:write(row,14, "IEPE")
            setWidth(14, widths, "IEPE")
        elseif mode == "4-20ma" then
            sheet:write(row,14, "4-20 mA")
            setWidth(14, widths, "4-20 mA")
        elseif mode == "lvdt" then
            sheet:write(row, 14, "LVDT")
            setWidth(14, widths, "LVDT")
        elseif mode == "pulse_frequency" then
            sheet:write(row, 14, "Pulse frequency")
            setWidth(14, widths, "Pulse Frequency")
        elseif mode == "encoder_frequency" then
            sheet:write(row, 14, "Encoder frequency")
            setWidth(14, widths, "Encoder Frequency")
        elseif mode == "pulse_counter" then
            sheet:write(row, 14, "Pulse counter")
            setWidth(14, widths, "Pulse counter")
        elseif mode == "encoder_counter" then
            sheet:write(row, 14, "Encoder counter")
            setWidth(14, widths, "Encoder counter")
        elseif mode == "ssi_encoder_counter" then
            sheet:write(row, 14, "SSI encoder counter")
            setWidth(14, widths, "SSI encoder counter")
        elseif mode == "potentiometer" then
            sheet:write(row, 14, "Potentiometer")
            setWidth(14, widths, "Potentiometer")
        elseif mode == "pt100" then
            sheet:write(row, 14, "PT100")
            setWidth(14, widths, "PT100")
        elseif mode == "pt1000" then
            sheet:write(row, 14, "PT1000")
            setWidth(14, widths, "PT1000")
        elseif mode == "volt" and max then
            if max >= 15 then
                sheet:write(row,14, "60V")
                setWidth(14, widths, "60V")
            elseif max == 10 then
                sheet:write(row,14,"10V")
                setWidth(14, widths, "10V")
            elseif max == .1 then
                sheet:write(row,14,"100mV")
                setWidth(14, widths, "100mV")
            end
        elseif mode == "inductive_bridge" and bridgeType then
            if bridgeType == "full" then
                sheet:write(row,14, "Inductive bridge full 6 Wire")
                setWidth(14, widths, "Inductive bridge full 6 wire")
            elseif bridgeType == "half" then
                sheet:write(row,14, "Inductive bridge half 5 Wire")
                setWidth(14, widths, "Inductive bridge half 5 wire")
            end
        elseif mode == "bridge" and (bridgeType or tmpType) and wires then
            if tmpType then bridgeType = tmpType end
            if bridgeType == "full" then
                if wires == 6 then
                    sheet:write(row, 14, "Bridge full 6 Wire")
                    setWidth(14, widths, "Bridge full 6 wire")
                elseif wires == 4 then
                    sheet:write(row, 14, "Bridge full 4 Wire (no sense lines)")
                    setWidth(14, widths, "Bridge full 4 wire (no sense lines)")
                end
            elseif bridgeType == "half" then
                if wires == 5 then
                    sheet:write(row, 14, "Bridge half 5 Wire")
                    setWidth(14, widths, "Bridge half 5 wire")
                elseif wires == 3 then
                    sheet:write(row, 14, "Bridge half 3 Wire (no sense lines)")
                    setWidth(14, widths, "Bridge half 3 wire (no sense lines)")
                end
            elseif bridgeType == "quarter" then
                if wires == 4 then
                    sheet:write(row, 14, "Bridge quarter 4 Wire")
                    setWidth(14, widths, "Bridge quarter 4 wire")
                elseif wires == 3 then
                    if regulation then
                        sheet:write(row, 14, "Bridge quarter 3 Wire")
                        setWidth(14, widths, "Bridge quarter 3 wire")
                    else 
                        sheet:write(row, 14, "Bridge quarter 3 Wire (no sense lines)")
                        setWidth(14, widths, "Bridge quarter 3 wire (no sense lines)")
                    end
                end
            end
        else 
            sheet:write(row, 14, "one" )
            setWidth(14, widths, "Bridge quarter 3 wire (no sense lines)")
        end      
    else 
        sheet:write(row, 14, "two")
        setWidth(14, widths, "Bridge quarter 3 wire (no sense lines)")
    end
end

local function writeThreshold(sheet, value, row, widths)
    for k,v in pairs(value) do
        if tonumber(v) then v = tonumber(v) end
        if v == cjson.null then v = nil end
        if k == "min" then
            sheet:write(row, 17, v)
            setWidth(17, widths, v)
        elseif k == "max" then
            sheet:write(row, 18, v)
            setWidth(18, widths, v)
        elseif k == "mode" then
            sheet:write(row, 16, v)
            setWidth(16, widths, v)
        end
    end
end
    
local function writeCalibrationMode(sheet, value, row, widths)
    if value == "defined_slope_intercept" then
        sheet:write(row, 32, "Defined slope intercept")
    elseif value == "defined_zero_span" then
        sheet:write(row, 32, "Defined zero span")
    elseif value == "defined_two_point" then
        sheet:write(row, 32, "Defined two point")
    elseif value == "experimental_two_point" then
        sheet:write(row, 32, "Experimental two point")
    elseif value == "shunt" then
        sheet:write(row, 32, "Shunt")
    elseif value == "strain_gage" then
        sheet:write(row, 32, "Strain gage")
    elseif value == "internal" then
        sheet:write(row, 32, "Internal")
    else
        sheet:write(row, 32, value)
    end
end

local function writeCalibration(sheet, value, row, widths)
    for k,v in pairs(value) do
        if tonumber(v) then v = tonumber(v) end
        if v == cjson.null then v = nil end
        if k == "mode" then
            writeCalibrationMode(sheet, v, row)
        elseif k == "electrical" then
            sheet:write(row, 35, tonumber(v[1]))
            sheet:write(row, 37, tonumber(v[2]))
        elseif k == "physical" then 
            sheet:write(row, 36, tonumber(v[1]))
            sheet:write(row, 38, tonumber(v[2]))
        elseif k == "electrical_zero" then
            sheet:write(row, 39, v)
        elseif k == "electrical_span" then
            sheet:write(row, 40, v)
        elseif k == "physical_span" then
            sheet:write(row, 41, v)
        elseif k == "slope" then
            sheet:write(row, 42, v)
        elseif k == "intercept" then
            sheet:write(row, 43, v)
        elseif k == "gage_factor" then
            sheet:write(row, 44, v)
        elseif k == "bridge_factor" then
            sheet:write(row, 45, v)
        elseif k == "resistor" then
            sheet:write(row, 46, "100 kOhms")
        elseif k == "polarity" then
            if v == "up_scale" then sheet:write(row, 46, "Up scale")
            else sheet:write(row,47, "Down scale") end            
        elseif k == "deviation_from_ideal" then
            sheet:write(row, 48, v)
        end
    end
end

local function writePrerunBalance(sheet, value, row, widths)
    local hasOffset, hasTarget = false, false
    for k,v in pairs(value) do
        if v == cjson.null then v = nil end
        if k == "mode" then
            if v == "interactive_only" then
                sheet:write(row, 51, "Interactive only")
            else
                sheet:write(row,51, "Not applicable")
            end
        elseif k == "offset" then
            hasOffset = true
            sheet:write(row, 52, tonumber(v))
        elseif k == "target" then
            hasTarget = true
            sheet:write(row, 53, tonumber(v))
        end
    end
    if not hasOffset then sheet:write(row, 52, 0) end
    if not hasTarget then sheet:write(row, 53, 0) end
end

local function writeFilterType(sheet, value, row, widths)
    if value == "none" then
        sheet:write(row, 10, "None") 
    elseif value == "linear_phase_fir" then
        sheet:write(row, 10, "Linear phase (FIR)") 
    elseif value == "butterworth_fir" then
        sheet:write(row, 10, "Butterworth (FIR)") 
    elseif value == "butterworth_iir" then
        sheet:write(row, 10, "Butterworth (IIR)") 
    elseif value == "bessel_iir" then
        sheet:write(row, 10, "Bessel (IIR)") 
    end
end

local function writeChartType(sheet, value, row, widths)
    if value == "gauge" then
        sheet:write(row, 55, "Analog Meter")
    elseif value == "digital" then
        sheet:write(row, 55, "Digital Meter")
    elseif value == "spectrum" then
        sheet:write(row,55, "Spectrum")
    else 
        sheet:write(row, 55, "Strip Chart")
    end
end 

local function writeExcitationMode(sheet, value, row, widths)
    if value == "dc" then
        sheet:write(row,21, "DC")
    elseif value == "ac_1200_hz_square" then
        sheet:write(row, 21, "AC 1200Hz square")
    elseif value == "ac_4800_hz_sine" then
        sheet:write(row, 21, "AC 4800Hz sine")
    end
end

local function writeChannels(sheet, table, row, colors, widths)
    local hasChart = false
    for k,v in pairs(table) do --for each key
        if tonumber(v) then v = tonumber(v) end
        if v == cjson.null then v = nil end
        if k == "sensor_input" then
            writeSensorInput(sheet,v,row, widths)
        elseif k == "threshold" then
            writeThreshold(sheet,v,row,widths)
        elseif k == "calibration" then
            writeCalibration(sheet,v[1],row, widths)
        elseif k == "prerun_balance" then
            writePrerunBalance(sheet,v,row, widths)
        elseif k == "connector" then
            writeDeviceName(sheet, v, row, widths)
        elseif k == "digital_filter_type" then
            writeFilterType(sheet, v, row, widths)
        elseif k == "sample_rate_legacy_uncorrected" then
            sheet:write(row, 9, v)
            setWidth(9, widths, v)
        elseif k == "decimals" then
            if v ~= "auto" then
                sheet:write(row, 56, string.upper(v))
                setWidth(56,widths, v)
            else 
                sheet:write(row, 56, "Auto")
                setWidth(56,widths, "Auto")
            end
        elseif k == "chart_type" then
            hasChart = true
            writeChartType(sheet, v, row, widths)
        elseif k == "excitation_mode" then
            writeExcitationMode(sheet, v, row, widths)
        elseif k == "excitation_range" then
            sheet:write(row,22, string.upper(v))
            setWidth(22,widths, v)
        elseif k == "speed_mode" then
            if v == "high" then
                sheet:write(row, 12, "High")
                setWidth(12, widths, "High")
            else
                sheet:write(row, 12, "Normal")
                setWidth(12, widths, "Normal")
            end
        elseif channelOrder[k] then
            sheet:write(row, channelOrder[k], v)
            setWidth(channelOrder[k],widths, v)
        end
    end
    if not hasChart then sheet:write(row, 55, "Strip Chart") end
end

local function print_xlsx(t, fileName)
    local Workbook = require "xlsxwriter.workbook"
    local workbook  = Workbook:new(fileName)
    local widths = {}
    local colors = {grey = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#ecedef"}),
    purple = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#b793d9"}),
    lightPurple = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#d6c8e4"}),
    brown = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#d8b490"}),
    lightBrown = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#e3d4c7"}),
    green = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#8fd897"}),
    lightGreen = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#c6e3c9"}),
    blue = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#93b4d9"}),
    lightBlue = workbook:add_format({bold = true, border = 1, border_color = "#ededed", pattern = 1, bg_color = "#c7d5e3"}),
    rowOdd = workbook:add_format({pattern = 1, border = 1, border_color = "#ededed", bg_color = "#e9f2f9"}),
    rowEven = workbook:add_format({border = 1, border_color = "#ededed", pattern = 1, bg_color = "#d2e0f0"})
    }
    local summaryWorksheet = workbook:add_worksheet("Test Setup")
    local worksheet = workbook:add_worksheet("Channels")
    for i =0,56 do
        widths[i] = 0
    end
    writeColumnHeadings(worksheet, colors, widths)
    local deadRows = 0
    for k,v in pairs (t) do
        k = tonumber(k) - 1
        v.category = tostring(v.category)
        v.connector = tostring(v.connector)
        v.isCan = v.connector:find("@can") or v.category:find("can") or v.category:find("mx471")   
        if k == 0 then
            writeSummary(summaryWorksheet, v, colors)
        else
            if v.category:find("^source") and not v.isCan then 
                if (k-deadRows)%2 == 0 then
                    worksheet:set_row(k-deadRows, 15, colors.rowEven)
                else worksheet:set_row(k-deadRows, 15, colors.rowOdd) 
                end
                writeChannels(worksheet, v, k-deadRows, colors, widths)
            else deadRows = deadRows + 1
            end
        end
    end
    setWidths(widths, worksheet)
    workbook:close()
end

-- Converts a Microsoft Excel .xlsx file to a test setup specification.
--
-- Parameters:
--     filename - A temporary filename from which .xlsx data should be read.
--
-- Returns:
--     On success, returns true, setup_spec.
--     On failure, returns false, error_message.

function m.import(filename)
    local workbook = xlsx.open(filename)
    local test = {}
    local summary = grabSummary(workbook[1])
    table.insert(test,summary)
    local isClassic = summary.sample_rate_domain == "hbm_legacy"
    grabChannels(workbook[2],test, isClassic)
    local sxr = writeSxr(test)
    return true, test, workbook --sxr
end

-- Converts a test setup specification into a Microsoft Excel .xlsx file.
--
-- Parameters:
--     setup - A Lua table containing the test setup.
--     filename - A temporary filename to which .xlsx data should be written.
--
-- Returns:
--     On success, returns true. On failure, returns false, error_message.

function m.export(setup, filename)
    print_xlsx(setup, filename)
    return true
end

return m
